!> \file
!! Contains subroutines cg(), vec_prod() and cal_dif()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine CG
!
!> This routine performs a preconditioned conjugate gradient
!!  method to solve AX=B
!!
!! Warning: the boundary conditions on first guess X must be
!! filled through the call to limites before calling cg
!!
!<
subroutine cg(X,B,icl,mat_prod,precond,cmp_precond,eps,nite,dif)

  use parameters, only : ndim, nxmin, nxmax
  use para      , only : mype
  use cputime   , only : temps_comm

  implicit none

  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: X, B
  integer, dimension(6) :: icl

  logical,intent(in ) :: precond !< if .true. perform matrix preconditionning
  real   ,intent(in ) :: eps     !< convergence criterion
  real   ,intent(out) :: dif     !< residual norm after convergence
  integer,intent(out) :: nite    !< number of iterations reached

  interface
     subroutine mat_prod(X,AX)
       use parameters, only : nxmin, nxmax
       implicit none
       real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: X, AX
     end subroutine mat_prod
     subroutine cmp_precond(R,Z)
       use parameters, only : nxmin, nxmax
       implicit none
       real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: R, Z
     end subroutine cmp_precond
  end interface

  integer               :: nitemax
  integer, dimension(3) :: ii

  logical :: BC

  real :: alph,pv1,pv2,pv3

  real, dimension(:,:,:), allocatable :: RR, PP, AX, ZZ
  real, dimension(1,1:2*ndim) :: X_l

  allocate(RR(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)))
  allocate(PP(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)))
  allocate(AX(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)))
  allocate(ZZ(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)))

  call Mat_prod(X,AX)
  RR = B - AX
  if(precond) then
     call cmp_precond(RR,ZZ)
     PP = ZZ
  else
     PP = RR
  endif

  BC = .true. ; nite = 0 ; nitemax = 1000 ; dif = 0.

  X_l = 0.

  do while (BC)
     
     nite = nite + 1

     call communications_gen(PP,1,temps_comm)
     call limites_gen(PP,icl,X_l,1)

     call Mat_prod(PP,AX)
     call Vec_prod(pv1,PP,AX)
     if(precond) then
        call Vec_prod(pv2,RR,ZZ)
     else
        call Vec_prod(pv2,RR,RR)
     endif

     if (pv1 .ne. 0.) then
        alph = pv2/pv1
     else
        write(*,*) 'pv1 = 0.'
        exit
     endif

     X = X + alph*PP
 
     call cal_dif(dif,X,PP,alph)

     RR = RR - alph*AX

     if(precond) then
        call cmp_precond(RR,ZZ)
        call Vec_prod(pv3,RR,ZZ)
        PP = ZZ + pv3/pv2 * PP
     else
        call Vec_prod(pv3,RR,RR)
        PP = RR + pv3/pv2 * PP
     endif
     
     if ((dif < eps .and. nite .ge. 2).or.nite.eq.nitemax) BC = .false.

  enddo

  if(nite.eq.nitemax) then
     ii = maxloc(abs(alph*PP/X))
     write(*,001) nite,maxval(abs(alph*PP/X)),nxmin(1)-1+ii(1),nxmin(2)-1+ii(2),nxmin(3)-1+ii(3), &
                                            X(nxmin(1)-1+ii(1),nxmin(2)-1+ii(2),nxmin(3)-1+ii(3))
001 format('nite,dif,i,j,k,X(i,j,k)= ',i4,E12.5,3(i4),E12.5)
     call close_program(mype,'Conjugate Gradient not converged: nitemax reached')
  endif

  deallocate(RR)
  deallocate(PP)
  deallocate(AX)
  deallocate(ZZ)

  return

end subroutine cg

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine VEC_PROD
!
!> Computes the vector product x1 * x2.
!<
subroutine vec_prod(pv,x1,x2)

  use parameters
  use divers    , only : verbose

  implicit none

#if PARA==1
  include 'mpif.h'
  integer :: info,un
#endif

  integer :: i,j,k

  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: x1,x2
  real :: pv,pv_loc,xx

  if (verbose) write(*,*) 'Start of vec_prod'

  pv_loc = 0. ; xx = 0.
  
  do k = 1,nx(3)
  do j = 1,nx(2)
  do i = 1,nx(1)
           
     pv_loc = pv_loc + x1(i,j,k)*x2(i,j,k)
        
  enddo
  enddo
  enddo

#if PARA==1
  info = 0 ; un = 1
  call MPI_ALLREDUCE(pv_loc,xx,un,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,info)
  pv = xx
#else
  pv = pv_loc
#endif

  if (verbose) write(*,*) 'End of vec_prod'

  return
  
end subroutine vec_prod

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine CAL_DIF
!
!> Computes the maximum relative difference for the X 
!! variable between two iterations
!<
subroutine cal_dif(dif,X,PP,alpha)

  use parameters
  use divers    , only : verbose

  implicit none

#if PARA==1
  include 'mpif.h'
  integer :: info,un
  real    :: xx
#endif

  integer :: i,j,k

  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: X,PP
  real :: dif,dif_loc,alpha

  if (verbose) write(*,*) 'Start of cal_dif'

  dif_loc = 0. ; dif = 0.
  
  do k = 1,nx(3)
  do j = 1,nx(2)
  do i = 1,nx(1)
           
     dif_loc = max(dif_loc,abs(alpha*PP(i,j,k)/X(i,j,k)))

  enddo
  enddo
  enddo

#if PARA==1
  info = 0 ; un = 1
  call MPI_ALLREDUCE(dif_loc,xx,un,MPI_DOUBLE_PRECISION,MPI_MAX,MPI_COMM_WORLD,info)
  dif = xx 
#else
  dif = dif_loc
#endif

  if (verbose) write(*,*) 'End of cal_dif'
  
  return

end subroutine cal_dif
