!> \file
!! Contains subroutines gc(), mat_prod_gra()
!! and cmp_precond_gra()
!!
!! This file contains the routines necessary to perform
!! conjugate gradients when including self-gravity in the
!! simulation.
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine GC
!
!> Performs the conjugate gradient algorithm for the
!! gravity potential calculation.
!<
subroutine gc(init)
  
  use var
  use geom      , only : x,Lbox
  use divers
  use parameters
  use unites    
  use gravity
  use cl        , only : icl
  use para      , only : mype
  use cputime   , only : temps_comm

  implicit none

  integer, dimension(6)               :: icl_gra
  integer                             :: i,j,k,nite,iface
  
  real, dimension(:,:,:), allocatable :: BB
  real                                :: dif,eps,xbar,ybar,zbar,pbar

  logical :: init,precond

  interface
     subroutine mat_prod_gra(X,AX)
       use parameters, only : nxmin, nxmax
       implicit none
       real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: X, AX
     end subroutine mat_prod_gra
     subroutine cmp_precond_gra(R,Z)
       use parameters, only : nxmin, nxmax
       implicit none
       real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: R, Z
     end subroutine cmp_precond_gra
  end interface


  if (verbose) write(*,*) 'start of potential computing by gc'

  allocate(BB(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)))

  Phi  = 0. ; BB   = 0.
  xbar = 0. ; ybar = 0. ; zbar = 0. ; pbar = 0. 

  call cal_barycentre(xbar,ybar,zbar,pbar)

  do k = 1,nx(3)
     do j = 1,nx(2)
        do i = 1,nx(1)
         
           if (init) then
              Phi(i,j,k) = - G*pbar/sqrt(Lbox(1)**2+Lbox(2)**2+Lbox(3)**2)
           else
              Phi(i,j,k) = Phiold(i,j,k)
           endif
         
           if(ISOLE == 1) then
              BB(i,j,k) = 4*Pi*G*rho(i,j,k) 
           else
              BB(i,j,k)= 4*Pi*G*(rho(i,j,k)-pbar/(lbox(1)*lbox(2)*lbox(3)))
           endif
            
        enddo
     enddo
  enddo

  call communications_gen(Phi,1,temps_comm)
  call limites_gra(Phi)

  ! Conjugate Gradient method
  eps = 1.e-6
  precond = .false.
  
  icl_gra = 0

  do iface = 1,6
     if (icl(iface).ne.0) then

        if (ISOLE==1) icl_gra(iface) = 4
        if (ISOLE==0) icl_gra(iface) = 2
        
     endif
  enddo

  call cg(phi,BB,icl_gra,mat_prod_gra,precond,cmp_precond_gra,eps,nite,dif)

  nitetot_grav = nitetot_grav + nite
  
  if (mype == 0) write(*,'(a,i4,a,es10.3,a,i9)') 'nite =',nite,'  err =',dif,'  nitetot',nitetot_grav
  
  deallocate(BB)

  if (verbose) write(*,*) 'end of potential computing by gc'

  return
  
end subroutine gc

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine MAT_PROD_GRA
!
!> Performs the matrix product.
!<
subroutine mat_prod_gra(T,AT)
      
  use var
  use divers    , only : verbose
  use parameters
  use gravity
  use geom      , only : x
  
  implicit none

  integer :: i,j,k

  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: T,AT
  real, dimension(3)                                                     :: gradp,gradm,grad2,dxc,dx
  
  do k = nxmin(3),nxmax(3)
     do j = nxmin(2),nxmax(2)
        do i = nxmin(1),nxmax(1)
        
           gradp = 0.
           gradm = 0.
           grad2 = 0.
           dxc   = 0.
           dx    = 0.
           
           dxc(1) = (x(i+1,1)-x(i-1,1))/2
           dxc(2) = (x(j+1,2)-x(j-1,2))/2
           dxc(3) = (x(k+1,3)-x(k-1,3))/2
           
           dx(1) = x(i+1,1)-x(i,1)
           dx(2) = x(j+1,2)-x(j,2)
           dx(3) = x(k+1,3)-x(k,3)
           
           gradp(1) = (T(i+1,j,k) - T(i,j,k))/dxc(1) 
           gradm(1) = (T(i,j,k) - T(i-1,j,k))/dxc(1)
           grad2(1) = (gradp(1)-gradm(1))/dx(1)
           
           gradp(2) = (T(i,j+1,k) - T(i,j,k))/dxc(2) 
           gradm(2) = (T(i,j,k) - T(i,j-1,k))/dxc(2)
           grad2(2) = (gradp(2)-gradm(2))/dx(2)
           
           gradp(3) = (T(i,j,k+1) - T(i,j,k))/dxc(3) 
           gradm(3) = (T(i,j,k) - T(i,j,k-1))/dxc(3)
           grad2(3) = (gradp(3)-gradm(3))/dx(3)
           
           AT(i,j,k) = (grad2(1)+grad2(2)+grad2(3))
       
        enddo
     enddo
  enddo
  
  return

end subroutine mat_prod_gra

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine CMP_PRECOND_GRA
!
!> This routine computes the preconditioned matrix.
!<
subroutine cmp_precond_gra(RR,ZZ)

  use parameters

  implicit none
  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: RR,ZZ

  ZZ = RR

end subroutine cmp_precond_gra
