!> \file
!! Contains subroutines init_eos(), pressure(),
!! temperature(), sound_speed(), energy() and
!! heat_capacity()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine INIT_EOS
!
!> Initialises the gas equation of state.
!<
subroutine init_eos
  
  use para  , only : mype
  use divers, only : isotherme

!  if(isotherme) then
!     if(mype==0) write(*,*) 'init_eos : Perfect gas - isothermal'
!  else
!     if(mype==0) write(*,*) 'init_eos : Perfect gas'
!  endif

  return

end subroutine init_eos

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine PRESSURE
!
!> Returns the pressure of an ideal gas array.
!<
subroutine Pressure(rho,e,P,nxmin,nxmax,nymin,nymax,nzmin,nzmax)

  use gammas, only : gamma,cs_iso2
  use divers, only : isotherme

  implicit none

  integer :: nxmin,nxmax,nymin,nymax,nzmin,nzmax
  real, dimension(nxmin:nxmax,nymin:nymax,nzmin:nzmax) :: rho, e, P

  if(isotherme) then
     P = rho*cs_iso2
  else
     P = (gamma-1.)*e
  endif

  return

end subroutine Pressure

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine TEMPERATURE
!
!> Returns the temperature of an ideal gas array.
!<
subroutine Temperature(rho,e,T,nxmin,nxmax,nymin,nymax,nzmin,nzmax)

  use gammas, only : gamma
  use unites, only : uma,kb
  use divers, only : mu,isotherme
  use param_ini, only : Temperature_isotherme

  implicit none

  integer :: nxmin,nxmax,nymin,nymax,nzmin,nzmax
  real, dimension(nxmin:nxmax,nymin:nymax,nzmin:nzmax) :: rho, e, T

  if(isotherme) then
     T = Temperature_isotherme
  else
     T = (gamma-1.)*e/rho *(mu*uma/kb)
  endif

  return

end subroutine Temperature

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine SOUND_SPEED
!
!> Returns the sound speed of an ideal gas array.
!<
subroutine Sound_speed(rho,e,cs,nxmin,nxmax,nymin,nymax,nzmin,nzmax)

  use gammas, only : gamma,cs_iso
  use divers, only : isotherme

  implicit none

  integer :: nxmin,nxmax,nymin,nymax,nzmin,nzmax
  real, dimension(nxmin:nxmax,nymin:nymax,nzmin:nzmax)  :: rho, e, cs

  if(isotherme) then
     cs = cs_iso
  else
     cs = sqrt(gamma*(gamma-1.)*e/rho)
  endif

  return

end subroutine Sound_speed

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine Energy
!
!> Returns the thermal energy of an ideal gas array.
!<
subroutine Energy(rho,T,e,nxmin,nxmax,nymin,nymax,nzmin,nzmax)

  use gammas, only : gamma
  use unites, only : uma,kb
  use divers, only : mu,isotherme
  use para  , only : mype

  implicit none

  integer :: nxmin,nxmax,nymin,nymax,nzmin,nzmax
  real, dimension(nxmin:nxmax,nymin:nymax,nzmin:nzmax) :: rho,T,e
  
  if(isotherme) then
     call close_program(mype,'Internal energy not defined in isothermal EOS...')
  else
     e = rho*kb*T/(mu*uma*(gamma-1.))
  endif

  return
  
end subroutine Energy

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine HEAT_CAPACITY
!
!> Returns the volumic heat capacity of an ideal gas array.
!<
subroutine heat_capacity(rho,T,cv,nxmin,nxmax,nymin,nymax,nzmin,nzmax)

  use gammas, only : gamma
  use unites, only : uma,kb
  use divers, only : mu,isotherme
  use para  , only : mype

  implicit none
  
  integer :: nxmin,nxmax,nymin,nymax,nzmin,nzmax
  real, dimension(nxmin:nxmax,nymin:nymax,nzmin:nzmax)  :: rho,T,cv

  if(isotherme) then
     call close_program(mype,'Heat capacity not defined in isothermal EOS...')
  else
     cv = rho*kb/(mu*uma*(gamma-1))
  endif

  return

end subroutine heat_capacity
