!> \file
!! Contains subroutine evol_hydro()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine EVOL_HYDRO
!
!> Performs the evolution of the hydrodynamic variables 
!! over one timestep. The update is done over the range
!! (imin:imax,jmin:jmax,kmin:kmax) which is not
!! necessarily the entire range of cells which the cpu
!! holds.
!!
!! The routine:
!!   - computes slopes and interface values
!!   - performs a half-timestep prediction
!!   - computes Riemann fluxes using the Riemann solver
!!   - updates the variables using computed fluxes
!!   - eventually computes explicit comoving terms for
!!     radiation coupling using computed interface velocity
!!     values.
!<
subroutine evol_hydro(imin,imax,jmin,jmax,kmin,kmax)
  
  use var
  use var_loc
  use varold
  use geom
  use divers
  use etat_GDS
  use parameters   , only : ndim,nx, nFx,slope_type
  use gammas       , only : gamma,Cs_iso2
  use cl
  use para         , only : mype
  use gravity      , only : phi
  use solvers_hydro

  implicit none
  
  integer, intent(in )  :: imin,imax,jmin,jmax,kmin,kmax

  integer, dimension(3,3) :: kron
  integer, dimension(6  ) :: icl_loc
  integer, dimension(3  ) :: indice
  integer                 :: nnx,nny,nnz,nn,nbuf
  integer                 :: i,j,k,idim,jdim,iface,id
  integer                 :: ip1,jp1,kp1,im1,jm1,km1
  integer                 :: nxmin,nxmax,nymin,nymax,nzmin,nzmax
  integer                 :: ixmin,ixmax,iymin,iymax,izmin,izmax

  real, dimension(:,:,:,:,:  ), allocatable :: RhoX,PX,EX 
  real, dimension(:,:,:,:,:  ), allocatable :: RhoXt,PXt,EXt
  real, dimension(:,:,:,:,:,:), allocatable :: UX,UXt,RhouX,RhouXt

  real, dimension(:,:,:,:,:  ), allocatable :: flux_masse,flux_ener,flux_pre
  real, dimension(:,:,:,:,:,:), allocatable :: flux_mom
#if RAYO==1 || DIF==1
  real, dimension(:,:,:,:    ), allocatable :: moment_ray
#endif

  real, dimension(:,:,:,:    ), allocatable :: prhog, ppreg
  real, dimension(:,:,:,:    ), allocatable :: pvg
  real, dimension(:,:,:,:,:  ), allocatable :: pviteg

  real, dimension(:,:,:,:    ), allocatable :: prhod, ppred
  real, dimension(:,:,:,:    ), allocatable :: pvd
  real, dimension(:,:,:,:,:  ), allocatable :: pvited

  real, dimension(:,:,:,:    ), allocatable :: pEing, pEind
  real, dimension(:,:,:,:,:  ), allocatable :: EinX, EinXt 
  real, dimension(:          ), allocatable :: ell,cll,err,crr
  
  real, dimension(:          ), allocatable :: dll,pll,drr,prr,dss,pss, ess  
  real, dimension(:,:        ), allocatable :: ull,urr,uss,fll,frr,fss

  real, dimension(3,2,3      )              :: Rayon
  real, dimension(1:ndim     )              :: flux_mom_tot,flux_pre_tot
  real, dimension(6          )              :: signe

  real :: V, drp,drm,f1,f2,Utot2,rr
  real :: Pression,cotanx,flux_masse_tot,flux_ener_tot

#if SCALP==1
  integer :: ifx
  real, dimension(:,:,:,:,:  ), allocatable :: pFxg,pFxd
  real, dimension(:,:,:,:,:,:), allocatable :: FxX, FxXt
  real, dimension(:,:,:,:,:,:), allocatable :: flux_fx
  real, dimension(1:Nfx      )              :: flux_fx_tot
#endif

  if (verbose) write(*,*) 'Begin of evol_hydro'
  if (verbose) write(*,*) 'Nfx = ',nfx

  ! Check size of the sub-arrays
  call check_size(imin,imax,jmin,jmax,kmin,kmax)

  ! Definition of icl_loc for boundary conditions on the sub-arrays
  icl_loc = 0

  if(imin.eq.1    ) icl_loc(1)=icl(1)
  if(imax.eq.nx(1)) icl_loc(2)=icl(2)
  if(jmin.eq.1    ) icl_loc(3)=icl(3)
  if(jmax.eq.nx(2)) icl_loc(4)=icl(4)
  if(kmin.eq.1    ) icl_loc(5)=icl(5)
  if(kmax.eq.nx(3)) icl_loc(6)=icl(6)


  ! Arrays allocation
  nnx = imax - imin + 1 ; nny =  jmax - jmin + 1 ; nnz = kmax - kmin + 1
  if(ndim.eq.1) then
     nxmin = -1 ; nxmax = nnx+2
     nymin =  1 ; nymax = 1
     nzmin =  1 ; nzmax = 1
     ixmin =  0 ; ixmax = nnx+1
     iymin =  1 ; iymax = 1
     izmin =  1 ; izmax = 1
  else if (ndim == 2) then
     nxmin = -1 ; nxmax = nnx+2
     nymin = -1 ; nymax = nny+2
     nzmin =  1 ; nzmax = 1
     ixmin =  0 ; ixmax = nnx+1
     iymin =  0 ; iymax = nny+1
     izmin =  1 ; izmax = 1
  else if (ndim == 3) then
     nxmin = -1 ; nxmax = nnx+2
     nymin = -1 ; nymax = nny+2
     nzmin = -1 ; nzmax = nnz+2
     ixmin =  0 ; ixmax = nnx+1
     iymin =  0 ; iymax = nny+1
     izmin =  0 ; izmax = nnz+1
  endif

  kron = 0
  do idim=1,3
     kron(idim,idim) = 1
  enddo

  nn  = max(nnx,nny,nnz)
  allocate(xloc(-1:nn+3,ndim)) 

  allocate(rholoc (nxmin:nxmax,nymin:nymax,nzmin:nzmax)       ,ploc(nxmin:nxmax,nymin:nymax,nzmin:nzmax       ))
  allocate(Eloc   (nxmin:nxmax,nymin:nymax,nzmin:nzmax)                                                        )
  allocate(rhouloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:ndim),uloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:ndim))

#if GRA==1
  if (GRAV) then
     allocate(philoc(ixmin:ixmax,iymin:iymax,izmin:izmax)) ; philoc = 0.
     endif
#endif

  allocate( prhog  (nxmin:nxmax,nymin:nymax,nzmin:nzmax,       1:ndim))
  allocate( ppreg  (nxmin:nxmax,nymin:nymax,nzmin:nzmax,       1:ndim))
  allocate (pvg    (nxmin:nxmax,nymin:nymax,nzmin:nzmax,       1:ndim))
  allocate( pviteg (nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:ndim,1:ndim))

  allocate( prhod  (nxmin:nxmax,nymin:nymax,nzmin:nzmax,       1:ndim))
  allocate( ppred  (nxmin:nxmax,nymin:nymax,nzmin:nzmax,       1:ndim))
  allocate (pvd    (nxmin:nxmax,nymin:nymax,nzmin:nzmax,       1:ndim))
  allocate( pvited (nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:ndim,1:ndim))

  allocate( rhoX (nxmin:nxmax,nymin:nymax,nzmin:nzmax,2,1:ndim))
  allocate(   EX (nxmin:nxmax,nymin:nymax,nzmin:nzmax,2,1:ndim))
  allocate(   PX (nxmin:nxmax,nymin:nymax,nzmin:nzmax,2,1:ndim))
  allocate(   uX (nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:ndim,2,1:ndim))   
  allocate(rhouX (nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:ndim,2,1:ndim))

  allocate( rhoXt(ixmin:ixmax,iymin:iymax,izmin:izmax,       2,1:ndim))
  allocate(   EXt(ixmin:ixmax,iymin:iymax,izmin:izmax,       2,1:ndim))
  allocate(   PXt(ixmin:ixmax,iymin:iymax,izmin:izmax,       2,1:ndim))
  allocate(   uXt(ixmin:ixmax,iymin:iymax,izmin:izmax,1:ndim,2,1:ndim))
  allocate(rhouXt(ixmin:ixmax,iymin:iymax,izmin:izmax,1:ndim,2,1:ndim))

  allocate(rhoS(nxmin:nxmax,nymin:nymax,nzmin:nzmax,       1:ndim))
  allocate(  pS(nxmin:nxmax,nymin:nymax,nzmin:nzmax,       1:ndim))
  allocate(  ES(nxmin:nxmax,nymin:nymax,nzmin:nzmax,       1:ndim))
  allocate(  uS(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:ndim,1:ndim))

  allocate(dll(1:nnx+1)       ,pll(1:nnx+1)       ,drr(1:nnx+1)       )
  allocate(prr(1:nnx+1)       ,dss(1:nnx+1)       ,pss(1:nnx+1), ess(1:nnx+1)       )
  allocate(ull(1:nnx+1,1:ndim),urr(1:nnx+1,1:ndim),uss(1:nnx+1,1:ndim))
  allocate(fll(1:nnx+1,1:nFx ),frr(1:nnx+1,1:nFx ),fss(1:nnx+1,1:nFx ))

  allocate( Einloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax         ))
  allocate(pEind  (nxmin:nxmax,nymin:nymax,nzmin:nzmax,  1:ndim))
  allocate(pEing  (nxmin:nxmax,nymin:nymax,nzmin:nzmax,  1:ndim))
  allocate( EinX  (nxmin:nxmax,nymin:nymax,nzmin:nzmax,2,1:ndim))
  allocate( EinXt (ixmin:ixmax,iymin:iymax,izmin:izmax,2,1:ndim))
  allocate(cll(1:nnx+1),ell(1:nnx+1),crr(1:nnx+1),err(1:nnx+1))

  if(verbose) write(*,*) 'Fill sub-arrays'

  ! Fill sub-arrays (with non actualized old values !!)
  if (ndim == 1) then
       rholoc(nxmin:nxmax,nymin:nymax,nzmin:nzmax       ) =  rhoold(imin-2:imax+2,1:1,1:1       )
         Eloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax       ) =    Eold(imin-2:imax+2,1:1,1:1       )
      rhouloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:ndim) = rhouold(imin-2:imax+2,1:1,1:1,1:ndim)
 else if (ndim ==2) then
      rholoc(nxmin:nxmax,nymin:nymax,nzmin:nzmax       ) =  rhoold(imin-2:imax+2,jmin-2:jmax+2,1:1       )
        Eloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax       ) =    Eold(imin-2:imax+2,jmin-2:jmax+2,1:1       )
     rhouloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:ndim) = rhouold(imin-2:imax+2,jmin-2:jmax+2,1:1,1:ndim)
 else if (ndim ==3) then
      rholoc(nxmin:nxmax,nymin:nymax,nzmin:nzmax       ) =  rhoold(imin-2:imax+2,jmin-2:jmax+2,kmin-2:kmax+2       )
        Eloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax       ) =    Eold(imin-2:imax+2,jmin-2:jmax+2,kmin-2:kmax+2       )
     rhouloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:ndim) = rhouold(imin-2:imax+2,jmin-2:jmax+2,kmin-2:kmax+2,1:ndim)
 endif


  !gravitational potential
#if GRA==1
  if (GRAV) then
     if (ndim==1) then
        philoc(ixmin:ixmax,iymin:iymax,izmin:izmax) = phi(imin-1:imax+1,1:1,1:1)
     else if (ndim==2) then
        philoc(ixmin:ixmax,iymin:iymax,izmin:izmax) = phi(imin-1:imax+1,jmin-1:jmax+1,1:1)
     else if (ndim==3) then
        philoc(ixmin:ixmax,iymin:iymax,izmin:izmax) = phi(imin-1:imax+1,jmin-1:jmax+1,kmin-1:kmax+1)
     endif
  endif
#endif

#if SCALP==1
  allocate(   fxloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:nFx           ))
  allocate(    pFxg(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:nFx ,1:ndim   ))
  allocate(    pFxd(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:nFx ,1:ndim   ))
  allocate(     FxX(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:nFx ,2, 1:ndim))
  allocate(    FxXt(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:nFx ,2, 1:ndim))
  allocate(     FxS(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:nFx ,1:ndim   ))

  if (ndim == 1) then
        Fxloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:nFx ) =   Fxold(imin-2:imax+2,1:1,1:1,1:nFx )
  else if (ndim == 2) then
       Fxloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:nFx  ) =   Fxold(imin-2:imax+2,jmin-2:jmax+2,1:1,1:nFx )
  else if (ndim == 3) then
       Fxloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:nFx  ) =   Fxold(imin-2:imax+2,jmin-2:jmax+2,kmin-2:kmax+2,1:nFx )
  endif 

#endif


                xloc(-1:nnx+3,1) = x(imin-2:imax+3,1)   
 if (ndim.gt.1) xloc(-1:nny+3,2) = x(jmin-2:jmax+3,2)   
 if (ndim.gt.2) xloc(-1:nnz+3,3) = x(kmin-2:kmax+3,3)   


  ! compute the geometry
  if(verbose) write(*,*) 'compute the geometry'
  call cal_geom(nnx,nny,nnz)


  ! compute P and U
  if(verbose) write(*,*) 'compute P and U'
    do idim=1,ndim
       Do k= nzmin,nzmax
       Do j= nymin,nymax
       Do i= nxmin,nxmax
          Uloc(i,j,k,idim) = rhouloc(i,j,k,idim)/rholoc(i,j,k)
       enddo
       enddo
       enddo
    enddo
 
    if(isotherme) then
       Do k= nzmin,nzmax
       Do j= nymin,nymax
       Do i= nxmin,nxmax
          Ploc(i,j,k) = Cs_iso2 * rholoc(i,j,k)
       enddo
       enddo
       enddo
    else
       Einloc=Eloc-0.5*sum(rhouloc**2,4)/rholoc
       call Pressure(rholoc,Einloc,Ploc,nxmin,nxmax,nymin,nymax,nzmin,nzmax)

       Do k= nzmin,nzmax
       Do j= nymin,nymax
       Do i= nxmin,nxmax
          if (Ploc(i,j,k) < 0.) then
             write(*,100) i,j,k,Ploc(i,j,k)
             Ploc(i,j,k) = 1.0E-8*Eloc(i,j,k)
          endif
       enddo
       enddo
       enddo
100 format('Ploc negative (i,j,k), p =',I5,I5,I5,2x,E10.3)
    endif

    
    ! compute slopes (May eventually have two slopes by cells, left and right (superbee)
    if(verbose) write(*,*) 'compute slopes'
    Call slopes( rholoc, prhog, prhod, nxmin,nxmax, nymin,nymax, nzmin,nzmax,slope_type) 
    if(.not.isotherme) Call slopes(   ploc, ppreg, ppred, nxmin,nxmax, nymin,nymax, nzmin,nzmax,slope_type) 
    Call slopes(Einloc, pEing, pEind, nxmin,nxmax, nymin,nymax, nzmin,nzmax,slope_type)

    do jdim = 1,ndim
       Call slopes(uloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,jdim), pvg, pvd,nxmin,nxmax, nymin,nymax, nzmin,nzmax,slope_type) 
       do idim=1,ndim
          pviteg(nxmin:nxmax,nymin:nymax,nzmin:nzmax,jdim,idim) = pvg(nxmin:nxmax,nymin:nymax,nzmin:nzmax,idim)
          pvited(nxmin:nxmax,nymin:nymax,nzmin:nzmax,jdim,idim) = pvd(nxmin:nxmax,nymin:nymax,nzmin:nzmax,idim)
       enddo
    enddo

#if SCALP==1
    pfxg = 0.
    pfxd = 0.
    
    do ifx = 1,Nfx
       Call slopes(Fxloc(nxmin:nxmax,nymin:nymax,nzmin:nzmax,ifx), pvg,pvd, nxmin,nxmax, nymin,nymax, nzmin,nzmax,slope_type) 
       do idim=1,ndim
          pFxd(nxmin:nxmax,nymin:nymax,nzmin:nzmax,ifx,idim) = pvd(nxmin:nxmax,nymin:nymax,nzmin:nzmax,idim)
          pFxg(nxmin:nxmax,nymin:nymax,nzmin:nzmax,ifx,idim) = pvg(nxmin:nxmax,nymin:nymax,nzmin:nzmax,idim)
       enddo
    enddo
#endif



    ! Compute interface values
    if (verbose) write(*,*) '-- Compute interface values --'

    do idim  = 1,ndim

       do k = nzmin,nzmax
       do j = nymin,nymax
       do i = nxmin,nxmax


          indice(1) = i
          indice(2) = j
          indice(3) = k
          
          id  = indice(idim)

          drp = ( xloc(id+1,idim) - xcloc(id,idim)) 
          drm = (xcloc(id  ,idim) -  xloc(id,idim)) 
       
          rhoX(i,j,k,1,idim) =  rholoc(i,j,k) - drm*prhog (i,j,k,idim)  
          rhoX(i,j,k,2,idim) =  rholoc(i,j,k) + drp*prhod (i,j,k,idim)  
          
          ! The Ein array is useless, it can be replaced by a local variable (2,ndim)
          EinX(i,j,k,1,idim) =  Einloc(i,j,k) - drm*pEing (i,j,k,idim)  
          EinX(i,j,k,2,idim) =  Einloc(i,j,k) + drp*pEind (i,j,k,idim)  

          do jdim=1,ndim
             uX(i,j,k,jdim,1,idim) =    uloc(i,j,k,jdim) - drm*pviteg(i,j,k,jdim,idim)  
             uX(i,j,k,jdim,2,idim) =    uloc(i,j,k,jdim) + drp*pvited(i,j,k,jdim,idim)  
          enddo
          
          do iface=1,2
             rhouX(i,j,k,1:ndim,iface,idim) = rhoX(i,j,k,iface,idim)*uX(i,j,k,1:ndim,iface,idim)
          enddo
           
#if SCALP==1
          do ifx = 1,nFx
             FxX(i,j,k,ifx,1,idim) =  Fxloc(i,j,k,ifx) - drm*pFxg(i,j,k,ifx,idim)  
             FxX(i,j,k,ifx,2,idim) =  Fxloc(i,j,k,ifx) + drp*pFxd(i,j,k,ifx,idim)  
          enddo
#endif

       enddo
       enddo
       enddo
    enddo

    if(isotherme) then
       do idim  = 1,ndim
          do k = nzmin,nzmax
          do j = nymin,nymax
          do i = nxmin,nxmax
             pX(i,j,k,1,idim) = Cs_iso2 * rhoX(i,j,k,1,idim)
             pX(i,j,k,2,idim) = Cs_iso2 * rhoX(i,j,k,2,idim)
          enddo
          enddo
          enddo
       enddo
    else
       do idim  = 1,ndim
          do k = nzmin,nzmax
          do j = nymin,nymax
          do i = nxmin,nxmax
             indice(1) = i
             indice(2) = j
             indice(3) = k
             id  = indice(idim)
             drp = ( xloc(id+1,idim) - xcloc(id,idim)) 
             drm = (xcloc(id  ,idim) -  xloc(id,idim)) 

             pX(i,j,k,1,idim) =    ploc(i,j,k) - drm*ppreg (i,j,k,idim)  
             pX(i,j,k,2,idim) =    ploc(i,j,k) + drp*ppred (i,j,k,idim)  
             do iface=1,2
                Utot2 = 0.
                do jdim=1,ndim
                   Utot2 = Utot2 + uX(i,j,k,jdim,iface,idim)**2
                enddo
                EX   (i,j,k       ,iface,idim) = EinX(i,j,k,iface,idim) + 0.5*RhoX(i,j,k,iface,idim)*Utot2
              enddo
          enddo
          enddo
          enddo
       enddo
    endif

  if (verbose) write(*,*) '-- end compute interface --'

  !  flux(i,j,1:ndim,1:2*ndim)  (flux imposes...)
  allocate(flux_masse(ixmin:ixmax,iymin:iymax,izmin:izmax       ,2,ndim))
  allocate(flux_ener (ixmin:ixmax,iymin:iymax,izmin:izmax       ,2,ndim))
  allocate(flux_pre  (ixmin:ixmax,iymin:iymax,izmin:izmax       ,2,ndim))
  allocate(flux_mom  (ixmin:ixmax,iymin:iymax,izmin:izmax,1:ndim,2,ndim))
#if SCALP==1
  allocate(flux_fx   (ixmin:ixmax,iymin:iymax,izmin:izmax,1:nfx ,2,ndim))
#endif

  if (verbose) write(*,*) '-- compute fluxes --'

  ! compute fluxes
  do idim = 1,ndim
     do k = izmin,izmax
     do j = iymin,iymax
     do i = ixmin,ixmax


        ip1 = i + kron(1,idim) ; jp1 = j + kron(2,idim) ; kp1 = k + kron(3,idim)


        flux_masse(i,j,k,1,idim) = RhoX(i,j,k,1,idim)*uX(i,j,k,idim,1,idim) *ds(i  ,j  ,k  ,idim)
        flux_masse(i,j,k,2,idim) = RhoX(i,j,k,2,idim)*uX(i,j,k,idim,2,idim) *ds(ip1,jp1,kp1,idim)

        do jdim=1,ndim
           flux_mom(i,j,k,jdim,1,idim) =(RhoX(i,j,k,1,idim)*uX(i,j,k,jdim,1,idim)*uX(i,j,k,idim,1,idim))*ds(i  ,j  ,k  ,idim)
           flux_mom(i,j,k,jdim,2,idim) =(RhoX(i,j,k,2,idim)*uX(i,j,k,jdim,2,idim)*uX(i,j,k,idim,2,idim))*ds(ip1,jp1,kp1,idim)
        enddo

        flux_pre(i,j,k,1,idim) = pX(i,j,k,1,idim)*ds(i  ,j  ,k  ,idim)
        flux_pre(i,j,k,2,idim) = pX(i,j,k,2,idim)*ds(ip1,jp1,kp1,idim)

#if SCALP==1
     do ifx=1,nfx
        flux_fx(i,j,k,ifx,1,idim) = fxX(i,j,k,ifx,1,idim) * uX(i,j,k,idim,1,idim) * ds(i  ,j  ,k  ,idim)
        flux_fx(i,j,k,ifx,2,idim) = fxX(i,j,k,ifx,2,idim) * uX(i,j,k,idim,2,idim) * ds(ip1,jp1,kp1,idim)
     enddo
#endif

     enddo
     enddo
     enddo
  enddo

  if(isotherme) then
     do idim = 1,ndim
        do k = izmin,izmax
        do j = iymin,iymax
        do i = ixmin,ixmax
           flux_ener(i,j,k,1,idim) = 0.
           flux_ener(i,j,k,2,idim) = 0.
        enddo
        enddo
        enddo
     enddo
  else
     do idim = 1,ndim
        do k = izmin,izmax
        do j = iymin,iymax
        do i = ixmin,ixmax
           ip1 = i + kron(1,idim) ; jp1 = j + kron(2,idim) ; kp1 = k + kron(3,idim)
           flux_ener(i,j,k,1,idim) =(EX(i,j,k,1,idim)+pX(i,j,k,1,idim))*uX(i,j,k,idim,1,idim) *ds(i  ,j  ,k  ,idim)
           flux_ener(i,j,k,2,idim) =(EX(i,j,k,2,idim)+pX(i,j,k,2,idim))*uX(i,j,k,idim,2,idim) *ds(ip1,jp1,kp1,idim)
        enddo
        enddo
        enddo
     enddo
  endif

  ! ecrasement eventuel si flux imposes
  ! A faire.......


  if (verbose) write(*,*) '-- Evolution over dt/2 --'

  ! for incoming or outgoing fluxes
  signe = 1. ; signe(2) = -1. ; signe(4) = -1. ; signe(6) = -1.

  ! dt/2
  do k = izmin,izmax
  do j = iymin,iymax
  do i = ixmin,ixmax

     flux_masse_tot       = 0.
     flux_mom_tot(1:ndim) = 0.
     flux_pre_tot(1:ndim) = 0.
     flux_ener_tot        = 0.
#if SCALP==1
     flux_fx_tot (1:nfx)  = 0.
#endif
     
     Rayon = 1.
     if (geom_dir(2) == 'phi_c' .or. geom_dir(2) == 'the_s') then
        rayon(2,1:2,1:3) = xcloc(i  ,1)
        rayon(2,1  ,1  ) =  xloc(i  ,1)
        rayon(2,2  ,1  ) =  xloc(i+1,1)
     endif

     if (geom_dir(3) == 'phi_s' ) then
        rayon(3,1  ,1  ) =  xloc(i  ,1) * sin(xcloc(j  ,2))
        rayon(3,2  ,1  ) =  xloc(i+1,1) * sin(xcloc(j  ,2))
        rayon(3,1  ,2  ) = xcloc(i  ,1) * sin( xloc(j  ,2))
        rayon(3,2  ,2  ) = xcloc(i  ,1) * sin( xloc(j+1,2))
        rayon(3,1  ,3  ) = xcloc(i  ,1) * sin(xcloc(j  ,2))
        rayon(3,2  ,3  ) = xcloc(i  ,1) * sin(xcloc(j  ,2))
     endif

     do idim  = 1,ndim
     do iface = 1,2
        
        flux_masse_tot  = flux_masse_tot  + signe(iface)* 0.5*dt*flux_masse(i,j,k,  iface,idim)
        flux_ener_tot   = flux_ener_tot   + signe(iface)* 0.5*dt*flux_ener (i,j,k,  iface,idim)
#if SCALP==1
        flux_fx_tot (1:nfx)  = flux_fx_tot (1:nfx)  + signe(iface)* 0.5*dt*flux_fx   (i,j,k,1:nfx ,iface,idim)
#endif
        do jdim=1,ndim
           flux_mom_tot(jdim) = flux_mom_tot(jdim) + signe(iface)* 0.5*dt*flux_mom  (i,j,k,jdim,iface,idim) * Rayon(jdim,iface,idim)
        enddo
        
        flux_pre_tot(idim) = flux_pre_tot(idim) +  signe(iface)* 0.5*dt*flux_pre(i,j,k,iface,idim)

     enddo
     enddo


     ! Geometry terms
     if (geom_dir(1) == 'rs' .or. geom_dir(1) == 'rc') then
        Pression = Ploc(i,j,k) * (ds(i+1,j,k,1) - ds(i,j,k,1))
        f1 = flux_pre_tot(1) + Pression  *0.5*dt

        f2 = - 0.5*dt* dv(i,j,k)* (pX(i,j,k,2,1) - pX(i,j,k,1,1))/(xloc(i+1,1)-xloc(i,1))

        flux_pre_tot(1) = f2
     endif

     if (geom_dir(2) == 'the_s') &
         flux_pre_tot(2) = - 0.5*dt* dv(i,j,k)* (pX(i,j,k,2,2) - pX(i,j,k,1,2))/(xloc(j+1,2)-xloc(j,2)) / xcloc(i,1)

     Pression = 0.
     if (geom_dir(2) == 'phi_c') Pression = Pression + rholoc(i,j,k)*Uloc(i,j,k,2)*Uloc(i,j,k,2) *     (ds(i+1,j,k,1) - ds(i,j,k,1))
     if (geom_dir(2) == 'the_s') Pression = Pression + rholoc(i,j,k)*Uloc(i,j,k,2)*Uloc(i,j,k,2) * 0.5*(ds(i+1,j,k,1) - ds(i,j,k,1))
     if (geom_dir(3) == 'phi_s') Pression = Pression + rholoc(i,j,k)*Uloc(i,j,k,3)*Uloc(i,j,k,3) * 0.5*(ds(i+1,j,k,1) - ds(i,j,k,1))

     flux_pre_tot(1) = flux_pre_tot(1) + Pression  *0.5*dt
     

     if (geom_dir(3) == 'phi_s')  then
        Pression = 0. 
        cotanx = cos(xcloc(j,2))/sin(xcloc(j,2))                    
        Pression = Pression + rholoc(i,j,k)*Uloc(i,j,k,3)*Uloc(i,j,k,3) * 0.5*(ds(i+1,j,k,1) - ds(i,j,k,1)) * cotanx
        flux_pre_tot(2) = flux_pre_tot(2) + Pression  *0.5*dt
     endif

     !---------------

     do idim  = 1,ndim
     do iface = 1,2
        rhoXt (i,j,k,  iface,idim) =  rhoX (i,j,k,  iface,idim) + flux_masse_tot      /dv(i,j,k)
        EXt   (i,j,k,  iface,idim) =  EX   (i,j,k,  iface,idim) + flux_ener_tot       /dv(i,j,k)     

        do jdim = 1,ndim
           rhouXt(i,j,k,jdim,iface,idim) =  rhouX(i,j,k,jdim,iface,idim) + flux_mom_tot(jdim)/dv(i,j,k) /Rayon(jdim,2,3)
        enddo

        do jdim = 1,ndim
           rhouXt(i,j,k,jdim,iface,idim) =  rhouXt(i,j,k,jdim,iface,idim) + flux_pre_tot(jdim)/dv(i,j,k) 
        enddo

        uXt   (i,j,k,1:ndim,iface,idim) = rhouXt(i,j,k,1:ndim,iface,idim)/rhoXt (i,j,k,iface,idim)

#if SCALP==1
        do ifx = 1,nFx
           FxXt(i,j,k,ifx,iface,idim) =  FxX(i,j,k,ifx,iface,idim) + flux_fx_tot(ifx)/dv(i,j,k)
        enddo
#endif

     enddo
     enddo
    
  enddo
  enddo
  enddo

  if(isotherme) then
     do idim  = 1,ndim
     do iface = 1,2
        do k = izmin,izmax
        do j = iymin,iymax
        do i = ixmin,ixmax
           pXt(i,j,k,iface,idim) = Cs_iso2 * rhoXt(i,j,k,iface,idim)
        enddo
        enddo
        enddo
     enddo
     enddo
  else

     EinXt=EXt-0.5*sum(rhouXt**2,4)/rhoXt

     do idim  = 1,ndim
     do iface = 1,2
        call Pressure(rhoXt(:,:,:,iface,idim),EinXt(:,:,:,iface,idim),PXt(:,:,:,iface,idim),ixmin,ixmax,iymin,iymax,izmin,izmax)
     enddo
     enddo
  endif

  !Gravity predictor step
#if GRA==1
  if (GRAV) then
     do iface=1,2
        do idim=1,ndim
           call grav_predictor(uXt(:,:,:,1:ndim,iface,idim),philoc,dt,ixmin,ixmax,iymin,iymax,izmin,izmax)
        end do
     end do
     do idim=1,ndim
        rhouXt(:,:,:,idim,:,:)=rhoXt*uXt(:,:,:,idim,:,:)
     end do
     if(.not.isotherme) EXt=EinXt+0.5*sum(rhouXt**2,4)/rhoXt
  endif
#endif

  ! Riemann and compute fluxes
  if (verbose) write(*,*) '-- Riemann problems --'
  do idim = 1,ndim

     do k=1,nnz+kron(idim,3)
     do j=1,nny+kron(idim,2)

        do i=1,nnx+kron(idim,1)

           im1 = i - kron(1,idim) ; jm1 = j - kron(2,idim) ; km1 = k - kron(3,idim)
           
           dll(i) = rhoXt(im1,jm1,km1,2,idim) ; ull(i,1:ndim) = uXt(im1,jm1,km1,1:ndim,2,idim) ; pll(i) = pXt(im1,jm1,km1,2,idim)
           drr(i) = rhoXt(i  ,j  ,k  ,1,idim) ; urr(i,1:ndim) = uXt(i  ,j  ,k  ,1:ndim,1,idim) ; prr(i) = pXt(i  ,j  ,k  ,1,idim)
           
#if SCALP==1
           do ifx = 1,nfx
              fll(i,ifx) = fxXt(im1,jm1,km1,ifx,2,idim) ; frr(i,ifx) = fxXt(i,j,k,ifx,1,idim) 
           enddo
#else
           fll = 1. ; frr = 1.
#endif           

        enddo
        
        V = 0. ; nbuf = nnx+1

        if(isotherme) then
           Call solver_iso(dll,ull,fll,drr,urr,frr,dss,uss,fss, V,idim,nbuf)
           if (v == -1.) then
              call close_program(mype,'called from EVOL_HYDRO: problem density < 0')
           endif

           do i=1,nnx+1
              rhoS(i,j,k,idim) = dss(i) ; uS(i,j,k,1:ndim,idim) = uss(i,1:ndim) ; pS(i,j,k,idim) = Cs_iso2 * dss(i)
#if SCALP==1
              do ifx = 1,nfx
                 fxS(i,j,k,ifx,idim) = fss(i,ifx)
              enddo
#endif
           enddo
        else
           ! ess is the internal energy, not the total energy.
           ! This variable is introduced since it is necessary for the relaxation solver.
           ! The values of the energy and pressure in the star region are not necessarily compatible with the EOS.

           SELECT CASE(Solver) ! hydrodynamic solver selection

           CASE('exact') 
              Call solver_exact   (dll,ull,pll,fll,drr,urr,prr,frr,dss,uss,pss,fss, V,idim,nbuf)
              ess = pss/(gamma-1.)

           CASE('acoustic')           
              Call solver_acoustic(dll,ull,pll,fll,drr,urr,prr,frr,dss,uss,pss,fss, V,idim,nbuf)
              ess = pss/(gamma-1.)

           CASE('relax')
              call solver_relax(dll=dll, ull=ull, pll=pll, fll=fll, &
                   drr=drr, urr=urr, prr=prr, frr=frr, &
                   dss=dss, uss=uss, pss=pss, ess=ess, fss=fss, &
                   vit=V  ,idim=idim,N=Nbuf)

           CASE('eos')           
              do i=1,nnx+kron(idim,1)
                 im1 = i - kron(1,idim) ; jm1 = j - kron(2,idim) ; km1 = k - kron(3,idim)
                 ell(i) = EinXt(im1,jm1,km1,2,idim) 
                 err(i) = EinXt(i  ,j  ,k  ,1,idim)
              enddo
              call Sound_speed(dll,ell,cll,1,nnx+kron(idim,1),1,1,1,1)
              call Sound_speed(drr,err,crr,1,nnx+kron(idim,1),1,1,1,1)

              call solver_eos(dll,ull,pll,cll,ell,fll,drr,urr,prr,crr,err,frr,dss,uss,pss,ess,fss,V,idim,nbuf) 

           CASE DEFAULT
	   
              call close_program(mype,'Unknown solver: '//Solver)
           
           END SELECT

           do i=1,nnx+1
              rhoS(i,j,k,idim) = dss(i) ; uS(i,j,k,1:ndim,idim) = uss(i,1:ndim) ; pS(i,j,k,idim) = pss(i)
             
                es(i,j,k,idim) = ess(i) + 0.5*dss(i)*SUM( uss(i,:)*uss(i,:) ) 

#if SCALP==1
              do ifx = 1,nfx
                 fxS(i,j,k,ifx,idim) = fss(i,ifx)
              enddo
#endif

           enddo

        endif
        
     enddo
  enddo
     
enddo

  if (verbose) write(*,*) '-- end of Riemann problems --'

  ! compute pressure and velocity cell-centered for geometric source terms
  if (.not.Cartesien) then
     PX   = 0.
     UX   = 0.
     RhoX = 0.

     do idim = 1,ndim

        do k=1,nnz
        do j=1,nny
        do i=1,nnx

           ip1 = i + kron(1,idim) ; jp1 = j + kron(2,idim) ; kp1 = k + kron(3,idim)

           PX  (i,j,k,1,idim) = pS  (i  ,j  ,k  ,idim) + pS  (ip1,jp1,kp1,idim)
           rhoX(i,j,k,1,idim) = rhoS(i  ,j  ,k  ,idim) + rhoS(ip1,jp1,kp1,idim)

           do jdim=1,ndim
              UX(i,j,k,jdim,1,idim) =  uS(i  ,j  ,k  ,jdim,idim)+ uS(ip1,jp1,kp1,jdim,idim)
           enddo

        enddo
        enddo
        enddo
     enddo

     PX   = 0.5*PX  
     UX   = 0.5*UX  
     rhoX = 0.5*RhoX
  endif

  ! ecrasement eventuel de certain flux avec les CL
  ! A FAIRE

  ! Variables update
  if (verbose) write(*,*) ' variables update'


  ! Fluxes
  do idim = 1,ndim

     do k=1,nnz
     do j=1,nny
     do i=1,nnx

        ip1 = i + kron(1,idim) ; jp1 = j + kron(2,idim) ; kp1 = k + kron(3,idim)

        flux_masse(i,j,k,1,idim) = RhoS(i  ,j  ,k  ,idim)*uS(i  ,j  ,k  ,idim,idim) *ds(i  ,j  ,k  ,idim)
        flux_masse(i,j,k,2,idim) = RhoS(ip1,jp1,kp1,idim)*uS(ip1,jp1,kp1,idim,idim) *ds(ip1,jp1,kp1,idim)

        do jdim=1,ndim
           flux_mom(i,j,k,jdim,1,idim) =(RhoS(i  ,j  ,k  ,idim)*uS(i  ,j  ,k  ,jdim,idim)*uS(i  ,j  ,k  ,idim,idim))*ds(i  ,j  ,k  ,idim) 
           flux_mom(i,j,k,jdim,2,idim) =(RhoS(ip1,jp1,kp1,idim)*uS(ip1,jp1,kp1,jdim,idim)*uS(ip1,jp1,kp1,idim,idim))*ds(ip1,jp1,kp1,idim)
        enddo

        flux_pre(i,j,k,1,idim) = pS(i  ,j  ,k  ,idim)*ds(i  ,j  ,k  ,idim) 
        flux_pre(i,j,k,2,idim) = pS(ip1,jp1,kp1,idim)*ds(ip1,jp1,kp1,idim)

#if SCALP==1
        do ifx=1,nfx
           flux_fx(i,j,k,ifx,1,idim) = fxS(i  ,j  ,k  ,ifx,idim) * uS(i  ,j  ,k  ,idim,idim) * ds(i  ,j  ,k  ,idim)  * RhoS(i  ,j  ,k  ,idim)
           flux_fx(i,j,k,ifx,2,idim) = fxS(ip1,jp1,kp1,ifx,idim) * uS(ip1,jp1,kp1,idim,idim) * ds(ip1,jp1,kp1,idim)  * RhoS(ip1,jp1,kp1,idim)
        enddo
#endif

     enddo
     enddo
     enddo
  enddo

  if(isotherme) then
     do idim = 1,ndim
        do k=1,nnz
        do j=1,nny
        do i=1,nnx
           flux_ener(i,j,k,1,idim) = 0.
           flux_ener(i,j,k,2,idim) = 0.
        enddo
        enddo
        enddo
     enddo
  else
     do idim = 1,ndim
        do k=1,nnz
        do j=1,nny
        do i=1,nnx
           ip1 = i + kron(1,idim) ; jp1 = j + kron(2,idim) ; kp1 = k + kron(3,idim)
           flux_ener(i,j,k,1,idim) =(eS(i  ,j  ,k  ,idim)+pS(i  ,j  ,k  ,idim))*uS(i  ,j  ,k  ,idim,idim) *ds(i  ,j  ,k  ,idim)
           flux_ener(i,j,k,2,idim) =(eS(ip1,jp1,kp1,idim)+pS(ip1,jp1,kp1,idim))*uS(ip1,jp1,kp1,idim,idim) *ds(ip1,jp1,kp1,idim)
        enddo
        enddo
        enddo
     enddo
  endif

#if RAYO==1 || DIF==1
  allocate(moment_ray(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:ndim))

  ! Computes the radiative source term rho*(kappa_abs+kappa_sca)F/c in moment and energy hydro equations
  if(RAY.or.DIFFU) call exp_source_ray(rholoc,Einloc,moment_ray,imin,imax,jmin,jmax,kmin,kmax,nxmin,nxmax,nymin,nymax,nzmin,nzmax)

  ! Explicit treatment of comoving terms for radiation
  if(RAY  ) call exp_comobile_ray(rholoc,Einloc,uS,imin,imax,jmin,jmax,kmin,kmax,nxmin,nxmax,nymin,nymax,nzmin,nzmax,'ray')
  if(DIFFU) call exp_comobile_ray(rholoc,Einloc,uS,imin,imax,jmin,jmax,kmin,kmax,nxmin,nxmax,nymin,nymax,nzmin,nzmax,'dif')
#endif

  do k=1,nnz
  do j=1,nny
  do i=1,nnx

     flux_masse_tot       = 0.
     flux_mom_tot(1:ndim) = 0.
     flux_pre_tot(1:ndim) = 0.
     flux_ener_tot        = 0.
#if SCALP==1
     flux_fx_tot (1:nfx)  = 0.
#endif
     Pression             = 0.

     Rayon = 1.
     if (geom_dir(2) == 'phi_c' .or. geom_dir(2) == 'the_s') then
        rayon(2,1:2,1:3) = xcloc(i  ,1)
        rayon(2,1  ,1  ) =  xloc(i  ,1)
        rayon(2,2  ,1  ) =  xloc(i+1,1)
     endif

     if (geom_dir(3) == 'phi_s' ) then
        rayon(3,1  ,1  ) =  xloc(i  ,1) * sin(xcloc(j  ,2))
        rayon(3,2  ,1  ) =  xloc(i+1,1) * sin(xcloc(j  ,2))
        rayon(3,1  ,2  ) = xcloc(i  ,1) * sin( xloc(j  ,2))
        rayon(3,2  ,2  ) = xcloc(i  ,1) * sin( xloc(j+1,2))
        rayon(3,1  ,3  ) = xcloc(i  ,1) * sin(xcloc(j  ,2))
        rayon(3,2  ,3  ) = xcloc(i  ,1) * sin(xcloc(j  ,2))
     endif

     do idim  = 1,ndim
     do iface = 1,2
        flux_masse_tot     = flux_masse_tot     + signe(iface)* dt*flux_masse(i,j,k,iface,idim)
        flux_ener_tot      = flux_ener_tot      + signe(iface)* dt*flux_ener (i,j,k,iface,idim)
        flux_pre_tot(idim) = flux_pre_tot(idim) + signe(iface)* dt*flux_pre  (i,j,k,iface,idim)

#if SCALP==1
        flux_fx_tot (1:nfx)  = flux_fx_tot (1:nfx)  + signe(iface)* dt*flux_fx(i,j,k,1:nfx ,iface,idim)
#endif

        do jdim = 1,ndim
           flux_mom_tot(jdim) = flux_mom_tot(jdim) + signe(iface)* dt*flux_mom(i,j,k,jdim,iface,idim) * Rayon(jdim,iface,idim)
        enddo

     enddo
     enddo

     ! Geometry terms
     if (geom_dir(1) == 'rs' .or.  geom_dir(1) == 'rc') then
        f2  = - dt* dv(i,j,k)* (pS(i+1,j  ,k,1) - pS(i,j,k,1))/(xloc(i+1,1)-xloc(i,1))

        Pression = PX(i,j,k,1,1) * (ds(i+1,j,k,1) - ds(i,j,k,1))
        f1 = flux_pre_tot(1) + Pression *dt

        flux_pre_tot(1) =  f2
     endif
     
     if (geom_dir(2) == 'the_s')                        &
                     flux_pre_tot(2) = - dt* dv(i,j,k)* (pS(i  ,j+1,k,2) - pS(i,j,k,2))/(xloc(j+1,2)-xloc(j,2)) / xcloc(i,1)


     Pression = 0.
     if (geom_dir(2) == 'phi_c') Pression = Pression + rhoX(i,j,k,1,1)*UX(i,j,k,2,1,1)*UX(i,j,k,2,1,1) *     (ds(i+1,j,k,1) - ds(i,j,k,1)) 
     if (geom_dir(2) == 'the_s') Pression = Pression + rhoX(i,j,k,1,1)*UX(i,j,k,2,1,1)*UX(i,j,k,2,1,1) * 0.5*(ds(i+1,j,k,1) - ds(i,j,k,1))
     if (geom_dir(3) == 'phi_s') Pression = Pression + rhoX(i,j,k,1,1)*UX(i,j,k,3,1,1)*UX(i,j,k,3,1,1) * 0.5*(ds(i+1,j,k,1) - ds(i,j,k,1))

     flux_pre_tot(1) = flux_pre_tot(1) + Pression * dt


     if (geom_dir(3) == 'phi_s')  then
        Pression = 0. 
        cotanx = cos(xcloc(j,2))/sin(xcloc(j,2))                    
        Pression = Pression + rhoX(i,j,k,1,2)*UX(i,j,k,3,1,2)*UX(i,j,k,3,1,2) * 0.5*(ds(i+1,j,k,1) - ds(i,j,k,1)) * cotanx
        flux_pre_tot(2) = flux_pre_tot(2) + Pression * dt
     endif

#if RAYO==1 || DIF==1
        if(RAY.or.DIFFU) then
           do jdim=1,ndim
              flux_mom_tot(jdim) = flux_mom_tot(jdim) + dt*dv(i,j,k)*moment_ray(i,j,k,jdim)
              flux_ener_tot      = flux_ener_tot      + dt*dv(i,j,k)*moment_ray(i,j,k,jdim)*rhouloc(i,j,k,jdim)/rholoc(i,j,k)
           enddo
        endif
#endif

      rr =   rholoc(i,j,k)
      rholoc(i,j,k  ) =   rholoc(i,j,k  ) + flux_masse_tot  /dv(i,j,k)
        Eloc(i,j,k  ) =     Eloc(i,j,k  ) + flux_ener_tot   /dv(i,j,k)
                                        
        do jdim = 1,ndim
           rhouloc(i,j,k,jdim) =  rhouloc(i,j,k,jdim) + flux_mom_tot(jdim) /dv(i,j,k) /Rayon(jdim,2,3)
        enddo
        do jdim = 1,ndim
           rhouloc(i,j,k,jdim) =  rhouloc(i,j,k,jdim) + flux_pre_tot(jdim) /dv(i,j,k)
        enddo

#if SCALP==1
       fxloc(i,j,k,1:nfx ) =(rr*fxloc(i,j,k,1:nfx ) + flux_fx_tot (1:nfx ) /dv(i,j,k))/rholoc(i,j,k)      
#endif

  enddo
  enddo
  enddo

  if(isotherme) then
     do k=1,nnz
     do j=1,nny
     do i=1,nnx
       ! When isotherm, the pressure is stored in the total energy variable
       Eloc(i,j,k) = Cs_iso2 * rholoc(i,j,k)
     enddo
     enddo
     enddo
  endif

  !read(*,*)

  if(verbose) write(*,*) 'end of update'


  ! Fill arrays with actualized sub-arrays
  rho(imin:imax,jmin:jmax,kmin:kmax       ) =  rholoc(1:nnx,1:nny,1:nnz       )
    E(imin:imax,jmin:jmax,kmin:kmax       ) =    Eloc(1:nnx,1:nny,1:nnz       )
 rhou(imin:imax,jmin:jmax,kmin:kmax,1:ndim) = rhouloc(1:nnx,1:nny,1:nnz,1:ndim)

#if SCALP==1
   fx(imin:imax,jmin:jmax,kmin:kmax,1:nfx ) =   fxloc(1:nnx,1:nny,1:nnz,1:nfx )
#endif

#if RAYO==1 || DIF==1
   deallocate(moment_ray)
#endif


 ! Deallocate arrays
 ! geom
  deallocate(dx,dxc)
  deallocate(ds,dv)
 ! others
  deallocate(xloc,xcloc)
  deallocate( rholoc,ploc, Eloc)
  deallocate(rhouloc,uloc)
#if GRA==1
  if (GRAV) deallocate(philoc)
#endif

  deallocate(Einloc,pEing,pEind,EinX,EinXt,ell,cll,err,crr)

  deallocate(  prhog,  ppreg)
  deallocate(pviteg,pvg)
  deallocate(  prhod,  ppred)
  deallocate(pvited,pvd)

  deallocate(rhoX )
  deallocate(rhoXt)
  deallocate(  EX )
  deallocate(  EXt)
  deallocate(  PX )
  deallocate(  PXt)

  deallocate(   uX )
  deallocate(   uXt)
  deallocate(rhouX )
  deallocate(rhouXt)


  deallocate(rhoS,PS,ES,uS)

  deallocate(flux_masse,flux_ener,flux_pre,flux_mom)

#if SCALP==1
  deallocate(Fxloc,pFxd, pfxg )
  deallocate(  FxX )
  deallocate(  FxXt)
  deallocate(  FxS, flux_fx)
#endif


  deallocate(dll , pll , drr       )  
  deallocate(prr , dss , pss , ess )
  deallocate(ull , urr , uss       )
  deallocate(fll , frr , fss       )

  if (verbose) write(*,*) 'end of evol_hydro'
  
  return

end subroutine evol_hydro
