!> \file
!! Contains module histo and subroutines init_histo()
!! and cal_histogramme()
!!
!! This file contains the routines to compute histogram
!! statistics on hydro variables.
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module HISTO
!
!> Contains variables for histogram computations.
!<
module histo
  integer                         :: Npoint      !< Npoint
  real, dimension(:), allocatable :: hP          !< hp
  real, dimension(:), allocatable :: hN          !< hn
  real, dimension(:), allocatable :: hT          !< ht
  real, dimension(:), allocatable :: P           !< p
  real, dimension(:), allocatable :: N           !< n
  real, dimension(:), allocatable :: T           !< t
  real, dimension(:), allocatable :: hPg         !< hPg
  real, dimension(:), allocatable :: hNg         !< hNg
  real, dimension(:), allocatable :: hTg         !< hTg
  real                            :: Pmin        !< Pmin
  real                            :: Pmax        !< Pmax
  real                            :: r_P         !< r_P
  real                            :: lr_P        !< lr_P
  real                            :: Nmin        !< Nmin
  real                            :: Nmax        !< Nmax
  real                            :: r_N         !< r_N
  real                            :: lr_N        !< lr_N
  real                            :: Tmin        !< Tmin
  real                            :: Tmax        !< Tmax
  real                            :: r_T         !< r_T
  real                            :: lr_T        !< lr_T
end module histo

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine INIT_HISTO
!
!> Initialises histogram parameters.
!<
subroutine init_histo

  use histo
  use unites, only : erg,gramme,cm3,seconde,kms,kelvin
  use para  , only : mype

  implicit none

  integer :: i
  logical :: existe

  Npoint = 25
  allocate(N(0:Npoint),P(0:Npoint),T(0:Npoint), hP (0:Npoint),hN (0:Npoint),hT (0:Npoint) )
  if (mype == 0) allocate(                      hPg(0:Npoint),hNg(0:Npoint),hTg(0:Npoint) )

  ! Pressure
  Pmin = 1.0E-14 !cgs *erg/cm3
  Pmax = 1.0E-10
  
  r_P = (Pmax/Pmin)**(1./float(Npoint)) ; lr_P = Log(r_P)
  P(0) = Pmin
  do i=1,Npoint
     P(i) = P(i-1)*r_P
  enddo

  ! Temperature
  Tmin = 1.0E1
  Tmax = 1.0E5
  r_T = (Tmax/Tmin)**(1./float(Npoint)) ; lr_T = Log(r_T)
  T(0) = Tmin
  do i=1,Npoint
     T(i) = T(i-1)*r_T
  enddo


  ! Density
  Nmin = 1.0E-1
  Nmax = 1.0E4
  r_N = (Nmax/Nmin)**(1./float(Npoint)) ; ; lr_N = Log(r_N)
  N(0) = Nmin
  do i=1,Npoint
     N(i) = N(i-1)*r_N
  enddo

  if (mype == 0) then
     Inquire(file='histo.dat',EXIST=existe)
     if (existe) then
        open(44,file='histo.dat',form='formatted',status='old',position='append',action='readwrite')
     else
        open(44,file='histo.dat',form='formatted',status='new')
        write(44,888) (N(i),i=0,Npoint)
        write(44,888) (P(i),i=0,Npoint)
        write(44,888) (T(i),i=0,Npoint)
     endif
  endif

888  format(100(E10.3,2x))

end subroutine init_histo

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine CAL_HISTOGRAMME
!
!> Computes histogram statistics.
!<
subroutine cal_histogramme

  use parameters, only : nx,ndim
  use var       , only : rho,E,rhou
  use unites    , only : erg,gramme,cm3,seconde,kms,kelvin,an
  use gammas    , only : gamma
  use divers    , only : mu,temps
  use histo
  use para      , only : mype
  
  implicit none

  integer :: i,j,k,idim
  integer :: iNN,ipp,itt

  real :: Masse,Masse_glob
  real :: uma,kb
  real :: U2
  real :: rr,pp,tt,nn

#if PARA==1
  include 'mpif.h'
  integer :: zero,Ncount,ierr
#endif


! This routine uses cgs units

  kb  =  1.38062e-16   ! erg/degre
  uma =  1.660531e-24  ! gramme
  
  Masse = 0. ; hN = 0. ; hT = 0. ; hP = 0.
  
  do k=1,nx(3)
  do j=1,nx(2)
  do i=1,nx(1)
     
     rr = rho(i,j,k) / (gramme/cm3)
     NN = rr         / (mu*uma)

     U2 = 0. 
     do idim=1,ndim
        U2 = U2 + (rhou(i,j,k,idim)/ rho(i,j,k))**2
     enddo
        
     PP = (gamma-1.)*(E(i,j,k) - 0.5*rho(i,j,k)*U2) /(erg/cm3)

     TT = (PP/rr) *(mu*uma/kb)
          
     
     iNN = Log(NN/Nmin)/lr_N
     iNN = max(iNN,0) ; iNN = min(iNN,Npoint)

     iPP = Log(PP/Pmin)/lr_P
     iPP = max(iPP,0) ; iPP = min(iPP,Npoint)

     iTT = Log(TT/Tmin)/lr_T
     iTT = max(iTT,0) ; iTT = min(iTT,Npoint)
     

     hN(iNN) = hN(iNN) + rr
     hP(iPP) = hP(iPP) + rr
     hT(iTT) = hT(iTT) + rr

     Masse = Masse + rr

  enddo
  enddo
  enddo
  

#if PARA==1
  Zero = 0 ; Ncount = Npoint+1
  call MPI_reduce(Masse, Masse_glob, 1     , MPI_DOUBLE_PRECISION, MPI_SUM,Zero,MPI_COMM_WORLD, ierr)
  call MPI_reduce(hN   , hNg       , Ncount, MPI_DOUBLE_PRECISION, MPI_SUM,Zero,MPI_COMM_WORLD, ierr)
  call MPI_reduce(hT   , hTg       , Ncount, MPI_DOUBLE_PRECISION, MPI_SUM,Zero,MPI_COMM_WORLD, ierr)
  call MPI_reduce(hP   , hPg       , Ncount, MPI_DOUBLE_PRECISION, MPI_SUM,Zero,MPI_COMM_WORLD, ierr)
#else
  Masse_glob = Masse
  hNg        = hN
  hPg        = hP
  hTg        = hT
#endif

  if (mype == 0) then
     hNg = hNg/masse_glob
     hPg = hPg/masse_glob
     hTg = hTg/masse_glob

     write(44,*)
     write(44,999) temps/an,(hNg(i),i=0,Npoint)
     write(44,999) temps/an,(hPg(i),i=0,Npoint)
     write(44,999) temps/an,(hTg(i),i=0,Npoint)

999  format(1pe10.3,3x,100(E10.3,2x))
  endif
  
  return

end subroutine cal_histogramme
