!> \file
!! Contains subroutines solver_relax(), solver_relaxation(),
!! solver_exact(), solver_cc(), solver_acoustic(),
!! solver_eos(), solver_g(), solver_iso(), starpu(),
!! guessp(), prefun(), sample(), starpu_iso(), guessp_iso(),
!! prefun_iso(), sample_iso() and cal_gamma()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

module solvers_hydro

  contains

     !  Subroutine SOLVER_RELAX
     !
     !> Relaxation Riemann solver provided by Benjamin
     !! Braconnier (26/06/2008).
     !<
     subroutine solver_relax(dll,ull,pll,cll,ell,fll,drr,urr,prr,crr,err,frr,dss,uss,pss,ess,fss,vit,idim,N) 
     
       use parameters, only : ndim,nfx
       use divers    , only : verbose
       use gammas
     
       implicit none
     
       integer                                     :: N
       real, dimension(1:N)          , intent(in ) :: dll,pll,drr,prr
       real, dimension(1:N), optional, intent(in ) :: cll,ell,crr,err
       real, dimension(1:N)                        :: ccll,eell,ccrr,eerr
       real, dimension(1:N,1:ndim)   , intent(in ) :: ull,urr
       real, dimension(1:N,1:nfx )   , intent(in ) :: fll,frr
       real, dimension(1:N)          , intent(out) :: dss,pss
       real, dimension(1:N), optional, intent(out) :: ess 
       real, dimension(1:N,1:ndim)   , intent(out) :: uss
       real, dimension(1:N,1:nfx )   , intent(out) :: fss
       integer                       , intent(in ) :: idim
       real                          , intent(in ) :: Vit
     
       integer                 :: i
       real, dimension(1:ndim) :: ul,ur,vno,us
       real                    :: dl,pl,dr,pr,cl,cr,el,er
       real                    :: ds,ps,es
     
       if (verbose) write(*,*) 'Start of solver_relax'
     
       if (present(cll)) then
     
          eell = ell
          ccll = cll
        
          eerr = err
          ccrr = crr
     
       else
     
          eell = pll/(gamma-1.)
          ccll = sqrt(gamma*pll/dll)
     
          eerr = prr/(gamma-1.)
          ccrr = sqrt(gamma*prr/drr)
     
       endif
     
     
       ! Caution! passive scalars are not yet treated!
       ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     
       do i = 1,N
     
          vno       = 0.
          vno(idim) = 1.
     
          ! Left state
          dl = dll(i)
          ul = ull(i,:)
          pl = pll(i)
     
          ! right state
          dr = drr(i)
          ur = urr(i,:)
          pr = prr(i)
     
          if (present(cll)) then
     
             el = ell(i)
             cl = cll(i)
     
             er = err(i)
             cr = crr(i)
     
          else
     
             el = pll(i)/(gamma-1.)
             cl = sqrt(gamma*pll(i)/dll(i))
     
             er = prr(i)/(gamma-1.)
             cr = sqrt(gamma*prr(i)/drr(i))                   
     
          endif
     
          call solver_relaxation(dl,ul,el,pl,cl,dr,ur,er,pr,cr,ds,us,es,ps,ndim,vno)
     
          dss(i)        = ds
          pss(i)        = ps
          uss(i,1:ndim) = us(1:ndim)
        
          if (vno(1) == 1.) then
             fss(i,1:nfx) = frr(i,1:nfx)
          else
             fss(i,1:nfx) = fll(i,1:nfx)
          endif
     
          if (present(ess)) ess(i) = es
        
       enddo
     
       if (verbose) write(*,*) 'End of solver_relax'
     
       return
     
     end subroutine solver_relax
     
!###########################################################
!###########################################################
!###########################################################
!###########################################################
     
     !  Subroutine SOLVER_RELAXATION
     !
     !> Relaxation Riemann solver provided by Benjamin
     !! Braconnier (26/06/2008).
     !<
     subroutine solver_relaxation(dl,ul,el,pl,cl,dr,ur,er,pr,cr,ds,us,es,ps,ndim,vno)
     
       use para, only : mype
     
       implicit none
     
       integer, intent(in   )                    :: ndim
       real   , intent(in   )                    :: dl,pl,dr,pr,cl,cr,el,er
       real   , intent(in   ), dimension(1:ndim) :: ul,ur
       real   , intent(inout), dimension(1:ndim) :: vno
       real   , intent(out  )                    :: ds,ps,es
       real   , intent(out  ), dimension(1:ndim) :: us
     
       ! local variables
       integer                 :: nvar
       real, dimension(1:ndim) :: ovitl,ovitr,ux,uy
       real                    :: pvitl,pvitr,pvitx,px,s1,s3,delt,a
       real                    :: dx,dy,etotl,etotr,etotx,etoty,epsl,epsr,ex,ey
     
       nvar = 2 + ndim
     
       ! description of left state
       ! we know dl, ul, pl
       epsl  = el/dl
       pvitl = sum(ul*vno)
       ovitl = ul - pvitl*vno
       etotl = dl*epsl + 0.5*dl*sum(ul*ul)
     
       ! description of right state
       ! we know dr, ur, pr
       epsr  = er/dr
       pvitr = sum(ur*vno)
       ovitr = ur - pvitr*vno
       etotr = dr*epsr + 0.5*dr*sum(ur*ur)
     
       ! wave speeds
       s1 = min( pvitr, pvitl) - max(cr, cl)
       s3 = max( pvitr, pvitl) + max(cr, cl)
     
       !a = max(dr*cr,dl*cl)
       !s1 = pvitl - a/dl
       !s3 = pvitr + a/dr
     
       ! calculate u*
       pvitx = ( dr*pvitr*(pvitr-s3) - dl*pvitl*(pvitl-s1) + pr - pl )/(dr*(pvitr-s3) - dl*(pvitl-s1))
     
       ! check the order of the waves
       ! apply an entropy fix/correction if the order is not correct
       if( pvitx .lt. s1 .or. pvitx .gt. s3 ) then
          ! consider a = constant
          write(*,*) 'entropic correction'
          write(*,*) pvitx, pvitl,pvitx
          read (*,*)
          a = max(dl*cl,dr*cr)
          delt = (pvitr-pvitl)**2 + 8.0*(pr-pl)/dl
          if(delt .ge. 0. ) then
             a = max(a,0.5*dl*( sqrt(delt)-(pvitr-pvitl) ) )
          end if
          delt = (pvitr-pvitl)**2 - 8.0*(pr-pl)/dr
          if(delt .GE. 0. ) then
             a = max(a, 0.5*dr*(sqrt(delt)-(pvitr-pvitl)) )
          end if
          a = a + epsilon(a)
          s1 = pvitl - a/dl
          s3 = pvitr + a/dr
       end if
     
       pvitx = ( dr*pvitr*(pvitr-s3) - dl*pvitl*(pvitl-s1) + pr - pl )/(dr*(pvitr-s3) - dl*(pvitl-s1))
     
       ! check the order of the waves
       ! stop if the order is not correct
       if( pvitx > s3) then
          write(*,*) 'pb in the wave order in relaxation solver' , 'pvitx > s3', pvitx, s3
          call close_program(mype,'called from SOLVER_RELAXATION')
       end if
     
       if( pvitx < s1) then
          write(*,*) 'pb in the wave order in relaxation solver' , 'pvitx < s1', pvitx, s1
          call close_program(mype,'called from SOLVER_RELAXATION')
       end if
     
       ! compute intermediate states
       px =  0.5*( pl + pr + dr*(pvitr-s3)*(pvitr-pvitx) + dl*(pvitl-s1)*(pvitl-pvitx) )
     
       etotx = ( (pvitl-s1)*etotl + pl*pvitl - px*pvitx )/(pvitx-s1)
       etoty = ( (pvitr-s3)*etotr + pr*pvitr - px*pvitx )/(pvitx-s3)
     
       dx = (s1-pvitl)*dl/(s1-pvitx)
       dy = (s3-pvitr)*dr/(s3-pvitx)
     
       ! compute global speeds
       ux = pvitx*vno + ovitl
       uy = pvitx*vno + ovitr
     
       ex = etotx - 0.5*dx*sum(ux*ux)
       ey = etoty - 0.5*dy*sum(uy*uy)
     
       if( s1 > 0.0 ) then 
          ds = dl; us = ul ; es = el ; ps = pl ; vno(1) = -1.  ! vno gives the direction of the flow
       elseif( pvitx > 0.0 ) then
          ds = dx; us = ux ; es = ex ; ps = px ; vno(1) = -1.
       elseif( s3 > 0.0 ) then
          ds = dy; us = uy ; es = ey ; ps = px ; vno(1) = +1.
       else
          ds = dr; us = ur ; es = er ; ps = pr ; vno(1) = +1.
       end if
     
       return
     
     end subroutine solver_relaxation
     
!###########################################################
!###########################################################
!###########################################################
!###########################################################
     
     !  Subroutine SOLVER_EXACT
     !
     !> Exact Riemann solver, taken from 'Riemann Solvers and
     !! Numerical Methods for Fluid Dynamics', E. F. Toro, 1999.
     !<
     subroutine solver_exact(dll,ull,pll,fll,drr,urr,prr,frr,ds,us,ps,fs,vit,idim,n) 
     
       use parameters, only : ndim,nfx
       use gammas
       use para      , only : mype
     
       implicit none
     
       integer                                  :: i,N,k
       integer                    , intent(in ) :: idim
       real, dimension(1:N,1:ndim), intent(in ) :: ull,urr
       real, dimension(1:N       ), intent(in ) :: pll,drr,prr
       real, dimension(1:N,1:nfx ), intent(in ) :: fll,frr
       real                       , intent(in ) :: Vit
       real, dimension(1:N,1:ndim), intent(out) :: us
       real, dimension(1:N,1:nfx ), intent(out) :: fs
       real, dimension(1:N       ), intent(out) :: ds,ps 
       real, dimension(1:N       )              :: dll
     
       real :: mpa,s,um,pm
       real :: dss,pss,uss
       real :: xx
       real :: dl,ul,pl,cl
       real :: dr,ur,pr,cr
       real :: eps
     
       do i = 1,N
     
          dl = dll(i) ; ul = ull(i,idim) ; pl = pll(i)
          dr = drr(i) ; ur = urr(i,idim) ; pr = prr(i)
     
          eps = 1.0e-08 * (dl+dr)*(ul**2+ur**2)
     
          pl = max(pl,eps)
          pr = max(pr,eps)
     
          if (pl <= 0. .or. pr <= 0. .or. dl<= 0. .or. dr <= 0.) then
             write(*,*) 'pl,pr,dl,dr =',pl,pr,dl,dr
             write(*,*) 'ul,ur,eps   =',ul,ur,eps
             call close_program(mype,'called from SOLVER_EXACT')
          endif
     
          mpa = 1.  ! constante de normalisation
     
          cl = sqrt(gamma*pl/dl)
          cr = sqrt(gamma*pr/dr)
     
          if (g4*(cl+cr) < (ur-ul) ) then
     
             xx = (ur-ul) - g4*(cl+cr) ; k = 0
             do while (xx > 0.)
                k = k+1
                pl = pl*1.05
                pr = pr*1.05
                cl = sqrt(gamma*pl/dl)
                cr = sqrt(gamma*pr/dr)
                xx = (ur-ul) - g4*(cl+cr)
                write(*,*) 'The temperature was increased in the solver_exact. k =',k
             enddo
     
          endif
     
          if (g4*(cl+cr) < (ur-ul) ) then
     
             write(*,*) 'The vacuum needs to be included: ur-ul =',(ur-ul)
     
             write(*,*) 'pl,pr,dl,dr =',pl,pr,dl,dr
             write(*,*) 'ul,ur,eps   =',ul,ur,eps
             write(*,*) 'i, idim     =',i,idim
             call close_program(mype,'called from SOLVER_EXACT')
     
             ! To avoid crashes
             us(i,1:ndim) = 0.
             fs(i,1:nfx ) = fll(i,1:nfx )
     
             us(i,idim) = 0.
             ds(i     ) = 1.0E-5*(dl+dr)
             ps(i     ) = 1.0E-5*(pl+pr)
             dll(i)     = -1.
     
          else
     
             call starpu(pm,um,mpa,dl,ul,pl,cl,dr,ur,pr,cr)
             uss = um
             pss = pm
     
             ! s = um  ! for Lagrangian
             ! s = 0.  ! for Eulerian
             s = Vit
             call sample(pm,um,s,dss,uss,pss,dl,ul,pl,cl,dr,ur,pr,cr)
     
             ! advection of tangential velocity
             if(Vit.lt.um) then
                us(i,1:ndim)=ull(i,1:ndim)
                fs(i,1:nfx )=fll(i,1:nfx ) 
             else
                us(i,1:ndim)=urr(i,1:ndim)
                fs(i,1:nfx )=frr(i,1:nfx ) 
             endif
     
             us(i,idim) = uss
             ds(i     ) = dss
             ps(i     ) = pss
     
          endif
     
       end do
     
       return
     
     end subroutine solver_exact
     
!###########################################################
!###########################################################
!###########################################################
!###########################################################
     
     !  Subroutine SOLVER_CC
     !
     !> Two shock Riemann solver.
     !<
     subroutine solver_cc(dll,ull,pll,fll,drr,urr,prr,frr,ds,us,ps,fs,vit,idim,n)
                           
       use parameters, only : ndim,nfx
       use gammas
     
       implicit none
     
       integer                                  :: i,N
       real, dimension(1:N       ), intent(in ) :: dll,pll,drr,prr
       real, dimension(1:N,1:nfx ), intent(in ) :: fll,frr
       real, dimension(1:N,1:ndim), intent(in ) :: ull,urr
       integer                    , intent(in ) :: idim
       real                       , intent(in ) :: vit
       real, dimension(1:N       ), intent(out) :: ds,ps
       real, dimension(1:N,1:ndim), intent(out) :: us
       real, dimension(1:N,1:nfx ), intent(out) :: fs
     
       real, dimension(1:ndim) :: uu
       real, dimension(1:nfx ) :: ff
       real                    :: wl,wr,qr,ql,usr,usl
       real                    :: pp,dd,ww,cc
       real                    :: pstar, ustar, dstar,cstar
       real                    :: delp,cor
       real                    :: sig,lamm,lams,xsi
       real                    :: dl,ul,pl,cl
       real                    :: dr,ur,pr,cr
     
       do i = 1,N
     
          dl = dll(i); ul = ull(i,idim); pl = pll(i)
          dr = drr(i); ur = urr(i,idim); pr = prr(i)
     
          ! Lagrangian sound speed
          cl = gamma*pl*dl
          cr = gamma*pr*dr
     
          ! First guess
          wl = sqrt(cl) ; wr = sqrt(cr)
          pstar = ((wr*pl+wl*pr)+wl*wr*(ul-ur))/(wl+wr)
     
          ! Newton-Raphson iterations to find pstar at the required accuracy
          ! for a two-shock Riemann problem
          cor = 1.
          do while(cor > 1.0E-6)
     
             wl = sqrt(cl*(1.+g2*(pstar-pl)/pl))
             wr = sqrt(cr*(1.+g2*(pstar-pr)/pr))
     
             ql = 2.*wl**3/(wl**2+cl)
             qr = 2.*wr**3/(wr**2+cr)
     
             usl = ul - (pstar-pl)/wl
             usr = ur + (pstar-pr)/wr
     
             delp = max(qr*ql/(qr+ql)*(usl-usr),-pstar)
     
             pstar = pstar + delp
     
             ! Convergence indicator
             cor = abs(delp/pstar)
     
          enddo
     
          ! Star region velocity
          ! for a two shock Riemann problem
          ustar = 0.5*(ul+(pl-pstar)/wl + ur-(pr-pstar)/wr)
     
          ! Left or right unperturbed state
          sig = sign(1.,ustar)
     
          if (sig > 0.) then
             uu(1:ndim) = ull(i,1:ndim)
             ff(1:nfx ) = fll(i,1:nfx ) 
             pp = pl
             dd = dl
             ww = wl
             cc = sqrt(gamma*pl/dl)
          else
             uu(1:ndim) = urr(i,1:ndim)
             ff(1:nfx ) = frr(i,1:nfx ) 
             pp = pr
             dd = dr
             ww = wr
             cc = sqrt(gamma*pr/dr)
          endif
     
          ! Shock
     
          if(pstar >= pp) then
             ! Shock
             dstar = 1./(1./dd - (pstar-pp)/ww/ww)
          else
             ! Rarefaction
             dstar = dd*(pstar/pp)**(1./gamma)
          end if
     
          ! Star region sound speed
          cstar = sqrt(gamma*pstar/dstar)
     
          if (pstar < pp) then
             cstar = sqrt(gamma*pstar/dstar) ! To be change for appropriate EOS
             lams = cstar - sig*ustar
             lamm = cc    - sig*uu(idim)
          else
             lams = ww/dd - sig*uu(idim)
             lamm = lams
          endif
     
          ! Sample the solution at x/t=0
          us(i,1:ndim) = uu(1:ndim) ! upwind transverse speeds
          fs(i,1:nfx ) = ff(1:nfx ) ! and passive scalars
          if(lams < 0.)then
             ds(i)       = dd
             us(i,idim)  = uu(idim)
             ps(i)       = pp
          else if(lams > 0)then
             ds(i)       = dstar
             us(i,idim)  = ustar
             ps(i)       = pstar
          else
             xsi=lamm/(lamm-lams)
             ds(i)      = xsi*dstar  + (1.-xsi)*dd
             ps(i)      = xsi*pstar  + (1.-xsi)*pp
             us(i,idim) = xsi*ustar  + (1.-xsi)*uu(idim)
          end if
     
       end do
      
       return
     
     end subroutine solver_cc
     
!###########################################################
!###########################################################
!###########################################################
!###########################################################
     
     !  Subroutine SOLVER_ACOUSTIC
     !
     !> Acoustic (linearised) Riemann solver.
     !<
     subroutine solver_acoustic(dll,ull,pll,fll,drr,urr,prr,frr,ds,us,ps,fs,vit,idim,n)
     
       use parameters, only : ndim,nfx
       use gammas
     
       implicit none
     
       integer                                  :: i,N
       real, dimension(1:N       ), intent(in ) :: dll,pll,drr,prr
       real, dimension(1:N,1:ndim), intent(in ) :: ull,urr
       real, dimension(1:N,1:nfx ), intent(in ) :: fll,frr
       integer                    , intent(in ) :: idim
       real                       , intent(in ) :: Vit
       real, dimension(1:N       ), intent(out) :: ds,ps 
       real, dimension(1:N,1:ndim), intent(out) :: us
       real, dimension(1:N,1:nfx ), intent(out) :: fs
     
       real, dimension(1:ndim) :: uu
       real, dimension(1:nfx ) :: ff
       real                    :: wl,wr
       real                    :: pp,dd,ww,cc
       real                    :: pstar,ustar,dstar,cstar
       real                    :: sig,lamm,lams,xsi,eps
       real                    :: dl,ul,pl,cl
       real                    :: dr,ur,pr,cr
     
       do i = 1,N
     
          dl = dll(i); ul = ull(i,idim); pl = pll(i)
          dr = drr(i); ur = urr(i,idim); pr = prr(i)
     
          eps = 1.0e-05 * (dl+dr)*(ul**2+ur**2)
         
          pl  = max(pl,eps)
          pr  = max(pr,eps)
     
          ! Acoustic star state
          cl = sqrt(gamma*pl/dl)
          cr = sqrt(gamma*pr/dr)
          wl = cl*dl
          wr = cr*dr
     
          pstar = ((wr*pl+wl*pr)+wl*wr*(ul-ur))/(wl+wr)
          
          ustar = ((wr*ur+wl*ul) + (pl-pr))/(wl+wr)
     
          ! Left or right unperturbed state
          sig = sign(1.,ustar)
     
          if (sig > 0.) then
             uu(1:ndim) = ull(i,1:ndim)
             ff(1:nfx ) = fll(i,1:nfx ) 
             pp = pl
             dd = dl
             ww = wl
             cc = cl
          else
             uu(1:ndim) = urr(i,1:ndim)
             ff(1:nfx ) = frr(i,1:nfx ) 
             pp = pr
             dd = dr
             ww = wr
             cc = cr
          endif
     
          ! density and sounds speed in the star region
          dstar = dd + (pstar - pp)/cc**2
     
          if (pstar <= 0. .or. dstar <= 1.0E-100) then
             write(*,800) dl,ul,pl
             write(*,801) dr,ur,pr
             write(*,802) wl,wr
             write(*,803) pstar,dstar
             pstar = max(pstar,eps)
          endif
     800   format('dl,ul,pl =',3(E10.3,2X))
     801   format('dr,ur,pr =',3(E10.3,2X))
     802   format('wl,wr    =',3(E10.3,2X))
     803   format('ps,ds    =',3(E10.3,2X))
      
          cstar = sqrt(gamma*pstar/dstar)
     
          ! wave speeds
          lams = cstar - sig*ustar
          lamm = cc    - sig*uu(idim)
         
          if (pstar > pp) then
             lams = 0.5*(lams+lamm)
             lamm = lams
          endif
         
          ! Sample the solution at x/t=0
          us(i,1:ndim) = uu(1:ndim) ! upwind transvers speeds
          fs(i,1:nfx ) = ff(1:nfx ) ! and passive scalars
          if(lamm < 0.)then
             ds(i)       = dd
             us(i,idim)  = uu(idim)
             ps(i)       = pp
          else if(lams > 0)then
             ds(i)       = dstar
             us(i,idim)  = ustar
             ps(i)       = pstar
          else
             xsi=lamm/(lamm-lams)
             ps(i)      = xsi*pstar  + (1.-xsi)*pp
             us(i,idim) = xsi*ustar  + (1.-xsi)*uu(idim)
             ds(i)      = xsi*dstar  + (1.-xsi)*dd
             ! ds       = dd*(ps/pp)**(1./gamma)
          end if
     
       end do
     
       return
     
     end subroutine solver_acoustic
     
!###########################################################
!###########################################################
!###########################################################
!###########################################################
     
     !  Subroutine SOLVER_EOS
     !
     !> Riemann solver from the FLASH paper.
     !<
     subroutine solver_eos(dll,ull,pll,cll,ell,fll,drr,urr,prr,crr,err,frr,dss,uss,pss,ess,fss,vit,idim,N) 
     
       use parameters, only : ndim,nfx
       use divers    , only : verbose
       use gammas
       use para      , only : mype
       
       implicit none
     
       integer                                  :: N
       real, dimension(1:N)       , intent(in ) :: dll,pll,cll,ell,drr,prr,crr,err
       real, dimension(1:N,1:ndim), intent(in ) :: ull,urr
       real, dimension(1:N,1:nfx ), intent(in ) :: fll,frr
       real, dimension(1:N)       , intent(out) :: dss,pss,ess 
       real, dimension(1:N,1:ndim), intent(out) :: uss
       real, dimension(1:N,1:nfx ), intent(out) :: fss
       integer                    , intent(in ) :: idim
       real                       , intent(in ) :: vit
     
       integer, dimension(1:n) :: Niter
       integer                 :: i,j,ii
     
       real, dimension(1:N)        :: css,pp,dd,ww,cc,ee,gams,sig
       real, dimension(1:N,1:ndim) :: uu
       real, dimension(1:N,1:nfx ) :: ff
       real, dimension(1:N)        :: grr,gll,gmin,gmax
       real, dimension(1:N)        :: prec
       real                        :: dl,ul,pl,cl,el,dr,ur,pr,cr,er,ds,us,ps
       real                        :: ggl,ggr,gpl,gpr,gprs,gpls,ps1,ps2
       real                        :: wl, wr,urs1,urs2,uls1,uls2,gpp
       real                        :: gghat,gphat,xsi,lams,lamm 
       real                        :: eps,dif,xp,xu
       real                        :: xx
     
       if (verbose) write(*,*) 'Start of solver_eos'
       
       eps = 1.0e-7 
     
       ! A revoir pour une grille cartesienne....
     
       ! calculate gmin and gmax in order to limit the variations of gamma
       do i = 1,n
          grr(i) =  1. + prr(i)/err(i)
          gll(i) =  1. + pll(i)/ell(i)
       enddo
       
       gmin(1:n) =  1000.
       gmax(1:n) = -1000.
       do j = -2,2
          do i = 3,n-2
             gmin(i) = min(gmin(i),grr(i+j))
             gmin(i) = min(gmin(i),gll(i+j))
             
             gmax(i) = max(gmax(i),grr(i+j))
             gmax(i) = max(gmax(i),gll(i+j))
          enddo
       enddo
       gmin(1  ) = gmin(3  ) ; gmin(2) = gmin(3  )
       gmax(1  ) = gmax(3  ) ; gmax(2) = gmax(3  )
     
       gmin(n-1) = gmin(n-2) ; gmin(n) = gmin(n-2)
       gmax(n-1) = gmax(n-2) ; gmax(n) = gmax(n-2)
     
       do i = 1,n
          gmin(i) = max(0.5*gmin(i),1.0001)
          gmax(i) = 1.5*gmax(i)
       enddo
     
     
       do i = 1,n
     
          dl = dll(i) ; ul = ull(i,idim) ; pl = pll(i) ; cl = cll(i) ; el = ell(i)
          dr = drr(i) ; ur = urr(i,idim) ; pr = prr(i) ; cr = crr(i) ; er = err(i)
     
          Ggl = dl*cl*cl/pl 
          Ggr = dr*cr*cr/pr  
     
          gpl = 1. + pl/el  ! ell : erg/cm3 (volumic energy)
          gpr = 1. + pr/er
     
     
          xp = 2.*abs(pl-pr)/(pl+pr)
          xu = abs(ur)+abs(ul)
          if (xu .ne. 0.) xu = 2.*abs(ur-ul)/xu
     
          ! if (verbose) write(*,*) '-- 1 --', dl,dr,ul,ur, pl,pr
          
          if (xp < 1.0E-10 .and. xu < 1.0E-10) then
             ps = pl
             us = ul
             wl = cl*dl ; wr = cr*dr
     
             gpls = gpl
             gprs = gpr
     
          else
     
             ! First guess for the pressure
             ps1 = (cr*dr*pl + cl*dl*pr + cr*dr*cl*dl*(ul-ur))/(cr*dr+cl*dl) 
             !if (verbose) write(*,*) 'ps1 =', ps1
     
             urs1 = ur + (ps1-pr)/(cr*dr)
             uls1 = ul - (ps1-pl)/(cl*dl)
     
             !if (verbose) then
             !   write(*,*) '-- 2a --'
             !   write(*,*) 'urs1, uls1 = ',urs1, uls1
             !endif
     
             ! Gamma jumps in the two shocks
             Gghat = 0.5*(Ggl+Ggr) ; gphat = 0.5*(gpl+gpr)
     
             gprs = gpr + 2.*(1.-gphat/Gghat)*(gphat-1.)*(ps1-pr)/(ps1+pr)
             gpls = gpl + 2.*(1.-gphat/Gghat)*(gphat-1.)*(ps1-pl)/(ps1+pl)
     
             gprs = max(gprs,gmin(i)) ; gprs = min(gprs,gmax(i))
             gpls = max(gpls,gmin(i)) ; gpls = min(gpls,gmax(i))
             
     
             ! Wave speeds
             xx = abs(gpl-gpls)
             if (xx < 1.0E-12) then
                Wl = dl*                  (ps1+0.5*(gpls-1.)*(ps1+pl)) ! suppose gpl = gpls
             else
                Wl = dl*(ps1-pl)*(gpl-1.)*(ps1+0.5*(gpls-1.)*(ps1+pl))/((gpl-1.)*ps1 -(gpls-1.)*pl)
             endif
     
             xx = abs(gpr-gprs)
             if (xx < 1.0E-12) then
                Wr = dr*                  (ps1+0.5*(gprs-1.)*(ps1+pr)) ! suppose gpr = gprs
             else
                Wr = dr*(ps1-pr)*(gpr-1.)*(ps1+0.5*(gprs-1.)*(ps1+pr))/((gpr-1.)*ps1 -(gprs-1.)*pr)
             endif
     
             !Wl = dl*(ps1-pl)*(gpl-1.)*(ps1+0.5*(gpls-1.)*(ps1+pl))/((gpl-1.)*ps1 -(gpls-1.)*pl)
             !Wr = dr*(ps1-pr)*(gpr-1.)*(ps1+0.5*(gprs-1.)*(ps1+pr))/((gpr-1.)*ps1 -(gprs-1.)*pr)
     
             Wl = sqrt(Wl) ; Wr = sqrt(Wr)
     
     
             ! Second guess for the pressure
             ps2 = (Wr*pl + Wl*pr + Wr*Wl*(ul-ur))/(Wr+Wl)
     
             ! Convergence iterations
             dif = 1. ; ii = 0
     
             urs2 = ur + (ps2-pr)/Wr
             uls2 = ul - (ps2-pl)/Wl
             if ((abs(urs2-urs1)+abs(uls2-uls1))==0. ) then
                ps  = ps2
                dif = 1.0E-30
             else
                ! included to avoid problems at the first iteration
                ps = ps2 - (urs2-uls2)*(abs(ps2-ps1)/(abs(urs2-urs1) + abs(uls2-uls1) ) ) 
                dif = abs(ps-ps2)/ps
             endif
     
             !if (verbose) then
             !   write(*,*) 'wl,wr =',wl,wr
             !   write(*,*) 'ps2 = ',ps2
             !   write(*,*) 'urs2, uls2 = ',urs2, uls2
             !   write(*,*) 'dif        = ',dif
             !endif
     
          
             do while (dif > eps ) 
     
                ii = ii + 1
     
                ps = ps2 - (urs2-uls2)*(abs(ps2-ps1)/(abs(urs2-urs1) + abs(uls2-uls1) ) ) 
     
                dif = abs(ps-ps2)/ps
     
                 ps1 =  ps2
                urs1 = urs2
                uls1 = uls2
     
                ps2 = ps
     
                gprs = gpr + 2.*(1.-gphat/Gghat)*(gphat-1.)*(ps-pr)/(ps+pr)
                gpls = gpl + 2.*(1.-gphat/Gghat)*(gphat-1.)*(ps-pl)/(ps+pl)
     
                gprs = max(gprs,gmin(i)) ; gprs = min(gprs,gmax(i))
                gpls = max(gpls,gmin(i)) ; gpls = min(gpls,gmax(i))
     
                Wl = dl*(ps-pl)*(gpl-1.)*(ps+0.5*(gpls-1.)*(ps+pl))/((gpl-1.)*ps -(gpls-1.)*pl)
                Wr = dr*(ps-pr)*(gpr-1.)*(ps+0.5*(gprs-1.)*(ps+pr))/((gpr-1.)*ps -(gprs-1.)*pr)
     
                Wl = sqrt(Wl) ; Wr = sqrt(Wr)
     
                urs2 = ur + (ps2-pr)/Wr
                uls2 = ul - (ps2-pl)/Wl
     
             end do
     
             Niter(i) = ii ; prec(i) = dif
     
             us = 0.5*(urs2+uls2) ! normally we should have urs2 = uls2...
             
     
          endif
     
          ! if (verbose) write(*,*) '-- 2 --'
     
          sig(i) = sign(1.,us-Vit)
     
          if (sig(i) > 0.) then
             uu(i,1:ndim) = ull(i,1:ndim)
             ff(i,1:nfx ) = fll(i,1:nfx ) 
             pp(i) = pl
             dd(i) = dl
             ww(i) = wl
             cc(i) = cl
             ee(i) = el
             gpp   = gpls ; gams(i) = gpls
          else
             uu(i,1:ndim) = urr(i,1:ndim)
             ff(i,1:nfx ) = frr(i,1:nfx ) 
             pp(i) = pr
             dd(i) = dr
             ww(i) = wr
             cc(i) = cr
             ee(i) = er
             gpp   = gprs ; gams(i) = gprs
          endif
     
          if ( (ps - pp(i)) > 0.) then
             ds = dd(i)*(1 + (ps - pp(i))*dd(i)/ww(i)/ww(i) )
          else
             ds = 1./(1./dd(i) - (ps-pp(i))/ww(i)/ww(i))
          endif
     
          if (ds < 0.) then
             write(*,*) 'i     = ',i
             write(*,*) 'dd(i) = ',dd(i)
             write(*,*) 'ps,pp = ',ps,pp(i)
             write(*,*) 'ww    = ',ww(i)
             write(*,*) 'ds    = ',ds
             write(*,*) 'urs2,uls2  =', urs2,uls2 
             write(*,*) 
             write(*,*) 'dr,dl =',drr(i),dll(i)
             write(*,*) 'ur,ul =',urr(i,1),ull(i,1)
             write(*,*) 'pr,pl =',prr(i),pll(i)
     
             do j=-2,2
                ii = max(i+j,1)
                write(*,*) prr(ii),pll(ii)
             enddo
             !call close_program(mype,'called from SOLVER_EOS: ds < 0')
          endif
     
          dss(i) = ds ; uss(i,idim) = us ; pss(i) = ps 
     
          ess(i) = ps/(gpp - 1.)
     
          ! Compute sound speed
          Css(i) = sqrt(gpp*pss(i)/dss(i))
       enddo
     
     
       do i=1,n
          if (dss(i) < 0.) then
             write(*,*) i,dss(i),ww(i),pss(i)
             call close_program(mype,'called from SOLVER_EOS: dss < 0')
          endif
       enddo
       
     
       ! if (verbose) write(*,*) '-- 3 --'
     
       do i=1,n
     
          if (pss(i) < pp(i)) then
             lams = css(i) - sig(i)*uss(i,idim)
             lamm = cc (i) - sig(i)*uu (i,idim)
          else
             lams = ww(i)/dd(i) - sig(i)*uu(i,idim)
             lamm = ww(i)/dd(i) - sig(i)*uu(i,idim)
          endif
     
     
     !     uss(i,1:ndim) = uu(i,1:ndim) ! upwind transverse speeds
          fss(i,1:nfx ) = ff(i,1:nfx ) ! and passive scalars
          if      (lams > -sig(i)*Vit) then
             dss(i     ) = dss(i     )
             uss(i,idim) = uss(i,idim)
             pss(i     ) = pss(i     )
             ess(i     ) = ess(i     )
          else if (lamm < -sig(i)*Vit) then
             dss(i     ) = dd(i     )
             uss(i,idim) = uu(i,idim)
             pss(i     ) = pp(i     )
             ess(i     ) = ee(i     )
          else  
             xsi = -sig(i)*Vit/(lams-lamm) + lamm/(lamm-lams)
             if (xsi > 1. .or. xsi <0.) then
                write(*,*) 'xsi= ',xsi
                call close_program(mype,'called from SOLVER_EOS: xsi > 1. .or. xsi < 0.')
             endif
             dss(i     )  = xsi*dss(i     ) + (1.-xsi)*dd(i     )
             uss(i,idim)  = xsi*uss(i,idim) + (1.-xsi)*uu(i,idim)
             pss(i     )  = xsi*pss(i     ) + (1.-xsi)*pp(i     )
             ess(i     )  = xsi*ess(i     ) + (1.-xsi)*ee(i     )
          endif
     
       enddo
     
       if (verbose) write(*,*) ' End of solver_eos'
     
       return
       
     end subroutine solver_eos
     
!###########################################################
!###########################################################
!###########################################################
!###########################################################
     
     !  Subroutine SOLVER_G
     !
     !> Variable gamma Riemann solver.
     !<
     subroutine solver_g(dll,ull,pll,fll,drr,urr,prr,frr,ds,us,ps,fs,vit,idim,n) 
     
       use parameters, only : ndim,nfx
       use gammas
       use para      , only : mype
     
       implicit none
     
       integer :: i,N
       
       real, dimension(1:N,1:ndim), intent(in ) :: ull,urr
       real, dimension(1:N,1:nfx ), intent(in ) :: fll,frr
       integer                    , intent(in ) :: idim
       real                       , intent(in ) :: Vit
       real, dimension(1:N       ), intent(in ) :: dll,pll,drr,prr
       real, dimension(1:N,1:ndim), intent(out) :: us
       real, dimension(1:N,1:nfx ), intent(out) :: fs
       real, dimension(1:N       ), intent(out) :: ds,ps 
     
       real :: mpa,s,um,pm
       real :: dss,pss,uss,yy
       real :: dl,ul,pl,cl
       real :: dr,ur,pr,cr
       real :: eps
     
     
       do i = 1,N
     
          dl = dll(i); ul = ull(i,idim); pl = pll(i)
          dr = drr(i); ur = urr(i,idim); pr = prr(i)
     
          eps = 1.0e-08 * (dl+dr)*(ul**2+ur**2)
          
          pl  = max(pl,eps)
          pr  = max(pr,eps)
       
          if (pl <= 0. .or. pr <= 0. .or. dl<= 0. .or. dr <= 0.) then
             write(*,*) 'pl,pr,dl,dr =',pl,pr,dl,dr
             write(*,*) 'ul,ur,eps   =',ul,ur,eps
             call close_program(mype,'called from SOLVER_G')
          endif
     
          mpa     = 1.  ! normalisation constant
     
          ! compute gamma
          yy = 0.5*(fll(i,1)+frr(i,1))
          call cal_gamma(yy)
     
          cl = sqrt(gamma*pl/dl)
          cr = sqrt(gamma*pr/dr)
     
          if (g4*(cl+cr) < (ur-ul) ) then
     
             write(*,*) 'Vacuum needs to be included: ur-ul =',(ur-ul)
             call close_program(mype,'called from SOLVER_G')
     
          else
     
             call starpu(pm,um,mpa,dl,ul,pl,cl,dr,ur,pr,cr)
             uss = um
             pss = pm
     
             ! s = um  ! for Lagrangian
             ! s = 0.  ! for Eulerian
             s = Vit
             call sample(pm,um,s,dss,uss,pss,dl,ul,pl,cl,dr,ur,pr,cr)
     
             ! advection of tangential velocity
             if(Vit.lt.um) then
                us(i,1:ndim)=ull(i,1:ndim)
                fs(i,1:nfx )=fll(i,1:nfx ) 
             else
                us(i,1:ndim)=urr(i,1:ndim)
                fs(i,1:nfx )=frr(i,1:nfx ) 
             endif
     
             us(i,idim) = uss
             ds(i     ) = dss
             ps(i     ) = pss
     
          endif
          
       end do
     
       return
       
     end subroutine solver_g
     
!###########################################################
!###########################################################
!###########################################################
!###########################################################
        
     !  Subroutine SOLVER_ISO
     !
     !> Isothermal Riemann solver.
     !<
     subroutine solver_iso(dll,ull,fll,drr,urr,frr,ds,us,fs,Vit,idim,N) 
     
       use parameters    , only : ndim,nfx
       use gammas        , only : Cs_iso,Cs_iso2
       use para          , only : mype
     
       implicit none
     
       integer :: i,N
        
       real, dimension(1:N       ), intent(in ) :: drr,dll
       real, dimension(1:N,1:ndim), intent(in ) :: ull,urr
       real, dimension(1:N,1:nfx ), intent(in ) :: fll,frr
       integer                    , intent(in ) :: idim
       real, dimension(1:N,1:ndim), intent(out) :: us
       real, dimension(1:N,1:nfx ), intent(out) :: fs
       real, dimension(1:N       ), intent(out) :: ds
       real                                     :: Vit
     
       real :: mpa,s,um,dm
       real :: dss,uss
       real :: dl,ul
       real :: dr,ur
       real :: cs,cs2
     
       Cs  = Cs_iso
       Cs2 = Cs_iso2
     
       do i = 1,N
     
          dl = dll(i); ul = ull(i,idim)
          dr = drr(i); ur = urr(i,idim)
                    
          if (dl < 0. .or. dr < 0.) then
             write(*,*) '**** dl, dr =',dr,dl
             vit = -1.
             return
          endif
     
          mpa     = 1.  ! normalisation constant
     
          call starpu_iso(dm,um,mpa,dl,ul,dr,ur,cs,cs2)
          uss = um
          dss = dm
     
          ! s = um  ! for Lagrangian
          ! s = 0.  ! for Eulerian
          s = Vit
          call sample_iso(dm,um,s,dss,uss,dl,ul,dr,ur,cs,cs2)
              
          ! advection of tangential velocity
          if(Vit.lt.um) then
             us(i,1:ndim)=ull(i,1:ndim)
             fs(i,1:nfx )=fll(i,1:nfx ) 
          else
             us(i,1:ndim)=urr(i,1:ndim)
             fs(i,1:nfx )=frr(i,1:nfx ) 
          endif
     
          us(i,idim) = uss
          ds(i     ) = dss
     
       end do
      
       return
        
     end subroutine solver_iso
     
!###########################################################
!###########################################################
!###########################################################
!###########################################################
     
     !  Subroutine STARPU
     !
     !> Computes the pressure and velocity in the star region.
     !<
     subroutine starpu(p,u,mpa,dl,ul,pl,cl,dr,ur,pr,cr)
     
       implicit none
     
       integer :: i,nriter
       real    :: p,u,mpa
       real    :: change,fl,fld,fr,frd,pold,pstart,tolpre,udiff,pmin
       real    :: dl,ul,pl,cl,dr,ur,pr,cr
     
       pmin   = 1.0e-10 * max((dl+dr)*(ul**2+ur**2),pr,pl)
       tolpre = 1.0E-8
       nriter = 20
     
       call guessp(pstart,dl,ul,pl,cl,dr,ur,pr,cr)
     
       pold  = max(pstart,pmin)
       udiff = ur - ul
     
       do i = 1,nriter 
          call prefun(fl,fld,pold,dl,pl,cl)
          call prefun(fr,frd,pold,dr,pr,cr)
     
          p = pold - (fl + fr + udiff)/(fld+frd)
          change = 2.0*abs(p-pold)/(p+pold)
          if(change < tolpre) goto 20
          if (p < pmin) p=pmin  
          pold = p
       enddo
     
       write(*,*) 'NR did not converge in STARPU'
       write(*,601) dl,dr
       write(*,602) pl,pr
       write(*,603) cl,cr
       write(*,604) change
       write(*,605) pmin
     
       601 format('Densite  : ',10(E12.4,2x))
       602 format('Pression : ',10(E12.4,2x))
       603 format('Cs       : ',10(E12.4,2x))
       604 format('Change   : ',10(E12.4,2x))
       605 format('Pmin     : ',10(E12.4,2x))
     
       20 continue
     
       u = 0.5*(ul + ur + fr - fl)
     
       return
     
     end subroutine starpu
     
!###########################################################
!###########################################################
!###########################################################
!###########################################################
     
     !  Subroutine GUESSP
     !
     !> Computes an intelligent first guess for the pressure.
     !<
     subroutine guessp(pm,dl,ul,pl,cl,dr,ur,pr,cr)
     
       use gammas
     
       implicit none
     
       real, intent(out) :: pm
       real              :: cup,gel,ger,pmax,pmin,ppv,pq,ptl,ptr,qmax,quser,um
       real              :: dl,ul,pl,cl,dr,ur,pr,cr
     
       quser = 2.
     
       ! Pressure from the PVRS solver
       cup  = 0.25*(dl+dr)*(cl+cr)
       ppv  = 0.5 *(pl+pr)+0.5*(ul-ur)*cup
       ppv  = max(0.0,ppv)
       pmin = min(pl,pr)
       pmax = max(pl,pr)
       qmax = pmax/pmin
     
       if (qmax < quser .and. (pmin <= ppv .and. ppv <= pmax)) then
       
          ! choose PVRS pressure
          pm = ppv
     
       else
      
          if (ppv < pmin) then
          
             ! choose two-rarefaction wave solver
             pq  = (pl/pr)**g1
             um  = (pq*ul/cl + ur/cr + g4*(pq-1.0))/(pq/cl + 1.0/cr)
             ptl = 1.0 + g7*(ul-um)/cl
             ptr = 1.0 + g7*(um-ur)/cr
             pm  = 0.5*(pl*ptl**g3 + pr*ptr**g3)
     
          else
      
             ! choose two-shock wave solver
             gel = sqrt((g5/dl)/(g6*pl+ppv))
             ger = sqrt((g5/dr)/(g6*pr+ppv))
             pm  = (gel*pl + ger*pr - (ur-ul))/(gel+ger)
         
          endif
     
       endif
     
       return
     
     end subroutine guessp
     
!###########################################################
!###########################################################
!###########################################################
!###########################################################
     
     !  Subroutine PREFUN
     !
     !> Evaluate pressure functions FL and FR.
     !<
     subroutine prefun(f,fd,p,dk,pk,ck)
     
       use gammas
     
       implicit none
     
       real :: f,fd,p,dk,pk,ck
       real :: ak,bk,prat,qrt
     
       if (p < pk) then   ! rarefaction wave
          prat = p/pk
          f    = g4*ck*(prat**g1 - 1.0)
          fd   = (1.0/(dk*ck))*prat**(-g2)
     
       else               ! shock
          ak  = g5/dk
          bk  = g6*pk
          qrt = sqrt(ak/(bk + p))
          f   = (p-pk)*qrt
          fd  = (1.0-0.5*(p-pk)/(bk+p))*qrt
       endif
     
       return
     
     end subroutine prefun
     
!###########################################################
!###########################################################
!###########################################################
!###########################################################
     
     !  Subroutine SAMPLE
     !
     !> Search through the different waves to find the correct
     !! solution.
     !<
     subroutine sample(pm,um,s,d,u,p,dl,ul,pl,cl,dr,ur,pr,cr)
     
       use gammas
     
       implicit none
     
       real, intent(in ) :: pm,um,s,dl,ul,pl,cl,dr,ur,pr,cr
       real, intent(out) :: d,p,u
     
       real :: c,cml,cmr,pml,pmr,shl,shr,sl,sr,stl,str
     
       if (s < um) then              ! We are to the LEFT of contact discontinuity
     
          if (pm < pl) then          ! Left rarefaction wave
         
             shl = ul - cl
         
             if  (s < shl) then      ! This is the left state
     
                d = dl ; u = ul ; p = pl
         
             else
            
                cml = cl*(pm/pl)**g1
                stl = um - cml
            
                if ( s > stl ) then  ! This is the left star state
     
                   d = dl*(pm/pl)**(1.0/gamma) ; u = um ; p = pm
               
                else                 ! This is the left set of solutions
                
                   c = g5*(cl + g7*(ul-s))  
                   d = dl*(c/cl)**g4
                   u = g5*(cl + g7*ul + s)
                   p = pl*(c/cl)**g3
               
                endif
     
             endif
       
          else                       ! Left shock
     
             pml = pm/pl
             sl  = ul - cl*sqrt(g2*pml + g1)
     
             if ( s < sl ) then      ! This is the left state
     
                d = dl ; u = ul ; p = pl
            
             else                    ! This is the left star state
     
                d = dl*(pml + g6)/(pml*g6 + 1.0) ; u = um ; p = pm
     
             endif
     
          endif
     
       else                          ! We are to the RIGHT of contact discontinuity
         
          if (pm > pr) then          ! Right shock
          
             pmr = pm/pr
             sr  = ur + cr*sqrt(g2*pmr + g1)
         
             if ( s > sr) then       ! This is the right state
     
                d = dr ; u = ur ; p = pr
     
             else                    ! This is the right star state
     
                d = dr*(pmr + g6)/(pmr*g6 + 1.0) ;  u = um ; p = pm
     
             endif
            
          else                       ! Right rarefaction wave
     
             shr = ur + cr
     
             if (s > shr) then       ! This is the right state
     
                d = dr; u = ur; p=pr
     
             else
     
                cmr = cr*(pm/pr)**g1
                str = um + cmr
            
                if ( s < str) then   ! This is the right star state
                
                   d = dr*(pm/pr)**(1.0/gamma) ; u = um ; p = pm
     
                else                 ! This is the right set of solutions
     
                   c = g5*(cr - g7*(ur-s))  
                   d = dr*(c/cr)**g4 ; u = g5*(-cr + g7*ur + s) ; p = pr*(c/cr)**g3
     
                endif                  
     
             endif
         
          endif
     
       endif
     
       return
     
     end subroutine sample
     
!###########################################################
!###########################################################
!###########################################################
!###########################################################
     
     !  Subroutine STARPU_ISO
     !
     !> Computes the pressure and velocity in the star region.
     !<
     subroutine starpu_iso(d,u,mpa,dl,ul,dr,ur,cs,cs2)
      
       use para, only : mype
     
       implicit none
      
       integer :: i,nriter
       real    :: change,fl,fld,fr,frd,dold,dstart,toldens,udiff,densmin
       real    :: d,u,mpa,dl,ul,dr,ur,cs,cs2
     
       densmin = 0.5*(dr+dl) * 1.0E-10
       toldens = 1.0E-06
       nriter  = 20
      
       call guessp_iso(dstart,dl,ul,dr,ur,cs,cs2)
     
       if (dstart == 0.) write(*,*) 'dstart =',dstart
       
       dold  = dstart
       udiff = ur - ul
      
       do i = 1,nriter 
          call prefun_iso(fl,fld,dold,dl,cs)
          call prefun_iso(fr,frd,dold,dr,cs)
     
          d = dold - (fl + fr + udiff)/(fld+frd)
          change = 2.0*(d-dold)/(d+dold)
          if(change < toldens) goto 20
          if (d < densmin) d=toldens      
          dold = d
       enddo
     
       write(*,*) 'NR did not converge in STARPU_ISO. mype =',mype
       write(*,*) 'DL, DR =',DL, DR
       write(*,*) 'UL, UR =',UL, UR
       write(*,*) 'dstart =',dstart
       write(*,*) 'cs_iso =',cs
     
       20 continue
      
       u = 0.5*(ul + ur + fr - fl)
     
       return
     
     end subroutine starpu_iso
     
!###########################################################
!###########################################################
!###########################################################
!###########################################################
     
     !  Subroutine GUESSP_ISO
     !
     !> Computes an intelligent 'first guess' for the pressure.
     !<
     subroutine guessp_iso(dm,dl,ul,dr,ur,cs,cs2)
     
       implicit none
     
       real, intent(out):: dm
     
       real :: cup,gel,ger,dmax,dmin,dpv,qmax,quser
       real :: dl,ul,dr,ur,cs,cs2
     
       quser = 2.
      
       ! Pressure from the PVRS solver
     
       cup  = 0.5*(dl+dr)/Cs
       dpv  = 0.5 *(dl+dr) +0.5*(ul-ur)*cup
       dpv  = max(0.0,dpv)
       dmin = min(dl,dr)
       dmax = max(dl,dr)
       qmax = dmax/dmin
     
       if (qmax < quser .and. (dmin <= dpv .and. dpv <= dmax)) then
          ! choose density given by PVRS
          dm = dpv
     
          if (dm == 0.) then
             write(*,*) 'dm = dpv = ',dm
             write(*,*) 'ul,ur,cs  =',ul,ur,cs
             write(*,*) 'dl,dr =',dl,dr
          endif
     
       else
     
          if (dpv < dmin) then
             ! choose two-rarefaction wave solver
             dm = sqrt(dr*dl) * exp((ul-ur)/(2.*Cs))
     
             if (dm == 0.) then
                write(*,*) 'dm = drr = ',dm
                write(*,*) 'ul,ur,cs =',ul,ur,cs
                write(*,*) 'dl,dr =',dl,dr
             endif
     
          else
             ! choose two-shock solver
             gel = sqrt(1./(dl*dpv))
             ger = sqrt(1./(dr*dpv))
             dm  = (gel*dl + ger*dr - (ur-ul)/Cs ) / (gel+ger)
     
             if (dm == 0.) then
                write(*,*) 'dm = dcc = ',dm
                write(*,*) 'ul,ur,cs =',ul,ur,cs
                write(*,*) 'dl,dr =',dl,dr
             endif
     
          endif
     
       endif
      
       return
     
     end subroutine guessp_iso
     
!###########################################################
!###########################################################
!###########################################################
!###########################################################
     
     !  Subroutine PREFUN_ISO
     !
     !> Computes pressure functions FL and FR.
     !<
     subroutine prefun_iso(f,fd,d,dk,cs)
     
       implicit none
     
       real :: f,fd,d,dk,cs
       real :: ak,drat,qrt
     
       if (d < dk) then   ! rarefaction wave 
          drat = d/dk
          f    = Cs*log(d/dk)
          fd   = Cs/d
     
       else               ! shock
          ak  = 1/dk
          qrt = sqrt(1/(dk*d))
          f   = Cs*(d-dk)*qrt
          fd  = Cs*(1.0-0.5*(d-dk)/d)*qrt
       endif
     
       return
       
     end subroutine prefun_iso
     
!###########################################################
!###########################################################
!###########################################################
!###########################################################
     
     !  Subroutine SAMPLE_ISO
     !
     !> Find solution looking through the different waves.
     !<
     subroutine sample_iso(dm,um,s,d,u,dl,ul,dr,ur,cs,cs2)
     
       implicit none
     
       real, intent(in ) :: dm,um,s 
       real, intent(out) :: d,u
     
       real :: dml,dmr,shl,shr,sl,sr,stl,str,lnd
       real :: dl,ul,dr,ur,cs,cs2
      
       if (s < um) then              ! We are to the LEFT of contact discontinuity
     
          if (dm < dl) then          ! Left rarefaction wave
             
             shl = ul - cs
             
             if (s < shl) then       ! This is the left state
             
                d = dl ; u = ul
             
             else
               
                stl = um - cs
               
                if (s > stl) then    ! This is the left star state
                
                   u = um ; d = dm
                   
                else                 ! This is the left set of solutions
                
                   lnd = (ul-cs-s)/cs + log(dl)
                   d = exp(lnd) ; u = (cs + s) 
                  
                endif
                
             endif
          
          else                       ! Left shock
     
             dml = dm/dl
             sl  = ul - cs*sqrt(dml)
     
             if (s < sl) then        ! This is the left state
             
                d = dl ; u = ul 
               
             else                    ! This is the left star state
             
                d = dm ; u = um
             
             endif
          
          endif
     
       else                          ! We are to the RIGHT of contact discontinuity
            
          if (dm > dr) then          ! Right shock
          
             dmr = dm/dr
             sr  = ur + cs*sqrt(dmr)
            
             if (s > sr) then        ! This is the right state
     
                d = dr ; u = ur
     
             else                    ! This is the right star state
     
                d = dm ;  u = um 
     
             endif
               
          else                       ! Right rarefaction wave
          
             shr = ur + cs
     
             if (s > shr) then       ! This is the right state
             
                d = dr ; u = ur
             
             else
     
                str = um + cs
               
                if (s < str) then    ! This is the right star state
     
                    u = um ; d = dm
     
                else                 ! This is the right set of solutions
     
                   lnd = (-ur-cs+s)/cs + log(dr)
                   d = exp(lnd) ; u = (-cs + s) 
     
                endif                  
     
             endif
            
          endif
     
       endif
     
       return
     
     end subroutine sample_iso

     !###########################################################
     !###########################################################
     !###########################################################
     !###########################################################
     
     !  Subroutine CAL_GAMMA
     !
     !> Computes the value of the specific heats ratio
     !! \f$\gamma\f$ according to the mass fraction y.
     !<
     subroutine cal_gamma(y)
       
       use gammas
       
       implicit none
       
       real :: y,A
       
       ! y  = mass fraction of 2
       ! mu = mu1/mu2 - 1 : reaction molar increment
       
       A = (1.-y)/(gamma1-1.) + (1.+mu_reac)*y/(gamma2-1.)
       
       gamma  = 5./3. !1. + (1.+mu_reac*y)/A
       
       g1 = (gamma - 1.)/(2.*gamma)
       g2 = (gamma + 1.)/(2.*gamma)
       g3 = 2.0*gamma/(gamma - 1.0)
       g4 = 2.0      /(gamma - 1.0)
       g5 = 2.0      /(gamma + 1.0)
       g6 = (gamma - 1.0)/(gamma + 1.0)
       g7 = (gamma - 1.0)/2.0
       g8 = (gamma - 1.0)
       
       return
  
     end subroutine cal_gamma

end module solvers_hydro
