!> \file
!! Contains modules parameters, var_loc, geom, var, varold,
!! etat_gds, unites, unites_sortie, divers, param_ini, cl,
!! gammas, cellules, const and cputime
!!
!! This file contains all the modules which hold all the
!! global variables.
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module PARAMETERS
!
!> Holds all the general simulation parameters.
!<
module parameters
  implicit none
  integer, dimension(:,:), allocatable :: nx_cpu                !< Number of cells in the x,y,z directions per cpu
  integer, dimension(3)                :: nx_glob               !< Global simulation dimensions
  integer, dimension(3)                :: nx                    !< Local simulation dimensions
  integer, dimension(3)                :: nxmin                 !< Lower limit for arrays
  integer, dimension(3)                :: nxmax                 !< Upper limit for arrays
  integer                              :: ndim                  !< Number of dimensions
  integer                              :: nx_max                !< Highest nx
  integer                              :: nx_glob_max           !< Global highest nx
  integer                              :: Nbuf                  !< Number of buffer/ghost cells
  integer                              :: N_vit                 !< Number of velocity components
  integer                              :: nvar                  !< Number of hydro/MHD variables
  integer                              :: nFx                   !< Number of passive scalars
  integer                              :: nvar_ray              !< Total number of radiative variables
  integer                              :: slope_type            !< Type of slope limiter for Hydro and MHD
  real   , parameter                   :: Pi = 3.1415926535898  !< \f$  \pi \f$
  real   , parameter                   :: quatre_Pi = 4.*pi     !< \f$ 4\pi \f$
  real   , parameter                   :: smallc = 1.d-30       !< Dimensioned small constant
  real   , parameter                   :: smallr = 1.d-40       !< Dimensioned small real number
  character(len=10)                    :: riemann               !< Type of Riemann solver
  character(len=10)                    :: riemann2d             !< Type of Riemann solver 2
end module parameters              

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module VAR_LOC
!
!> Contains the local variables.
!<
module var_loc
  implicit none
  real, dimension(:,:,:,:), allocatable :: rhouloc              !< Local momentum array
  real, dimension(:,:,:,:), allocatable :: uloc                 !< Local velocity array
  real, dimension(:,:,:,:), allocatable :: Fxloc                !< Local passive scalar array
  real, dimension(:,:,:  ), allocatable :: rholoc               !< Local density array
  real, dimension(:,:,:  ), allocatable :: Ploc                 !< Local pressure array
  real, dimension(:,:,:  ), allocatable :: Eloc                 !< Local total gas energy array
  real, dimension(:,:,:  ), allocatable :: Tloc                 !< Local temperature array
  real, dimension(:,:,:  ), allocatable :: Einloc               !< Local internal gas energy array
  real, dimension(:,:,:  ), allocatable :: philoc               !< Local gravitational potential array
  real, dimension(:,:    ), allocatable :: xloc                 !< Local coordinate array
end module var_loc

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module GEOM
!
!> Contains all the grid variables and coordinates.
!<
module geom
  implicit none
  integer, dimension(3      )              :: nshift_gr         !< nshift
  real   , dimension(:,:,:,:), allocatable :: ds                !< Surface element
  real   , dimension(:,:,:  ), allocatable :: dv                !< Volume element
  real   , dimension(:,:    ), allocatable :: x                 !< Local cell interface coordinate array
  real   , dimension(:,:    ), allocatable :: x_glob            !< Global cell interface coordinate array
  real   , dimension(:,:    ), allocatable :: dx                !< Line element
  real   , dimension(:,:    ), allocatable :: dxc               !< Cell centred line element
  real   , dimension(:,:    ), allocatable :: xcloc             !< Local cell centre coordinate array
  real   , dimension(3      )              :: shift_gr          !< shiftgr
  real   , dimension(3      )              :: Lbox              !< Size of computational box
  real   , dimension(3      )              :: box_min           !< Negative coordinate offset
  real   , dimension(3      )              :: box_max           !< Positive coordinate offset
  character(len=5), dimension(3)           :: geom_dir          !< Geometry directions
  character(len=5)                         :: geometrie         !< Grid geometry
  logical                                  :: Cartesien         !< Cartesian grid if .true.
  logical                                  :: Cylindrique       !< Cylindrical grid if .true.
  logical                                  :: Spherique         !< Spherical grid if .true.
end module geom

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module VAR
!
!> Contains all the main variable arrays.
!<
module var
  implicit none
  real, dimension(:,:,:,:  ), allocatable :: rhou               !< Gas momentum array \f$ \rho(u_{x},u_{y},u_{z})\f$
  real, dimension(:,:,:,:  ), allocatable :: B                  !< Magnetic field
  real, dimension(:,:,:,:  ), allocatable :: Fx                 !< Passive scalar array
  real, dimension(:,:,:,:  ), allocatable :: supp1              !< Additional array 1
  real, dimension(:,:,:,:  ), allocatable :: supp2              !< Additional array 2
  real, dimension(:,:,:,:  ), allocatable :: supp3              !< Additional array 3
  real, dimension(:,:,:,:  ), allocatable :: supp4              !< Additional array 4
  real, dimension(:,:,:,:  ), allocatable :: supp5              !< Additional array 5
  real, dimension(:,:,:    ), allocatable :: rho                !< Gas density array \f$ \rho \f$
  real, dimension(:,:,:    ), allocatable :: E                  !< Total gas energy \f$ E = \displaystyle\frac{1}{2} \rho u^{2} + \frac{P}{\gamma-1} \f$
  real, dimension(:,:,:    ), allocatable :: Tgaz               !< Gas temperature array
 !real, dimension(:,:,:    ), allocatable :: P                  !< Gas pressure array
 !real, dimension(:,:,:,:  ), allocatable :: U                  !< Gas velocity array
end module var

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module VAROLD
!
!> Contains all the old (non-updated) variable arrays.
!<
module varold
  implicit none
  real, dimension(:,:,:,:), allocatable :: rhouold              !< Old gas momentum
  real, dimension(:,:,:,:), allocatable :: Bold                 !< Old magnetic field
  real, dimension(:,:,:,:), allocatable :: Fxold                !< Old passive scalar
  real, dimension(:,:,:  ), allocatable :: rhoold               !< Old gas density
  real, dimension(:,:,:  ), allocatable :: Eold                 !< Old total gas energy
  real, dimension(:,:,:  ), allocatable :: Tgazold              !< Old gas temperature
 !real, dimension(:,:,:  ), allocatable :: Pold                 !< Old gas pressure
 !real, dimension(:,:,:,:), allocatable :: Uold                 !< Old gas velocity
end module varold

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module etat_gds
!
!> Contains the interface values arrays.
!<
module Etat_GDS
  implicit none
  real, dimension(:,:,:,:,:), allocatable :: uS                 !< Value of the gas velocity at the cell interface
  real, dimension(:,:,:,:,:), allocatable :: FxS                !< Value of the passive scalar at the cell interface
  real, dimension(:,:,:,:  ), allocatable :: rhoS               !< Value of the gas momentum at the cell interface
  real, dimension(:,:,:,:  ), allocatable :: pS                 !< Value of the gas pressure at the cell interface
  real, dimension(:,:,:,:  ), allocatable :: eS                 !< Value of the total gas energy at the cell interface
end module Etat_GDS

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module OUTPUT
!
!> This routine calls the relevant output routines to
!! write binary or HDF5 files.
!<
module unites
  implicit none
  real :: gramme                                                !< Gram unit
  real :: centimetre                                            !< Centimetre unit
  real :: seconde                                               !< Second unit
  real :: dyne                                                  !< Dyne unit
  real :: erg                                                   !< Erg unit
  real :: degre                                                 !< Temperature unit
  real :: unitm                                                 !< Mass unit
  real :: unitl                                                 !< Length unit
  real :: unitt                                                 !< Time unit
  real :: unitf                                                 !< Force unit
  real :: unite                                                 !< Energy unit
  real :: unitd                                                 !< Temperature unit
  real :: unitmom                                               !< Momentum unit
  real :: cm2                                                   !< \f$\mathrm{cm}^{2}\f$
  real :: cm3                                                   !< \f$\mathrm{cm}^{3}\f$
  real :: ms                                                    !< Millisecond
  real :: kms                                                   !< Kilometer
  real :: an                                                    !< Year
  real :: Mans                                                  !< Million years
  real :: Gans                                                  !< Billion years
  real :: Msol                                                  !< Solar mass \f$ M_{\odot} \f$
  real :: Rsol                                                  !< Solar radius \f$ R_{\odot} \f$
  real :: Lsol                                                  !< Solar luminosity \f$ L_{\odot} \f$
  real :: pc                                                    !< Parsec
  real :: kpc                                                   !< Kiloparsec
  real :: Mpc                                                   !< Megaparsec
  real :: Kelvin                                                !< Kelvin
  real :: c                                                     !< Speed of light \f$ c \f$
  real :: c2                                                    !< Square of the speed of light \f$ c^{2} \f$
  real :: eV                                                    !< Electron volt \f$ eV \f$
  real :: G                                                     !< Gravitational constant \f$ G \f$
  real :: hplanck                                               !< Planck constant \f$ h \f$
  real :: kB                                                    !< Boltzmann constant \f$ k_{B} \f$
  real :: uma                                                   !< Atomic mass unit \f$ u \f$
  real :: H0                                                    !< Hubble's constant \f$ H_{0} \f$
  real :: a_R                                                   !< Stefan-Boltzmann constant \f$ a_{R} \f$
  real :: Pascal                                                !< Pascal
  real :: bar                                                   !< Bar
  real :: kg                                                    !< Kilogram
  real :: metre                                                 !< Metre
  real :: m2                                                    !< \f$ \mathrm{m}^2 \f$
  real :: m3                                                    !< \f$ \mathrm{m}^3 \f$
  real :: joule                                                 !< Joule
  real :: ns                                                    !< Nanosecond
  real :: micron                                                !< Micron
  real :: Gauss                                                 !< Gauss
  real :: mu0                                                   !< Reduced atomic mass
end module unites

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module UNITES_SORTIE
!
!> Contains the units for output file dumping.
!<
module unites_sortie
  implicit none
  real :: u_M                                                   !< Output mass unit 
  real :: u_L                                                   !< Output length unit
  real :: u_T                                                   !< Output time unit
  real :: u_E                                                   !< Output energy unit
  real :: u_Vol                                                 !< Output volume unit
  real :: u_dens                                                !< Output density unit
  real :: u_evol                                                !< Output volumic energy unit
  real :: u_Fr                                                  !< Output force unit
  real :: u_Vit                                                 !< Output velocity unit
  real :: u_Mom                                                 !< Output momentum unit
end module unites_sortie

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module DIVERS
!
!> Contains various variables.
!<
module divers
  implicit none
  integer, dimension(20)            :: fourdim                   !< f
  integer, dimension(5 )            :: suppdim                   !< f
  integer                           :: Nimp                      !< Write 000 output file every Nimp timesteps
  integer                           :: i_restart                 !< File number to restart from
  integer                           :: nsx                       !< x domain subdivision factor 
  integer                           :: nsy                       !< y domain subdivision factor 
  integer                           :: nsz                       !< z domain subdivision factor 
  integer                           :: nrepr                     !< nr
  integer                           :: irepr                     !< ir
  integer                           :: nreprd                    !< nr
  integer                           :: ndata                     !< nd
  integer                           :: nsupp                     !< Number of additional arrays required
  integer                           :: nout                      !< Total number of outputs
  integer                           :: iout                      !< iout
  integer                           :: noutd                     !< no
  integer                           :: noutput                   !< Number of terminal outputs
  real   , dimension(1000)          :: tout                      !< Output time
  real   , dimension(1000)          :: trepr                     !< Restart time
  real                              :: temps                     !< Time
  real                              :: dt                        !< Timestep
  real                              :: dtold                     !< Old timestep
  real                              :: tend                      !< End time
  real                              :: alpha                     !< \f$ \alpha \f$
  real                              :: dt_max                    !< Maximum timestep
  real                              :: dt_glob                   !< Global timestep
  real                              :: dt_mhd                    !< MHD timestep
  real                              :: dt_cfl                    !< CFL limited timestep
  real                              :: dt_ff                     !< Gravitational free-fall limited timestep
  real                              :: tcpu_max                  !< tcpu_max
  real                              :: tcpu_ini                  !< tcpu_ini
  real                              :: tcpu_last                 !< tcpu_last en seconde
  real                              :: mu                        !< \f$ \mu \f$
  character(len=80 ), dimension(20) :: datanames                 !< datanames
  character(len=80 ), dimension(20) :: dataunits                 !< dataunits
  character(len=132), dimension(40) :: terminal_output           !< Strings to hold terminal outputs
  character(len=80 ), dimension(5 ) :: suppnames                 !< Names of additional arrays
  character(len=80 ), dimension(5 ) :: suppunits                 !< Units of additional arrays
  character(len=80 )                :: comments                  !< comments
  character(len=10 )                :: Solver                    !< Type of Hydro/MHD Riemann solver
  character(len=5  )                :: name_of_job               !< Name of job
  character(len=3  )                :: sortie                    !< Output
  logical                           :: reprise                   !< Perform restart from previous run if .true.
  logical                           :: COM                       !< com
  logical                           :: limdx                     !< limdx
  logical                           :: limnite                   !< limnite
  logical                           :: limfr                     !< limfr
  logical                           :: REF                       !< Include cooling if .true.
  logical                           :: DETO                      !< Include detonation if .true.
  logical                           :: CONDUC                    !< Include thermal conduction if .true.
  logical                           :: DIFFU                     !< Include flux limited diffusion if .true.
  logical                           :: RAY                       !< Include radiative transfer if .true.
  logical                           :: HYDRO                     !< Include hydrodynamics if .true.
  logical                           :: GRAV                      !< Include gravity if .true.
  logical                           :: HISTOGRAMME               !< Computes histograms if .true.
  logical                           :: ISOTHERME                 !< Uses isothermal equation of state if .true.
  logical                           :: STIR                      !< Include stirring if .true.
  logical                           :: verbose                   !< Prints out extra information for debugging if .true.
end module divers

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module PARAM_INI
!
!> Contains variables used for initialisation of the
!! simulation.
!<
module param_ini
  implicit none
  real :: rho1                                                  !< \f$ \rho_{1} \f$
  real :: rho2                                                  !< \f$ \rho_{2} \f$
  real :: rho3                                                  !< \f$ \rho_{3} \f$
  real :: rho4                                                  !< \f$ \rho_{4} \f$
  real :: p1                                                    !< \f$ p_{1} \f$
  real :: p2                                                    !< \f$ p_{2} \f$
  real :: p3                                                    !< \f$ p_{3} \f$
  real :: p4                                                    !< \f$ p_{4} \f$
  real :: T1                                                    !< \f$ T_{1} \f$
  real :: T2                                                    !< \f$ T_{2} \f$
  real :: T3                                                    !< \f$ T_{3} \f$
  real :: T4                                                    !< \f$ T_{4} \f$
  real :: E1                                                    !< \f$ E_{1} \f$
  real :: E2                                                    !< \f$ E_{2} \f$
  real :: E3                                                    !< \f$ E_{3} \f$
  real :: E4                                                    !< \f$ E_{4} \f$
  real :: u1                                                    !< \f$ u_{1} \f$
  real :: u2                                                    !< \f$ u_{2} \f$
  real :: u3                                                    !< \f$ u_{3} \f$
  real :: u4                                                    !< \f$ u_{4} \f$
  real :: v1                                                    !< \f$ v_{1} \f$
  real :: v2                                                    !< \f$ v_{2} \f$
  real :: v3                                                    !< \f$ v_{3} \f$
  real :: v4                                                    !< \f$ v_{4} \f$
  real :: cs1                                                   !< \f$ cs_{1} \f$
  real :: x1                                                    !< \f$ x_{1} \f$
  real :: x2                                                    !< \f$ x_{2} \f$
  real :: d1                                                    !< \f$ d_{1} \f$
  real :: d2                                                    !< \f$ d_{1} \f$
  real :: ww                                                    !< ?
  real :: B0                                                    !< \f$ B_{0} \f$
  real :: Temperature_isotherme                                 !< Isothermal temperature
end module param_ini

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module CL
!
!> Contains the variables used to define the boundary
!! conditions.
!<
module cl
  implicit none
  integer, dimension(6  )              :: icl_glob              !< Global boundaries types for faces 1 to 6
  integer, dimension(6  )              :: icl                   !< Local  boundaries types for faces 1 to 6
  integer, dimension(6  )              :: icl0                  !< icl
  real   , dimension(:,:), allocatable :: rhou_l                !< Imposed value for the gas momentum    inside the ghost cells
  real   , dimension(:,:), allocatable :: u_l                   !< Imposed value for the gas velocity    inside the ghost cells
  real   , dimension(:  ), allocatable :: rho_l                 !< Imposed value for the gas density     inside the ghost cells
  real   , dimension(:  ), allocatable :: E_l                   !< Imposed value for the gas energy      inside the ghost cells
  real   , dimension(:  ), allocatable :: P_l                   !< Imposed value for the gas pressure    inside the ghost cells
  real   , dimension(:  ), allocatable :: T_l                   !< Imposed value for the gas temperature inside the ghost cells
  real   , dimension(:,:), allocatable :: fx_l                  !< Imposed value for the passive scalar  inside the ghost cells
  real   , dimension(:,:), allocatable :: B_l                   !< Imposed value for the magnetic field  inside the ghost cells
end module cl

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module GAMMAS
!
!> Contains the variables determining the heat capacity
!! ratio and thermodynamic relations.
!<
module gammas
  implicit none
  real :: mu_reac                                               !< \f$ \mu_{reac} \f$
  real :: gamma                                                 !< Ratio of specific heats \f$ \gamma \f$
  real :: gamma1                                                !< \f$ \gamma_{1} \f$
  real :: gamma2                                                !< \f$ \gamma_{1} \f$
  real :: g1                                                    !< \f$ \displaystyle \frac{\gamma-1}{2\gamma} \f$
  real :: g2                                                    !< \f$ \displaystyle \frac{\gamma+1}{2\gamma} \f$
  real :: g3                                                    !< \f$ \displaystyle \frac{2\gamma}{\gamma-1} \f$
  real :: g4                                                    !< \f$ \displaystyle \frac{2}{\gamma-1} \f$
  real :: g5                                                    !< \f$ \displaystyle \frac{2}{\gamma+1} \f$
  real :: g6                                                    !< \f$ \displaystyle \frac{\gamma-1}{\gamma+1} \f$
  real :: g7                                                    !< \f$ \displaystyle \frac{\gamma-1}{2} \f$
  real :: g8                                                    !< \f$ \gamma-1 \f$
  real :: Cs_iso                                                !< Isothermal sound speed
  real :: Cs_iso2                                               !< Isothermal sound speed 2
end module gammas

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module CELLULES
!
!> Contains ?
!<
module cellules
  implicit none
  integer, dimension(:,:  ), allocatable :: Imax                !< Imax
  integer, dimension(:,:  ), allocatable :: Imax2               !< Imax2
  integer, dimension(:    ), allocatable :: fff_t               !< fff_t
  integer                                :: N_cell              !< N_cell
  integer                                :: fff                 !< fff
  integer                                :: ioffset             !< ioffset
  integer                                :: ifich               !< ifich
  integer                                :: size_offset         !< size_offset
  integer                                :: ifilm = 0           !< ifilm
  real   , dimension(:,:,:), allocatable :: Pcell               !< Pcell
  real   , dimension(:,:  ), allocatable :: Pmax                !< Pmax
  real   , dimension(:,:  ), allocatable :: Pmax2               !< Pmax2
  real                                   :: dx_cell             !< dx_cell
  real                                   :: t_film              !< t_film
  real                                   :: dt_film             !< dt_film
end module cellules

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module CONST
!
!> Contains useful constants.
!<
module const
  implicit none
  real, parameter :: bigreal = 1.0e+30                          !< A big real number
  real, parameter :: zero    = 0.0                              !< 0
  real, parameter :: one     = 1.0                              !< 1
  real, parameter :: two     = 2.0                              !< 2
  real, parameter :: three   = 3.0                              !< 3
  real, parameter :: four    = 4.0                              !< 4
  real, parameter :: two3rd  = 2.0/3.0                          !< 2/3
  real, parameter :: half    = 0.5                              !< 1/2
  real, parameter :: third   = 1.0/3.0                          !< 1/3
  real, parameter :: forth   = 0.25                             !< 1/4
  real, parameter :: sixth   = 1.0/6.0                          !< 1/6
end module const

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module CPUTIME
!
!> Contains variables used to compute the cpu time spent
!! is each part of the code.
!<
module cputime
  implicit none
  real :: temps1                                                !< Time 1
  real :: temps_ref                                             !< Time spent computing cooling
  real :: temps_conduc                                          !< Time spent computing thermal conduction
  real :: temps_ray                                             !< Time spent computing radiative transfer
  real :: temps_hydro                                           !< Time spent computing hydrodynamics
  real :: temps_io                                              !< Time spent performing outputs
  real :: temps_comm                                            !< Time spent performing communications
  real :: temps_dif                                             !< Time spent computing flux limited diffusion
  real :: temps_grav                                            !< Time spent computing gravity
end module cputime
