!> \file
!! contains subroutines upwind(), lax_friedrich(), hll(),
!! hlld(), find_mhd_flux(), find_mhd_flux2(),
!! find_speed_info(), find_speed_fast(),
!! find_speed_alfven(), hydro_acoustic(), athena_roe(),
!! eigenvalues() and eigen_cons()
!!
!! This file contains all the subroutines used to compute
!! mhd fluxes.
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine UPWIND
!
!> 1D upwind Riemann solver.
!<
subroutine upwind(qleft,qright,fgdnv,zero_flux)

  use parameters
  use const
  
  implicit none  

  real, dimension(1:nvar) :: qleft,qright,fgdnv
  real, dimension(1:nvar) :: fleft,fright,fmean
  real, dimension(1:nvar) :: uleft,uright,udiff
  real                    :: vleft,bx_mean,zero_flux
  
  ! enforce continuity of normal component
  bx_mean=half*(qleft(4)+qright(4))
  qleft (4)=bx_mean
  qright(4)=bx_mean

  call find_mhd_flux(qleft ,uleft ,fleft )
  call find_mhd_flux(qright,uright,fright)
  
  ! find the mean flux
  fmean =  half * ( fright + fleft ) * zero_flux
  
  ! find the mean normal velocity
  vleft = half * ( qleft(3) + qright(3) )
  
  ! difference between the 2 states
  udiff = half * ( uright - uleft )
  
  ! the upwind flux
  fgdnv = fmean - abs(vleft) * udiff
  
  return

end subroutine upwind

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine LAX_FRIEDRICH
!
!> Lax -Friedrich Riemann solver.
!<
subroutine lax_friedrich(qleft,qright,fmean,zero_flux)

  use parameters
  use const

  implicit none

  real, dimension(1:nvar) :: qleft,qright
  real, dimension(1:nvar) :: fleft,fright,fmean,uleft,uright,udiff
  real                    :: vel_info,vel_left,vel_right,zero_flux

  !take mean value of the transverse field
  qleft (4)=half*(qleft(4)+qright(4))
  qright(4)=qleft(4)
  
  call find_mhd_flux(qleft ,uleft ,fleft )
  call find_mhd_flux(qright,uright,fright)
  
  !find the mean flux
  fmean = half*(fright+fleft)*zero_flux

  ! find the largest eigenvalue in the normal direction to the interface
  call find_speed_info(qleft ,vel_left )
  call find_speed_info(qright,vel_right)

  vel_info=max(vel_left,vel_right)
              
  !difference between the 2 states
  udiff = half*(uright-uleft)
  
  !the local lax-friedrich flux
  fmean = fmean - vel_info*udiff
  
  return

end subroutine lax_friedrich

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine HLL
!
!> HLL Riemann solver.
!<
subroutine hll(qleft,qright,fgdnv)

  use parameters
  use const

  implicit none  

  real, dimension(1:nvar) :: qleft,qright,fgdnv
  real, dimension(1:nvar) :: fleft,fright
  real, dimension(1:nvar) :: uleft,uright
  real                    :: vleft,vright,bx_mean,cfleft,cfright,sl,sr
  
  ! enforce continuity of normal component
  bx_mean=half*(qleft(4)+qright(4))
  qleft (4)=bx_mean
  qright(4)=bx_mean

  call find_mhd_flux(qleft ,uleft ,fleft )
  call find_mhd_flux(qright,uright,fright)
  
  ! find the largest eigenvalue in the normal direction to the interface
  call find_speed_fast(qleft ,cfleft ,nvar)
  call find_speed_fast(qright,cfright,nvar)
  vleft =qleft (3)
  vright=qright(3)
  sl=min(min(vleft,vright)-max(cfleft,cfright),zero)
  sr=max(max(vleft,vright)+max(cfleft,cfright),zero)

  ! the hll flux
  fgdnv = (sr*fleft-sl*fright+sr*sl*(uright-uleft))/(sr-sl)
  
  return

end subroutine hll

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine HLLD
!
!> HLLD Riemann solver.
!! (Miyoshi & Kusano, 2005, JCP, 208, 315)
!<
subroutine hlld(qleft,qright,fgdnv,ro,uo,vo,wo,bo,co,ptoto,nvar)

  use gammas
  use const
  
  implicit none
 
  integer :: nvar,ivar
  real, dimension(1:nvar) :: qleft,qright,fgdnv
  real :: sl,sr,sal,sar
  real :: entho,a,sgnm
  real :: rl,pl,ul,vl,wl,cl,ecinl,emagl,etotl,ptotl,vdotbl,bl,el
  real :: rr,pr,ur,vr,wr,cr,ecinr,emagr,etotr,ptotr,vdotbr,br,er
  real :: cfastl,calfvenl,rcl,rstarl,vstarl,wstarl,bstarl,cstarl,vdotbstarl
  real :: cfastr,calfvenr,rcr,rstarr,vstarr,wstarr,bstarr,cstarr,vdotbstarr
  real :: sqrrstarl,etotstarl,etotstarstarl
  real :: sqrrstarr,etotstarr,etotstarstarr
  real :: ustar,ptotstar,estar,vstarstar,wstarstar,bstarstar,cstarstar,vdotbstarstar
  real :: ro,uo,vo,wo,bo,co,ptoto,etoto,vdotbo

  entho = one/(gamma-one)

  ! enforce continuity of normal component
  a=half*(qleft(4)+qright(4))
  sgnm=sign(one,a)
  qleft(4)=a; qright(4)=a
  
#ifdef ISO
  qleft (2)=qleft (1)*cs_iso**2
  qright(2)=qright(1)*cs_iso**2
#endif

  ! left variables
  rl=qleft(1); pl=qleft(2); ul=qleft(3)
  vl=qleft(5); bl=qleft(6); wl=qleft(7); cl=qleft(8)
  ecinl = half*(ul*ul+vl*vl+wl*wl)*rl
  emagl = half*(a*a+bl*bl+cl*cl)
  etotl = pl*entho+ecinl+emagl
  ptotl = pl + emagl
  vdotbl= ul*a+vl*bl+wl*cl

  ! right variables
  rr=qright(1); pr=qright(2); ur=qright(3)
  vr=qright(5); br=qright(6); wr=qright(7); cr=qright(8)
  ecinr = half*(ur*ur+vr*vr+wr*wr)*rr
  emagr = half*(a*a+br*br+cr*cr)
  etotr = pr*entho+ecinr+emagr
  ptotr = pr + emagr
  vdotbr= ur*a+vr*br+wr*cr

  ! find the largest eigenvalues in the normal direction to the interface
  call find_speed_fast(qleft ,cfastl,nvar)
  call find_speed_fast(qright,cfastr,nvar)

  ! compute hll wave speed
  sl=min(ul,ur)-max(cfastl,cfastr)
  sr=max(ul,ur)+max(cfastl,cfastr)
  
  ! compute lagrangian sound speed
  rcl=rl*(ul-sl)
  rcr=rr*(sr-ur)

  ! compute acoustic star state
  ustar   =(rcr*ur   +rcl*ul   +  (ptotl-ptotr))/(rcr+rcl)
  ptotstar=(rcr*ptotl+rcl*ptotr+rcl*rcr*(ul-ur))/(rcr+rcl)

  ! left star region variables
  rstarl=rl*(sl-ul)/(sl-ustar)
  estar =rl*(sl-ul)*(sl-ustar)-a**2
  el    =rl*(sl-ul)*(sl-ul   )-a**2
  if(estar==0)then
     vstarl=vl
     bstarl=bl
     wstarl=wl
     cstarl=cl
  else
     vstarl=vl-a*bl*(ustar-ul)/estar
     bstarl=bl*el/estar
     wstarl=wl-a*cl*(ustar-ul)/estar
     cstarl=cl*el/estar
  endif
  vdotbstarl=ustar*a+vstarl*bstarl+wstarl*cstarl
  etotstarl=((sl-ul)*etotl-ptotl*ul+ptotstar*ustar+a*(vdotbl-vdotbstarl))/(sl-ustar)
  sqrrstarl=sqrt(rstarl)
  calfvenl=abs(a)/sqrrstarl
  sal=ustar-calfvenl

  ! right star region variables
  rstarr=rr*(sr-ur)/(sr-ustar)
  estar =rr*(sr-ur)*(sr-ustar)-a**2
  er    =rr*(sr-ur)*(sr-ur   )-a**2
  if(estar==0)then
     vstarr=vr
     bstarr=br
     wstarr=wr
     cstarr=cr
  else
     vstarr=vr-a*br*(ustar-ur)/estar
     bstarr=br*er/estar
     wstarr=wr-a*cr*(ustar-ur)/estar
     cstarr=cr*er/estar
  endif
  vdotbstarr=ustar*a+vstarr*bstarr+wstarr*cstarr
  etotstarr=((sr-ur)*etotr-ptotr*ur+ptotstar*ustar+a*(vdotbr-vdotbstarr))/(sr-ustar)
  sqrrstarr=sqrt(rstarr)
  calfvenr=abs(a)/sqrrstarr
  sar=ustar+calfvenr

  ! double star region variables
  vstarstar=(sqrrstarl*vstarl+sqrrstarr*vstarr+sgnm*(bstarr-bstarl))/(sqrrstarl+sqrrstarr)
  wstarstar=(sqrrstarl*wstarl+sqrrstarr*wstarr+sgnm*(cstarr-cstarl))/(sqrrstarl+sqrrstarr)
  bstarstar=(sqrrstarl*bstarr+sqrrstarr*bstarl+sgnm*sqrrstarl*sqrrstarr*(vstarr-vstarl))/(sqrrstarl+sqrrstarr)
  cstarstar=(sqrrstarl*cstarr+sqrrstarr*cstarl+sgnm*sqrrstarl*sqrrstarr*(wstarr-wstarl))/(sqrrstarl+sqrrstarr)
  vdotbstarstar=ustar*a+vstarstar*bstarstar+wstarstar*cstarstar
  etotstarstarl=etotstarl-sgnm*sqrrstarl*(vdotbstarl-vdotbstarstar)
  etotstarstarr=etotstarr+sgnm*sqrrstarr*(vdotbstarr-vdotbstarstar)
  
  ! sample the solution at x/t=0
  if(sl>0d0)then
     ro=rl
     uo=ul
     vo=vl
     wo=wl
     bo=bl
     co=cl
     ptoto=ptotl
     etoto=etotl
     vdotbo=vdotbl
  else if(sal>0d0)then
     ro=rstarl
     uo=ustar
     vo=vstarl
     wo=wstarl
     bo=bstarl
     co=cstarl
     ptoto=ptotstar
     etoto=etotstarl
     vdotbo=vdotbstarl
  else if(ustar>0d0)then
     ro=rstarl
     uo=ustar
     vo=vstarstar
     wo=wstarstar
     bo=bstarstar
     co=cstarstar
     ptoto=ptotstar
     etoto=etotstarstarl
     vdotbo=vdotbstarstar
  else if(sar>0d0)then
     ro=rstarr
     uo=ustar
     vo=vstarstar
     wo=wstarstar
     bo=bstarstar
     co=cstarstar
     ptoto=ptotstar
     etoto=etotstarstarr
     vdotbo=vdotbstarstar
  else if (sr>0d0)then
     ro=rstarr
     uo=ustar
     vo=vstarr
     wo=wstarr
     bo=bstarr
     co=cstarr
     ptoto=ptotstar
     etoto=etotstarr
     vdotbo=vdotbstarr
  else
     ro=rr
     uo=ur
     vo=vr
     wo=wr
     bo=br
     co=cr
     ptoto=ptotr
     etoto=etotr
     vdotbo=vdotbr
  end if

  ! compute the godunov flux
  fgdnv(1) = ro*uo
  fgdnv(2) = (etoto+ptoto)*uo-a*vdotbo
  fgdnv(3) = ro*uo*uo-a*a
  fgdnv(4) = zero
  fgdnv(5) = ro*uo*vo-a*bo
  fgdnv(6) = bo*uo-a*vo
  fgdnv(7) = ro*uo*wo-a*co
  fgdnv(8) = co*uo-a*wo
  do ivar = 9,nvar
     if(fgdnv(1)>0)then
        fgdnv(ivar) = fgdnv(1)*qleft (ivar)
     else
        fgdnv(ivar) = fgdnv(1)*qright(ivar)
     endif
  end do
  
  return

end subroutine hlld

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine FIND_MHD_FLUX
!
!> Compute the 1d mhd fluxes from the conservative
!! variables. The structure of qvar is : rho, pressure,
!! vnormal, bnormal, vtransverse1, btransverse1,
!! vtransverse2, btransverse2.
!<
subroutine find_mhd_flux(qvar,cvar,ff)

  use parameters
  use gammas
  use const
  
  implicit none
   
  integer                 :: ivar
  real, dimension(1:nvar) :: qvar,cvar,ff
  real                    :: ecin,emag,etot,d,u,v,w,a,b,c,p,ptot,entho

#ifdef ISO
  ! recompute pressure
  qvar(2) = qvar(1)*cs_iso**2
#endif

  ! local variables
  entho = one/(gamma-one)
  d=qvar(1); p=qvar(2); u=qvar(3); a=qvar(4)
  v=qvar(5); b=qvar(6); w=qvar(7); c=qvar(8)
  ecin = half*(u*u+v*v+w*w)*d
  emag = half*(a*a+b*b+c*c)
  etot = p*entho+ecin+emag
  ptot = p + emag
  
  ! compute conservative variables
  cvar(1) = d
  cvar(2) = etot
  cvar(3) = d*u
  cvar(4) = a
  cvar(5) = d*v
  cvar(6) = b
  cvar(7) = d*w
  cvar(8) = c
  do ivar = 9,nvar
     cvar(ivar) = d*qvar(ivar)
  end do

  ! compute fluxes
  ff(1) = d*u
  ff(2) = (etot+ptot)*u-a*(a*u+b*v+c*w)
  ff(3) = d*u*u+ptot-a*a
  ff(4) = zero
  ff(5) = d*u*v-a*b
  ff(6) = b*u-a*v
  ff(7) = d*u*w-a*c
  ff(8) = c*u-a*w
  do ivar = 9,nvar
     ff(ivar) = d*u*qvar(ivar)
  end do
  
  return

end subroutine find_mhd_flux

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine FIND_MHD_FLUX2
!
!> Compute the 1d mhd fluxes from the conservative
!! variables.
!! qvar: rho,p,vn,bn,vt,bt1,vt2,bt2
!<
subroutine find_mhd_flux2(qvar,cvar,ff)

  use parameters
  use gammas
  use const

  implicit none
   
  integer                 :: i,iv,ib
  real, dimension(1:nvar) :: qvar,ff,cvar 
  real                    :: ecin,emag,magp,vb,entho

  entho = one / (gamma-one)

#ifdef ISO
  ! recompute pressure
  qvar(2) = qvar(1)*cs_iso**2
#endif

  ! density & density flux
  cvar(1) = qvar(1)
  ff(1) = qvar(1)*qvar(3)   !rho*vn

  ecin = qvar(3)**2                               !vn^2
  emag = half*qvar(4)**2                          !bn^2/2
  magp = 0.
  vb   = qvar(3)*qvar(4)                          !vn*bn  

  do i=1,n_vit -1 
   iv = 3 + 2*i
    ib = 4 + 2*i
    ecin = ecin + qvar(iv)**2                     !vt^2
    magp = magp + qvar(ib)**2                     !bt^2
    vb   = vb   + qvar(iv) * qvar(ib)             !vt*bt
  enddo

  ecin = half * qvar(1) * ecin                    !rho*v^2 / 2.
  emag = emag + half*magp                         !b^2/(8 pi)
  magp = half*magp                                !bperp^2/(8 pi) 

  !normal momentum & normal momentum flux
  cvar(3) = qvar(1)*qvar(3)
  ff(3) = qvar(1)*(qvar(3))**2 + & !(rho*vn^2)
&                       qvar(2) + &               !p
&                       magp -   half*qvar(4)**2  !-bx**2/2
  

  !transverse momemtum and transverse momemtum flux
  do i=1,n_vit-1
    iv = 3 + 2*i
    ib = 4 + 2*i
    cvar(iv) = qvar(1)*qvar(iv)
    ff(iv) = qvar(1)*qvar(3)*qvar(iv)  -  &        !rho*vn*vt : flux of transverse velocity
&                           qvar(4)*qvar(ib)       !bn*bt     : magnetic tension
  enddo

  !energy & energy flux (eint + p = gamma/(gamma-1)*p)
  cvar(2) =  entho*qvar(2) + ecin + emag
  ff(2) = (qvar(2) + cvar(2) + emag )*qvar(3) - &  !(eint+p+ecin+emag)*vn
&                vb * qvar(4)                      !(v*b) * bn 

  !passive scalars
  do i=n_vit*2+3,nvar
     cvar(i) = qvar(i)
     ff(i) = qvar(i)*qvar(3)
  enddo

  ! normal magnetic field; normal magnetic flux is exactly 0
  cvar(4) = qvar(4) 
  ff(4) = 0.

  !transverse magnetic field & transverse magnetic flux
  do i=1,n_vit-1
   iv = 3 + 2*i
   ib = 4 + 2*i
   cvar(ib) = qvar(ib)
   ff(ib) = qvar(ib)*qvar(3) - qvar(4)*qvar(iv)  ! bt*vx - bx*vt 
  enddo

  return

end subroutine find_mhd_flux2

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine FIND_SPEED_INFO
!
!> Calculate the fastest velocity at which information is
!! exchanged at the interface.
!! The structure of qvar is : rho, pressure, vnormal,
!! bnormal, vtransverse1, btransverse1, vtransverse2,
!! btransverse2
!<
subroutine find_speed_info(qvar,vel_info)

  use parameters
  use gammas    , only : gamma
  use const
  
  implicit none
   
  real, dimension(1:nvar) :: qvar  
  real                    :: vel_info
  real                    :: d,p,u,v,w,a,b,c,b2,c2,d2,cf

  d=qvar(1); p=qvar(2); u=qvar(3); a=qvar(4)
  v=qvar(5); b=qvar(6); w=qvar(7); c=qvar(8)
  b2 = a*a+b*b+c*c
  c2 = gamma*p/d
  d2 = half*(b2/d+c2)
  cf = sqrt( d2 + sqrt(d2**2-c2*a*a/d) )
  vel_info = cf+abs(u)
  
  return

end subroutine find_speed_info

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine FIND_SPEED_FAST
!
!> Calculate the fast magnetosonic velocity.
!! The structure of qvar is : rho, pressure, vnormal,
!! bnormal, vtransverse1, btransverse1, vtransverse2,
!! btransverse2
!<
subroutine find_speed_fast(qvar,vel_info,nvar)
  
  use gammas, only: gamma
  use const
  
  implicit none

  integer                 :: nvar
  real, dimension(1:nvar) :: qvar  
  real                    :: vel_info
  real                    :: d,p,u,v,w,a,b,c,b2,c2,d2,cf

  d=qvar(1); p=qvar(2); u=qvar(3); a=qvar(4)
  v=qvar(5); b=qvar(6); w=qvar(7); c=qvar(8)
  b2 = a*a+b*b+c*c
  c2 = gamma*p/d
  d2 = half*(b2/d+c2)
  cf = sqrt( d2 + sqrt(d2**2-c2*a*a/d) )
  vel_info = cf
  
  return

end subroutine find_speed_fast

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine FIND_SPEED_ALFVEN
!
!> Calculate the Alfven velocity.
!! The structure of qvar is : rho, pressure, vnormal,
!! bnormal, vtransverse1, btransverse1, vtransverse2,
!! btransverse2
!<
subroutine find_speed_alfven(qvar,vel_info,nvar)
  
  use const
  
  implicit none
   
  integer                 :: nvar
  real, dimension(1:nvar) :: qvar  
  real                    :: vel_info
  real                    :: d,a

  d=qvar(1); a=qvar(4)
  vel_info = sqrt(a*a/d)
  
  return

end subroutine find_speed_alfven

!###########################################################
!###########################################################
!###########################################################
!########################################################### 

!  Subroutine HYDRO_ACOUSTIC
!
!> Acoustic Riemann solver.
!<
subroutine hydro_acoustic(qleft,qright,fgdnv)

  use parameters
  use gammas    , only : gamma
  use const

  implicit none

  ! dummy arguments
  real, dimension(1:nvar) :: qleft,qright,fgdnv,qgdnv,ugdnv

  ! local variables
  integer :: n
  real    :: smallp, bx_mean
  real    :: rl   ,ul   ,pl   ,cl
  real    :: rr   ,ur   ,pr   ,cr   
  real    :: ro   ,uo   ,po   ,co   
  real    :: rstar,ustar,pstar,cstar
  real    :: wl   ,wr   ,wo   
  real    :: sgnm ,spin ,spout,ushock
  real    :: frac

  ! constants
  smallp = smallr*smallc**2

  ! enforce continuity of normal component
  bx_mean=half*(qleft(4)+qright(4))
  qleft (4)=bx_mean
  qright(4)=bx_mean

  ! initial states pressure, density and velocity
  rl=max(qleft (1),smallr)
  rr=max(qright(1),smallr)

  pl=max(qleft (2),smallp)
  pr=max(qright(2),smallp)

  ul = qleft (3)
  ur = qright(3)

  ! acoustic star state
  cl = sqrt(gamma*pl/rl)
  cr = sqrt(gamma*pr/rr)
  wl = cl*rl
  wr = cr*rr
  pstar = ((wr*pl+wl*pr)+wl*wr*(ul-ur))/(wl+wr)
  ustar = ((wr*ur+wl*ul)+(pl-pr))/(wl+wr)

  ! left going or right going contact wave
  sgnm = sign(one,ustar)

  ! left or right unperturbed state
  if(sgnm==one)then
     ro = rl
     uo = ul
     po = pl
     wo = wl
     co = cl
  else
     ro = rr
     uo = ur
     po = pr
     wo = wr
     co = cr
  end if

  ! star region density and sound speed
  rstar = ro+(pstar-po)/co**2
  rstar = max(rstar,smallr)
  cstar = sqrt(abs(gamma*pstar/rstar))
  cstar = max(cstar,smallc)

  ! head and tail speed of rarefaction
  spout = co   -sgnm*uo   
  spin  = cstar-sgnm*ustar

  ! shock speed
  ushock = half*(spin+spout)
  ushock = max(ushock,-sgnm*ustar)
  if(pstar>=po)then
     spout=ushock
     spin =spout 
  end if

  ! sample the solution at x/t=0
  if(spout<zero)then      ! initial state
     qgdnv(1) = ro
     qgdnv(2) = po
     qgdnv(3) = uo
  else if(spin>=zero)then  ! star region
     qgdnv(1) = rstar
     qgdnv(2) = pstar
     qgdnv(3) = ustar
  else                        ! rarefaction
     frac = spout/(spout-spin)
     qgdnv(1) = frac*rstar + (one - frac)*ro
     qgdnv(2) = frac*pstar + (one - frac)*po
     qgdnv(3) = frac*ustar + (one - frac)*uo
  end if

  ! passive scalars
  do n = 4,nvar
     if(sgnm==one)then
        qgdnv(n) = qleft (n)
     else
        qgdnv(n) = qright(n)
     end if
  end do

  call find_mhd_flux(qgdnv,ugdnv,fgdnv)
  
  return

end subroutine hydro_acoustic

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine ATHENA_ROE
!
!> Athena Roe Riemann solver.
!<
subroutine athena_roe(qleft,qright,fmean,zero_flux)

  use parameters
  use const
  
  implicit none
  
  integer :: n

  real, dimension(1:nvar       ) :: qleft,qright
  real, dimension(1:nvar       ) :: fleft,fright,fmean,uleft,uright,udiff
  real, dimension(nvar-1,nvar-1) :: lem,rem
  real, dimension(nvar-1       ) :: lambda,lambdal,lambdar,a

  real :: droe,vxroe,vyroe,vzroe,sqrtdl,sqrtdr
  real :: hroe,byroe,bzroe,xfactor,yfactor
  real :: pbr,pbl
  real :: fluxd,fluxmx,fluxmy,fluxmz,fluxe
  real :: fluxby,fluxbz,byl,byr,bzl,bzr,bx
  real :: dl,vxl,vyl,vzl,pl,l1,l2
  real :: dr,vxr,vyr,vzr,pr
  real :: mxl,myl,mzl,el
  real :: mxr,myr,mzr,er
  real :: coef,vright,vleft
  real :: zero_flux,dim,eim,mxm,mym,mzm,bym,bzm,etm

  logical :: llf

  !take mean value of the transverse field
  qleft (4)=0.5d0*(qleft(4)+qright(4))
  qright(4)=qleft(4)

  ! compute the fluxes and the conserved variables
  !remember the convention : rho, e, rhovx, bx, rhovy, by, vz, rhobz
  call find_mhd_flux(qleft ,uleft ,fleft )
  call find_mhd_flux(qright,uright,fright)

  !define the primitive quantities explicitly 
  dl  = qleft (1)
  dr  = qright(1)

  pl  = qleft (2)
  pr  = qright(2)

  vxl = qleft (3)
  vxr = qright(3)

  vyl = qleft (5)
  vyr = qright(5)

  byl = qleft (6)
  byr = qright(6)

  vzl = qleft (7)
  vzr = qright(7)

  bzl = qleft (8)
  bzr = qright(8)

  !finally attribute bx value (left and right are identical)
  bx  = 0.5d0*(qleft(4)+qright(4))

  !define explicitly the conserved quantities
  el  = uleft (2)
  er  = uright(2)

  mxl = uleft (3)
  mxr = uright(3)

  myl = uleft (5)
  myr = uright(5)


  mzl = uleft (7)
  mzr = uright(7)

  !the magnetic pressure
  pbl = half * (bx*bx + byl*byl + bzl*bzl)
  pbr = half * (bx*bx + byr*byr + bzr*bzr)

! step 2 : compute roe-averaged data from left and right states
!   these averages will be the input variables to the eigen problem

  sqrtdl  = sqrt(dl)
  sqrtdr  = sqrt(dr)
  droe    = sqrtdl*sqrtdr
  vxroe   = (sqrtdl*vxl + sqrtdr*vxr)/(sqrtdl+sqrtdr)
  vyroe   = (sqrtdl*vyl + sqrtdr*vyr)/(sqrtdl+sqrtdr)
  vzroe   = (sqrtdl*vzl + sqrtdr*vzr)/(sqrtdl+sqrtdr)
  byroe   = (sqrtdr*byl + sqrtdl*byr)/(sqrtdl+sqrtdr)
  bzroe   = (sqrtdr*bzl + sqrtdl*bzr)/(sqrtdl+sqrtdr)
  hroe    = ((el+pl+pbl)/sqrtdl+(er+pr+pbr)/sqrtdr)/(sqrtdl+sqrtdr)

  xfactor  = ((byroe*byroe-byl*byr)+(bzroe*bzroe-bzl*bzr))/(2*droe)
  yfactor  = (dl+dr)/(2*droe)
!
! step 3 : compute eigenvalues and eigenmatrices from roe-averaged values
!

  call eigen_cons(droe,vxroe,vyroe,vzroe,hroe,&
&          bx,byroe,bzroe,xfactor,yfactor,lambda,rem,lem)

  !
  !  compute eigenvalues from left and right states
  !  
  call eigenvalues(dl,vxl,vyl,vzl,pl,bx,byl,bzl,lambdal)
  call eigenvalues(dr,vxr,vyr,vzr,pr,bx,byr,bzr,lambdar)

!
! step 4 : create intermediate states from eigenmatrices
! remember that the convention of athena's eigenvalues is : rho,rhovx,rhovy,rhovz,e,by,bz

  a  = 0
  a  = a  + (dr -dl ) * lem(1,:)
  a  = a  + (mxr-mxl) * lem(2,:)
  a  = a  + (myr-myl) * lem(3,:)
  a  = a  + (mzr-mzl) * lem(4,:)
  a  = a  + (er - el) * lem(5,:)
  a  = a  + (byr-byl) * lem(6,:)
  a  = a  + (bzr-bzl) * lem(7,:)

  llf=.false.
  dim = dl 
  mxm = mxl
  mym = myl
  mzm = mzl
  eim = el 
  bym = byl
  bzm = bzl
  do n = 1, nvar-1
     dim = dim + a(n)*rem(n,1)
     mxm = mxm + a(n)*rem(n,2)
     mym = mym + a(n)*rem(n,3)
     mzm = mzm + a(n)*rem(n,4)
     eim = eim + a(n)*rem(n,5)
     bym = bym + a(n)*rem(n,6)
     bzm = bzm + a(n)*rem(n,7)
     etm = eim-0.5*(mxm*mxm+mym*mym+mzm*mzm)/dim-0.5*(bx*bx+bym*bym+bzm*bzm)
     if(dim.le.zero.or.etm.le.zero)llf=.true.
      if (llf) write(*,*) 'problem in roe density or pressure is < 0', dim, etm ,dl, pl
  end do

  if( llf ) then
     fmean = half * ( fright + fleft ) * zero_flux
     call find_speed_info(qleft ,vleft )
     call find_speed_info(qright,vright)
     udiff = half * ( uright - uleft )
     fmean = fmean - max(vleft,vright) * udiff
     return
  endif

  !
  ! step 6 : entropy fix for genuinely non linear waves
  !
  do n = 1, nvar-1, 2
     l1 = min(lambdal(n),lambda(n))
     l2 = max(lambdar(n),lambda(n))
     if(l1.lt.zero.and.l2.gt.zero)then
        lambda(n)=(lambda(n)*(l2+l1)-two*l2*l1)/(l2-l1)
     endif
  end do

!
! step 6 : compute fluxes at interface using  roe  solver
!
  !add the left and right fluxes
  !remember the convention : rho, e, rhovx, bx, rhovy, by, vz, rhobz
  fleft  = fleft  * zero_flux
  fright = fright * zero_flux
  fluxd  = fleft(1) + fright(1)
  fluxe  = fleft(2) + fright(2)
  fluxmx = fleft(3) + fright(3)
  fluxmy = fleft(5) + fright(5)
  fluxby = fleft(6) + fright(6)
  fluxmz = fleft(7) + fright(7)
  fluxbz = fleft(8) + fright(8)

!now compute the roe fluxes
! remember that the convention of athena's eigenvalues is : rho,rhovx,rhovy,rhovz,e,by,bz
  do n=1,nvar-1
      coef = abs(lambda(n))*a(n)
      fluxd   = fluxd  - coef*rem(n,1)
      fluxe   = fluxe  - coef*rem(n,5)
      fluxmx  = fluxmx - coef*rem(n,2)
      fluxmy  = fluxmy - coef*rem(n,3)
      fluxby  = fluxby - coef*rem(n,6)
      fluxmz  = fluxmz - coef*rem(n,4)
      fluxbz  = fluxbz - coef*rem(n,7)
  enddo

  !take half and put into the fmean variables 
  fmean(1) = half * fluxd 
  fmean(2) = half * fluxe 
  fmean(3) = half * fluxmx
  fmean(4) = zero
  fmean(5) = half * fluxmy
  fmean(6) = half * fluxby
  fmean(7) = half * fluxmz
  fmean(8) = half * fluxbz
  
  return

end subroutine athena_roe

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine EIGENVALUES
!
!> Computes mhd adiabatic eigenvalues.
!!
!! Input arguments:
!!   bx    = magnetic field in sweep direction
!!   d     = density
!!   vx    = x velocity 
!!   vy    = y velocity 
!!   vz    = z velocity 
!!   by    = y magnetic field
!!   bz    = z magnetic field
!!   p     = thermal pressure
!!
!! Output arguments:
!!
!!   lambda  = eigenvalues
!<
subroutine eigenvalues(d,vx,vy,vz,p,bx,by,bz,lambda)

  use parameters
  use gammas    , only : gamma
  use const
  
  implicit none

  real, intent(in) :: d,vx,vy,vz,p
  real, intent(in) :: bx,by,bz
  real, dimension(1:7), intent(out) :: lambda

  ! local variables
  real  :: btsq,bt,vsq,vax,vaxsq
  real  :: asq,astarsq,cfsq,cfast,cssq,cslow

  vsq = vx**2+vy**2+vz**2
  btsq = by**2+bz**2
  bt = sqrt(btsq)
  vaxsq = bx**2/d
  vax = sqrt(vaxsq)
  asq = gamma*p/d
  asq = max(asq,smallc**2)
  astarsq = asq+vaxsq+btsq/d
  
  cfsq = .5*(astarsq + sqrt(astarsq**2-4.0*asq*vaxsq))
  cfast = sqrt(cfsq)
  
  cssq = .5*(astarsq - sqrt(astarsq**2-4.0*asq*vaxsq))
  if (cssq .le. 0.) cssq = 0.
  cslow = sqrt(cssq)
  
  lambda(1) = vx - cfast
  lambda(2) = vx - vax
  lambda(3) = vx - cslow
  lambda(4) = vx
  lambda(5) = vx + cslow
  lambda(6) = vx + vax
  lambda(7) = vx + cfast
  
  return

end subroutine eigenvalues

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine EIGEN_CONS
!
!> Computes Roe eigenvalues.
!!
!! Input arguments:
!!   d     = roe density
!!   vx    = roe x velocity 
!!   vy    = roe y velocity 
!!   vz    = roe z velocity 
!!   bx    = magnetic field in sweep direction
!!   by    = roe y magnetic field
!!   bz    = roe z magnetic field
!!   h     = roe enthalpy
!!   xfac  = ((by^2-byl*byr)+bz^2-bzl*bzr))/(2*rho)
!!   yfac  = (rho_l+rho_r)/(2*rho)
!!
!! output arguments:
!!
!!   lambda = eigenvalues
!!   rem     = right eigenmatrix
!!   lem     = left  eigenmatrix
!<
subroutine eigen_cons(d,vx,vy,vz,h,bx,by,bz,xfac,yfac,lambda,rem,lem)

  use parameters
  use gammas    , only : gamma
  use const
  
  implicit none
  
  real, dimension(7,7) :: lem,rem
  real, dimension(7  ) :: lambda
  real                 :: d,vx,vy,vz,h
  real                 :: bx,by,bz,xfac,yfac
  real                 :: btsq, bt_starsq,bt,bt_star
  real                 :: vsq,vax,vaxsq,hp,twid_asq,q_starsq 
  real                 :: cfsq,cfast,cssq,cslow
  real                 :: beta_y,beta_z,beta_ystar,beta_zstar,beta_starsq,vbeta
  real                 :: alpha_f,alpha_s,droot,s,twid_a
  real                 :: qfast,qslow,af_prime,as_prime,afpbb,aspbb,na
  real                 :: cff,css,af,as,afpb,aspb,vqstr,norm
  real                 :: q_ystar,q_zstar
  
  vsq = vx*vx+vy*vy+vz*vz
  btsq = by*by+bz*bz
  bt_starsq = (gamma-1. - (gamma-2.)*yfac)*btsq
  bt=sqrt(btsq)
  bt_star = sqrt(bt_starsq)
  
  vaxsq = bx*bx/d
  vax = sqrt(vaxsq)
  
  hp = h - (vaxsq + btsq/d)
  twid_asq = ((gamma-1.)*(hp-.5*vsq)-(gamma-2.)*xfac)
  twid_asq = max(twid_asq,smallc*smallc)
  q_starsq = twid_asq+(vaxsq+bt_starsq/d)
  
  cfsq = .5*(q_starsq + sqrt(q_starsq*q_starsq-4.0*twid_asq*vaxsq))
  cfast = sqrt(cfsq)
  
  cssq = .5*(q_starsq - sqrt(q_starsq*q_starsq-4.0*twid_asq*vaxsq))
  if (cssq .le. 0.) cssq = 0.
  cslow = sqrt(cssq)
  
  if (bt .eq. 0) then
     beta_y = .5*sqrt(2.)
     beta_z = .5*sqrt(2.)
     beta_ystar = .5*sqrt(2.)
     beta_zstar = .5*sqrt(2.)
  else
     beta_y = by/bt
     beta_z = bz/bt
     beta_ystar = by/bt_star
     beta_zstar = bz/bt_star
  endif
  beta_starsq = beta_ystar*beta_ystar + beta_zstar*beta_zstar
  vbeta = vy*beta_ystar + vz*beta_zstar
  
  if ( (cfsq - cssq) .eq. 0.) then
     alpha_f = 1.0
     alpha_s = 0.0
  else if ( (twid_asq - cssq) .le. 0.) then
     alpha_f = 0.0
     alpha_s = 1.0
  else if ( (cfsq - twid_asq) .le. 0.) then
     alpha_f = 1.0
     alpha_s = 0.0
  else
     alpha_f = sqrt((twid_asq-cssq)/(cfsq-cssq))
     alpha_s = sqrt((cfsq-twid_asq)/(cfsq-cssq))
  endif
  !
  ! compute qs and as for eigenmatrices
  !
  droot = sqrt(d)
  s  = sign(one,bx)
  twid_a = sqrt(twid_asq)
  qfast = s*cfast*alpha_f
  qslow = s*cslow*alpha_s
  af_prime = twid_a*alpha_f/droot
  as_prime = twid_a*alpha_s/droot
  afpbb = af_prime*bt_star*beta_starsq
  aspbb = as_prime*bt_star*beta_starsq
  !
  ! eigenvalues
  !
  lambda(1) = vx - cfast
  lambda(2) = vx - vax
  lambda(3) = vx - cslow
  lambda(4) = vx
  lambda(5) = vx + cslow
  lambda(6) = vx + vax
  lambda(7) = vx + cfast
  !
  ! eigenmatrix
  !
  rem(1,1) = alpha_f
  rem(1,2) = alpha_f*(vx-cfast)
  rem(1,3) = alpha_f*vy + qslow*beta_ystar
  rem(1,4) = alpha_f*vz + qslow*beta_zstar
  rem(1,5) = alpha_f*(hp-vx*cfast) + qslow*vbeta + aspbb
  rem(1,6) = as_prime*beta_ystar
  rem(1,7) = as_prime*beta_zstar

  rem(2,1) = 0.
  rem(2,2) = 0.
  rem(2,3) = -beta_z
  rem(2,4) =  beta_y
  rem(2,5) = -(vy*beta_z - vz*beta_y)
  rem(2,6) = -s*beta_z/droot
  rem(2,7) =  s*beta_y/droot

  rem(3,1) = alpha_s
  rem(3,2) = alpha_s*(vx-cslow)
  rem(3,3) = alpha_s*vy - qfast*beta_ystar
  rem(3,4) = alpha_s*vz - qfast*beta_zstar
  rem(3,5) = alpha_s*(hp - vx*cslow) - qfast*vbeta - afpbb
  rem(3,6) = -af_prime*beta_ystar
  rem(3,7) = -af_prime*beta_zstar

  rem(4,1) = 1.0
  rem(4,2) = vx
  rem(4,3) = vy
  rem(4,4) = vz
  rem(4,5) = 0.5*vsq + (gamma-2.)*xfac/(gamma-1.)
  rem(4,6) = 0.
  rem(4,7) = 0.

  rem(5,1) =  alpha_s
  rem(5,2) =  alpha_s*(vx+cslow)
  rem(5,3) =  alpha_s*vy + qfast*beta_ystar
  rem(5,4) =  alpha_s*vz + qfast*beta_zstar
  rem(5,5) =  alpha_s*(hp+vx*cslow) + qfast*vbeta - afpbb
  rem(5,6) =  rem(3,6)
  rem(5,7) =  rem(3,7)

  rem(6,1) = 0.
  rem(6,2) = 0.
  rem(6,3) =  beta_z
  rem(6,4) = -beta_y
  rem(6,5) = -rem(2,5)
  rem(6,6) =  rem(2,6)
  rem(6,7) =  rem(2,7)

  rem(7,1) =  alpha_f
  rem(7,2) = alpha_f*(vx+cfast)
  rem(7,3) = alpha_f*vy - qslow*beta_ystar
  rem(7,4) = alpha_f*vz - qslow*beta_zstar
  rem(7,5) = alpha_f*(hp + vx*cfast) - qslow*vbeta + aspbb
  rem(7,6) =  rem(1,6)
  rem(7,7) =  rem(1,7)
  ! 
  ! left eignematrix
  !
  ! normalize some of the quantities by 1/(2a^2)
  ! some by (gamma-1)/2a^2
  !
  na  = 0.5/twid_asq
  cff = na*alpha_f*cfast
  css = na*alpha_s*cslow
  qfast = qfast*na
  qslow = qslow*na
  af = na*af_prime*d
  as = na*as_prime*d
  afpb = na*af_prime*bt_star
  aspb = na*as_prime*bt_star
  !
  alpha_f = (gamma-1.)*na*alpha_f
  alpha_s = (gamma-1.)*na*alpha_s
  q_ystar = beta_ystar/beta_starsq
  q_zstar = beta_zstar/beta_starsq
  vqstr   = (vy*q_ystar+vz*q_zstar)
  norm = (gamma-1.)*2.*na

  lem(1,1) = alpha_f*(vsq-hp) + cff*(cfast+vx) - qslow*vqstr - aspb
  lem(2,1) = -alpha_f*vx - cff
  lem(3,1) = -alpha_f*vy + qslow*q_ystar
  lem(4,1) = -alpha_f*vz + qslow*q_zstar
  lem(5,1) =  alpha_f
  lem(6,1) = as*q_ystar - alpha_f*by
  lem(7,1) = as*q_zstar - alpha_f*bz

  lem(1,2) =  0.5*(vy*beta_z - vz*beta_y)
  lem(2,2) =  0.
  lem(3,2) = -0.5*beta_z
  lem(4,2) =  0.5*beta_y
  lem(5,2) =  0.
  lem(6,2) = -0.5*droot*beta_z*s
  lem(7,2) =  0.5*droot*beta_y*s

  lem(1,3) =  alpha_s*(vsq-hp) + css*(cslow+vx) + qfast*vqstr + afpb
  lem(2,3) = -alpha_s*vx - css
  lem(3,3) = -alpha_s*vy - qfast*q_ystar
  lem(4,3) = -alpha_s*vz - qfast*q_zstar
  lem(5,3) =  alpha_s
  lem(6,3) = -af*q_ystar - alpha_s*by
  lem(7,3) = -af*q_zstar - alpha_s*bz

  ! CAUTION! There is a difference in sign compared to RAMSES
  lem(1,4) =  1. - norm*(.5*vsq - (gamma-2.)*xfac/(gamma-1.))
  ! this is the old version...
  !  lem(1,4) =  1. - norm*(.5*vsq + (gamma-2.)*xfac/(gamma-1.))
  lem(2,4) =  norm*vx
  lem(3,4) =  norm*vy
  lem(4,4) =  norm*vz
  lem(5,4) = -norm
  lem(6,4) =  norm*by
  lem(7,4) =  norm*bz

  lem(1,5) =  alpha_s*(vsq-hp) + css*(cslow-vx) - qfast*vqstr + afpb
  lem(2,5) = -alpha_s*vx + css 
  lem(3,5) = -alpha_s*vy + qfast*q_ystar 
  lem(4,5) = -alpha_s*vz + qfast*q_zstar 
  lem(5,5) =  alpha_s
  lem(6,5) =  lem(6,3)
  lem(7,5) =  lem(7,3)

  lem(1,6) = -lem(1,2)
  lem(2,6) =  0.
  lem(3,6) = -lem(3,2)
  lem(4,6) = -lem(4,2)
  lem(5,6) =  0.
  lem(6,6) =  lem(6,2)
  lem(7,6) =  lem(7,2)

  lem(1,7) =  alpha_f*(vsq-hp) + cff*(cfast-vx) + qslow*vqstr -aspb
  lem(2,7) = -alpha_f*vx + cff
  lem(3,7) = -alpha_f*vy - qslow*q_ystar
  lem(4,7) = -alpha_f*vz - qslow*q_zstar
  lem(5,7) =  alpha_f
  lem(6,7) =  lem(6,1)
  lem(7,7) =  lem(7,1)
  
  return

end subroutine eigen_cons
