!> \file
!! Contains subroutine read_valp() and function
!! interpol_valp()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine READ_VALP
!
!> Reads the tabulated eigenvalues for the M1 radiative
!! transfer solver.
!<
subroutine read_valp

  use valeurs_propres, only :valp,n_points

  implicit none
  
  integer            :: i,j,k
  real, dimension(4) :: x
  character(len=100) :: file,dir_valp
  character(len=1  ) :: dir_eos        ! not used here, just needed to read the namelist

  namelist/dir/dir_valp,dir_eos

  dir_valp='.'
  open  (1,file='liste.nml')
  rewind(1                 )
  read  (1,NML=dir         )
  close (1                 )

  file=TRIM(dir_valp)//'/valp_spher_eps.dat'

  open(unit=10,file=file,form='formatted',status='old',action='read')
  read(10,100) n_points

  i = 0 ; j = 0 ; k = 0

  allocate(valp(0:n_points,0:n_points,0:n_points,4))

  do while (i.ne.n_points .or. j.ne.n_points .or. k.ne.n_points)
     read(10,200) i,j,k,x
     valp(i,j,k,1) = x(1)
     valp(i,j,k,2) = x(2)
     valp(i,j,k,3) = x(3)
     valp(i,j,k,4) = x(4)
  enddo
  close(10)

100 format(i3)
200 format(3(i3,1x),4(e13.5))

  return

end subroutine read_valp

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Function INTERPOL_VALP
!
!> Performs interpolation between tabulated point on the
!! eigenvalues curves to find a particular eigenvalue
!! during the simulation.
!<
function interpol_valp(f,t,e,ivalp)

  use parameters     , only : pi
  use valeurs_propres, only : n_points,valp

  implicit none
  
  integer :: ivalp
  integer :: i_f,i_t,i_e
  integer :: i_f_p,i_t_p,i_e_p
  
  real :: f,t,e
  real :: interpol_valp
  real :: fmin,tmin,emin,df,dt,de,lf,lt,le,dd1,dd2,dd3,de1,de2,de3

  fmin = 0.
  tmin = 0.
  emin = 0.

  df = 1./n_points
  dt = pi/n_points
  de = 1./n_points

  lf = (f -fmin )/df   ! No +0.5 since we calculate them at the cell interfaces and not the centres
  lt = (t -tmin )/dt
  le = (e -emin )/de

  i_f = int(lf) ; i_t = int(lt) ; i_e = int(le)

  dd1 = lf - float(i_f) 
  dd2 = lt - float(i_t)
  dd3 = le - float(i_e)

  de1 = 1.0 - dd1
  de2 = 1.0 - dd2
  de3 = 1.0 - dd3

  ! Boundary conditions: at the boundaries, dd = 0 so value of b is unimportant as long as b < 1
  i_f_p = min(n_points,i_f+1)
  i_t_p = min(n_points,i_t+1)
  i_e_p = min(n_points,i_e+1)

  interpol_valp = 0. 
  interpol_valp = interpol_valp + de1*de2*de3*valp(i_f  ,i_t  ,i_e  ,ivalp)
  interpol_valp = interpol_valp + dd1*de2*de3*valp(i_f_p,i_t  ,i_e  ,ivalp)
  interpol_valp = interpol_valp + de1*dd2*de3*valp(i_f  ,i_t_p,i_e  ,ivalp)
  interpol_valp = interpol_valp + dd1*dd2*de3*valp(i_f_p,i_t_p,i_e  ,ivalp)
  interpol_valp = interpol_valp + de1*de2*dd3*valp(i_f  ,i_t  ,i_e_p,ivalp)
  interpol_valp = interpol_valp + dd1*de2*dd3*valp(i_f_p,i_t  ,i_e_p,ivalp)
  interpol_valp = interpol_valp + de1*dd2*dd3*valp(i_f  ,i_t_p,i_e_p,ivalp)
  interpol_valp = interpol_valp + dd1*dd2*dd3*valp(i_f_p,i_t_p,i_e_p,ivalp)

end function interpol_valp
