## @Package reading heracles output
#

import tables
import numpy as np

from utils import *
import fortranfile as ff

## reading routine for Heracles Outputs 
# 
# idump: number of the output
# form: if 'hdf' then read hdf5 else read binary
# opt: if 1 read the corresponding array
# opt[0]: rho
# opt[1]: u
# opt[2]: E
# opt[3]: fx
# opt[4]: er,fr
# opt[5]: ed
# opt[6]: B
#
# verbose :  mode 
#        
# endian : character, optional
# Specify the endian-ness of the file.  Possible values are
# '>', '<', '@' and '='.  See the documentation of Python's
# struct module for their meanings.  The deafult is '>' (native
# byte order)
# header_prec : character, optional
# Specify the precision used for the record headers.  Possible
# values are 'h', 'i', 'l' and 'q' with their meanings from
# Python's struct module.  The default is 'i' (the system's
# default integer).
#
# dir : directory where the data are stored (Default: current directory)
#
# the structure contains:
# self.xg,yg,zg: global mesh
# self.nxglob,nyglob,nzglob,nfx,ndim: global size
# self.time,dt,dtold: time variable
# self.nite: number of iteration
# self.gamma: adiabatic exponent
# self.mu: molecular weight
# self.u_M: mass unit
# self.u_L: length unit
# self.u_T: time unit
# self.mhd,diffu,ray,conduc,grav: physics flags
# self.rho,u,E,fx,er,fr,ed,B: physical fields
class HeraclesData: 
 
    def __init__(self,idump=1,form='hdf',opt=[1,1,1,1,1,1,1],verbose=True,endian='@', header_prec='i',dir='.'):
        
        #read basic file data

        if form=='hdf':

            #from hdf file
            
            if verbose:
                print('Reading Heracles hdf5 outputs')

            job_params=JobParams(dir)
            dir=dir+'/'+str_suffix(idump,length=3)+'/'

            ncpu   = job_params.ncpu

            self.ndim     = job_params.ndim            
            self.nfx      = job_params.nfx
            self.ngrp_ray = job_params.ngrp_ray
            self.ngrp_dif = job_params.ngrp_dif

            ndim = self.ndim ; nfx = self.nfx ; N_vit = ndim ; ngrp_ray=self.ngrp_ray ; ngrp_dif=self.ngrp_dif
            
            self.nxglob = job_params.nxglob
            self.nyglob = job_params.nyglob
            self.nzglob = job_params.nzglob

            nxglob=self.nxglob ; nyglob=self.nyglob ; nzglob=self.nzglob 
            
            self.geometry = job_params.geometry

            self.xg     = job_params.xg
            self.yg     = job_params.yg
            self.zg     = job_params.zg
            
            self.nite   = job_params.nite
            self.time   = job_params.time
            self.dt     = job_params.dt
            self.dtold  = job_params.dtold
            self.gamma  = job_params.gamma
            self.mu     = job_params.mu
	    self.u_M    = job_params.u_M
	    self.u_L    = job_params.u_L
	    self.u_T    = job_params.u_T

            self.mhd    = job_params.mhd
            self.ray    = job_params.ray
            self.diffu  = job_params.diffu
            self.conduc = job_params.conduc
            self.grav   = job_params.grav

            nbuf = job_params.nbuf

            mhd=self.mhd ; ray=self.ray ; diffu=self.diffu
            
            if mhd:
                N_vit = 3

        else:

            #from binary file
            
            if verbose:
                print("Reading Heracles binary outputs")
            
            dir=dir+'/'+str_suffix(idump,length=3)+'/'
           
            #Read first output to get basic info
            filename=dir+'liste-00001_'+str_suffix(idump,length=3)+'.Bin'
            
            f=ff.FortranFile(filename,endian=endian,header_prec=header_prec)

            [nxglob,nyglob,nzglob,ncpu,geometry]= f.readType('4i5s')

            self.nxglob=nxglob ; self.nyglob=nyglob ; self.nzglob=nzglob ; self.geometry=geometry
            

            mesh = f.readReals('d')
            
            self.xg  = mesh[0              :nxglob+1              ]
            self.yg  = mesh[nxglob+1       :nxglob+nyglob+2       ]
            self.zg  = mesh[nxglob+nyglob+2:nxglob+nyglob+nzglob+3]
                      
            [ndim,nx,ny,nz,nfx,nbuf,ngrp_ray,ngrp_dif] = f.readInts('i')
            self.nfx=nfx ; self.ndim=ndim ; self.ngrp_ray=ngrp_ray ; self.ngrp_dif=ngrp_dif
            
            N_vit = ndim
        
            [mhd,ray,diffu,conduc,grav] = f.readInts('i')
            
            self.mhd=mhd ; self.ray=ray ; self.diffu=diffu ; self.conduc=conduc ; self.grav=grav

            if mhd:
                N_vit = 3
            
            [istart,jstart,kstart,iend,jend,kend] = f.readInts('i')
        
            local_mesh = f.readReals('d')

            shift_gr  = f.readReals('d')
            nshift_gr = f.readInts('i')
        
            self.nite                      = f.readInts('i')
            [self.time,self.dt,self.dtold] = f.readReals('d')
            [self.gamma,self.mu]           = f.readReals('d')
	    [self.u_M,self.u_L,self.u_T]   = f.readReals('d')
            
            f.close()



        # buffer zone
        
        nbufx = nbuf
        nbufy = 0
        nbufz = 0
        
        if ndim>=2:
            nbufy = nbuf 
        if ndim>=3:
            nbufz = nbuf
            
        #Define output arrays
        if opt[0]==1:
            self.rho = np.zeros((nxglob,nyglob,nzglob      ))
        if opt[1]==1:
            self.u   = np.zeros((nxglob,nyglob,nzglob,N_vit))
        if opt[2]==1:
            self.E   = np.zeros((nxglob,nyglob,nzglob      ))
        if opt[3]==1 and nfx>0:
            self.fx = np.zeros((nxglob,nyglob,nzglob,nfx    ))
        if opt[4]==1 and ray:
            self.er = np.zeros((nxglob,nyglob,nzglob        ))
            self.fr = np.zeros((nxglob,nyglob,nzglob,ndim   ))
        if opt[5]==1 and diffu:
            self.ed = np.zeros((nxglob,nyglob,nzglob        ))
        if opt[6]==1 and mhd:
            self.B  = np.zeros((nxglob,nyglob,nzglob,N_vit  ))
            
        #Define centered mesh
        self.xcg  = np.zeros(nxglob)
        self.ycg  = np.zeros(nyglob)
        self.zcg  = np.zeros(nzglob)

        for i in range(nxglob):
            self.xcg[i] = 0.5*(self.xg[i]+self.xg[i+1])
        for i in range(nyglob):
            self.ycg[i] = 0.5*(self.yg[i]+self.yg[i+1])
        for i in range(nzglob):
            self.zcg[i] = 0.5*(self.zg[i]+self.zg[i+1])

        #Loop on cpus
        if verbose:
            print("Reading "+str(ncpu)+" files")

        for icpu in range(ncpu):

            if form=='hdf':

              
                filename=dir+"myjob-"+str_suffix(icpu+1,length=5)+"_"+\
                    str_suffix(idump ,length=3)+".hdf"
               
                data=HdfFile(filename,opt=opt)
                
            else:

                mype = icpu+1
                filename=dir+'liste-'+str_suffix(mype,length=5)+'_'+str_suffix(idump,length=3)+'.Bin'
                
                data=BinFile(filename,opt=opt,endian=endian,header_prec=header_prec)
                
            istart=data.istart ; iend=data.iend
            jstart=data.jstart ; jend=data.jend
            kstart=data.kstart ; kend=data.kend
            
            nx = data.nx ; ny = data.ny ; nz = data.nz 
               
            if ndim==1:
                if opt[0]==1:
                    self.rho [istart-1:iend,0,0  ] = data.rho [nbuf:nx+nbuf,0,0  ]
                if opt[1]==1:
                    for i in range(N_vit):
                        self.u   [istart-1:iend,0,0,i] = data.rhou[nbuf:nx+nbuf,0,0,i]/data.rho[nbuf:nx+nbuf,0,0]
                if opt[2]==1:
                    self.E   [istart-1:iend,0,0  ] = data.E   [nbuf:nx+nbuf,0,0  ]
                if opt[3]==1 and nfx>0:
                    self.fx  [istart-1:iend,0,0,:] = data.fx  [nbuf:nx+nbuf,0,0,:]
                if opt[4]==1 and ray:
                    self.er  [istart-1:iend,0,0  ] = data.er  [nbuf:nx+nbuf,0,0  ]
                    self.fr  [istart-1:iend,0,0,:] = data.fr  [nbuf:nx+nbuf,0,0,:]
                if opt[5]==1 and diffu:
                    self.ed  [istart-1:iend,0,0  ] = data.ed  [nbuf:nx+nbuf,0,0  ]
                if opt[6]==1 and mhd:
                    self.B   [istart-1:iend,0,0,:] = data.B   [nbuf:nx+nbuf,0,0,:]
            if ndim==2:
                if opt[0]==1:
                    self.rho [istart-1:iend,jstart-1:jend,0  ] = data.rho [nbuf:nx+nbuf,nbuf:ny+nbuf,0  ]
                if opt[1]==1:
                    for i in range(N_vit):
                        self.u   [istart-1:iend,jstart-1:jend,0,i] = data.rhou[nbuf:nx+nbuf,nbuf:ny+nbuf,0,i]/data.rho [nbuf:nx+nbuf,nbuf:ny+nbuf,0  ]
                if opt[2]==1:
                    self.E   [istart-1:iend,jstart-1:jend,0  ] = data.E   [nbuf:nx+nbuf,nbuf:ny+nbuf,0  ]
                if opt[3]==1 and nfx>0:
                    self.fx  [istart-1:iend,jstart-1:jend,0,:] = data.fx  [nbuf:nx+nbuf,nbuf:ny+nbuf,0,:]
                if opt[4]==1 and ray:
                    self.er  [istart-1:iend,jstart-1:jend,0  ] = data.er  [nbuf:nx+nbuf,nbuf:ny+nbuf,0  ]
                    self.fr  [istart-1:iend,jstart-1:jend,0,:] = data.fr  [nbuf:nx+nbuf,nbuf:ny+nbuf,0,:]
                if opt[5]==1 and diffu:
                    self.ed  [istart-1:iend,jstart-1:jend,0  ] = data.ed  [nbuf:nx+nbuf,nbuf:ny+nbuf,0  ]
                if opt[6]==1 and mhd:
                    self.B   [istart-1:iend,jstart-1:jend,0,:] = data.B   [nbuf:nx+nbuf,nbuf:ny+nbuf,0,:]
                    
            if ndim==3:
                if opt[0]==1:
                    self.rho [istart-1:iend,jstart-1:jend,kstart-1:kend  ] = data.rho [nbuf:nx+nbuf,nbuf:ny+nbuf,nbuf:nz+nbuf  ]
                if opt[1]==1:
                    for i in range(N_vit):
                        self.u   [istart-1:iend,jstart-1:jend,kstart-1:kend,i] = data.rhou[nbuf:nx+nbuf,nbuf:ny+nbuf,nbuf:nz+nbuf,i]/data.rho [nbuf:nx+nbuf,nbuf:ny+nbuf,nbuf:nz+nbuf  ]
                if opt[2]==1:
                    self.E   [istart-1:iend,jstart-1:jend,kstart-1:kend  ] = data.E   [nbuf:nx+nbuf,nbuf:ny+nbuf,nbuf:nz+nbuf  ]
                if opt[3]==1 and nfx>0:
                    self.fx  [istart-1:iend,jstart-1:jend,kstart-1:kend,:] = data.fx  [nbuf:nx+nbuf,nbuf:ny+nbuf,nbuf:nz+nbuf,:]
                if opt[4]==1 and ray:
                    self.er  [istart-1:iend,jstart-1:jend,kstart-1:kend  ] = data.er  [nbuf:nx+nbuf,nbuf:ny+nbuf,nbuf:nz+nbuf  ]
                    self.fr  [istart-1:iend,jstart-1:jend,kstart-1:kend,:] = data.fr  [nbuf:nx+nbuf,nbuf:ny+nbuf,nbuf:nz+nbuf,:]
                if opt[5]==1 and diffu:
                    self.ed  [istart-1:iend,jstart-1:jend,kstart-1:kend  ] = data.ed  [nbuf:nx+nbuf,nbuf:ny+nbuf,nbuf:nz+nbuf  ]
                if opt[6]==1 and mhd:
                    self.B   [istart-1:iend,jstart-1:jend,kstart-1:kend,:] = data.B   [nbuf:nx+nbuf,nbuf:ny+nbuf,nbuf:nz+nbuf,:]
            del data

    def get_eint(self):
        u2=self.u[:,:,:,0]**2
        for i in range(1,self.ndim):
            u2=u2+self.u[:,:,:,i]**2
        eint=self.E[:,:,:]-0.5*self.rho[:,:,:]*u2
        if self.mhd:
            eint=eint-0.5*(self.B[:,:,:,0]**2+self.B[:,:,:,1]**2+self.B[:,:,:,2]**2)
        return eint

    def get_P(self):
        P=self.get_eint()*(self.gamma-1)
        return P

    def get_T(self):
        kb  = 1.38062e-16 
        uma = 1.660531e-24
        T=self.get_P()/self.rho*(self.mu*uma/kb)
        return T

    def get_slice_x(self,type='rho',ycut=0,zcut=0,ifx=0):
        if type=='rho':
            slice=self.rho[:,ycut,zcut]
        elif type=='vx':
            slice=self.u[:,ycut,zcut,0]
        elif type=='vy':
            slice=self.u[:,ycut,zcut,1]
        elif type=='vz':
            slice=self.u[:,ycut,zcut,2]
        elif type=='bx':
            slice=self.B[:,ycut,zcut,0]
        elif type=='by':
            slice=self.B[:,ycut,zcut,1]
        elif type=='bz':
            slice=self.B[:,ycut,zcut,2]
        elif type=='E':
            slice=self.E[:,ycut,zcut]
        elif type=='P':
            P=self.get_P()
            slice=P[:,ycut,zcut]
        elif type=='T':
            T=self.get_T()
            slice=T[:,ycut,zcut]
        elif type=='fx':
            slice=self.fx[:,ycut,zcut,ifx]
        elif type=='Ed':
            slice=self.ed[:,ycut,zcut]
        elif type=='Td':
            a_r = 7.56464e-15 #* erg/centimetre**3/degre**4 
            slice=(self.ed[:,ycut,zcut]/a_r)**0.25
        elif type=='Er':
            slice=self.er[:,ycut,zcut]
        elif type=='Tr':
            a_r = 7.56464e-15 #* erg/centimetre**3/degre**4 
            slice=(self.er[:,ycut,zcut]/a_r)**0.25
        elif type=='Frx':
            slice=self.fr[:,ycut,zcut,0]
        elif type=='Fry':
            slice=self.fr[:,ycut,zcut,1]
        elif type=='Frz':
            slice=self.fr[:,ycut,zcut,2]
        elif type=='frx':
            c   = 2.997925e+10 #* centimetre/seconde 
            slice=self.fr[:,ycut,zcut,0]/(c*self.er[:,ycut,zcut])
        elif type=='fry':
            c   = 2.997925e+10 #* centimetre/seconde 
            slice=self.fr[:,ycut,zcut,1]/(c*self.er[:,ycut,zcut])
        elif type=='frz':
            c   = 2.997925e+10 #* centimetre/seconde 
            slice=self.fr[:,ycut,zcut,2]/(c*self.er[:,ycut,zcut])
            
        return slice

    def get_slice_y(self,type='rho',xcut=0,zcut=0,ifx=0):
        if type=='rho':
            slice=self.rho[xcut,:,zcut]
        elif type=='vx':
            slice=self.u[xcut,:,zcut,0]
        elif type=='vy':
            slice=self.u[xcut,:,zcut,1]
        elif type=='vz':
            slice=self.u[xcut,:,zcut,2]
        elif type=='bx':
            slice=self.B[xcut,:,zcut,0]
        elif type=='by':
            slice=self.B[xcut,:,zcut,1]
        elif type=='bz':
            slice=self.B[xcut,:,zcut,2]
        elif type=='E':
            slice=self.E[xcut,:,zcut]
        elif type=='P':
            P=self.get_P()
            slice=P[xcut,:,zcut]
        elif type=='T':
            T=self.get_T()
            slice=T[xcut,:,zcut]
        elif type=='fx':
            slice=self.fx[xcut,:,zcut,ifx]
        elif type=='Ed':
            slice=self.ed[xcut,:,zcut]
        elif type=='Td':
            a_r = 7.56464e-15 #* erg/centimetre**3/degre**4 
            slice=(self.ed[xcut,:,zcut]/a_r)**0.25
        elif type=='Er':
            slice=self.er[xcut,:,zcut]
        elif type=='Tr':
            a_r = 7.56464e-15 #* erg/centimetre**3/degre**4 
            slice=(self.er[xcut,:,zcut]/a_r)**0.25
        elif type=='Frx':
            slice=self.fr[xcut,:,zcut,0]
        elif type=='Fry':
            slice=self.fr[xcut,:,zcut,1]
        elif type=='Frz':
            slice=self.fr[xcut,:,zcut,2]
        elif type=='frx':
            c   = 2.997925e+10 #* centimetre/seconde 
            slice=self.fr[xcut,:,zcut,0]/(c*self.er[xcut,:,zcut])
        elif type=='fry':
            c   = 2.997925e+10 #* centimetre/seconde 
            slice=self.fr[xcut,:,zcut,1]/(c*self.er[xcut,:,zcut])
        elif type=='frz':
            c   = 2.997925e+10 #* centimetre/seconde 
            slice=self.fr[xcut,:,zcut,2]/(c*self.er[xcut,:,zcut])
            
        return slice

    def get_slice_z(self,type='rho',xcut=0,ycut=0,ifx=0):
        if type=='rho':
            slice=self.rho[xcut,ycut,:]
        elif type=='vx':
            slice=self.u[xcut,ycut,:,0]
        elif type=='vy':
            slice=self.u[xcut,ycut,:,1]
        elif type=='vz':
            slice=self.u[xcut,ycut,:,2]
        elif type=='bx':
            slice=self.B[xcut,ycut,:,0]
        elif type=='by':
            slice=self.B[xcut,ycut,:,1]
        elif type=='bz':
            slice=self.B[xcut,ycut,:,2]
        elif type=='E':
            slice=self.E[xcut,ycut,:]
        elif type=='P':
            P=self.get_P()
            slice=P[xcut,ycut,:]
        elif type=='T':
            T=self.get_T()
            slice=T[xcut,ycut,:]
        elif type=='fx':
            slice=self.fx[xcut,ycut,:,ifx]
        elif type=='Ed':
            slice=self.ed[xcut,ycut,:]
        elif type=='Td':
            a_r = 7.56464e-15 #* erg/centimetre**3/degre**4 
            slice=(self.ed[xcut,ycut,:]/a_r)**0.25
        elif type=='Er':
            slice=self.er[xcut,ycut,:]
        elif type=='Tr':
            a_r = 7.56464e-15 #* erg/centimetre**3/degre**4 
            slice=(self.er[xcut,ycut,:]/a_r)**0.25
        elif type=='Frx':
            slice=self.fr[xcut,ycut,:,0]
        elif type=='Fry':
            slice=self.fr[xcut,ycut,:,1]
        elif type=='Frz':
            slice=self.fr[xcut,ycut,:,2]
        elif type=='frx':
            c   = 2.997925e+10 #* centimetre/seconde 
            slice=self.fr[xcut,ycut,:,0]/(c*self.er[xcut,ycut,:])
        elif type=='fry':
            c   = 2.997925e+10 #* centimetre/seconde 
            slice=self.fr[xcut,ycut,:,1]/(c*self.er[xcut,ycut,:])
        elif type=='frz':
            c   = 2.997925e+10 #* centimetre/seconde 
            slice=self.fr[xcut,ycut,:,2]/(c*self.er[xcut,ycut,:])
            
        return slice

    def get_slice_xy(self,type='rho',zcut=0,ifx=0):
        if type=='rho':
            slice=self.rho[:,:,zcut]
        elif type=='vx':
            slice=self.u[:,:,zcut,0]
        elif type=='vy':
            slice=self.u[:,:,zcut,1]
        elif type=='vz':
            slice=self.u[:,:,zcut,2]
        elif type=='bx':
            slice=self.B[:,:,zcut,0]
        elif type=='by':
            slice=self.B[:,:,zcut,1]
        elif type=='bz':
            slice=self.B[:,:,zcut,2]
        elif type=='E':
            slice=self.E[:,:,zcut]
        elif type=='P':
            P=self.get_P()
            slice=P[:,:,zcut]
        elif type=='T':
            T=self.get_T()
            slice=T[:,:,zcut]
        elif type=='fx':
            slice=self.fx[:,:,zcut,ifx]
        elif type=='Ed':
            slice=self.ed[:,:,zcut]
        elif type=='Td':
            a_r = 7.56464e-15 #* erg/centimetre**3/degre**4 
            slice=(self.ed[:,:,zcut]/a_r)**0.25
        elif type=='Er':
            slice=self.er[:,:,zcut]
        elif type=='Tr':
            a_r = 7.56464e-15 #* erg/centimetre**3/degre**4 
            slice=(self.er[:,:,zcut]/a_r)**0.25
        elif type=='Frx':
            slice=self.fr[:,:,zcut,0]
        elif type=='Fry':
            slice=self.fr[:,:,zcut,1]
        elif type=='Frz':
            slice=self.fr[:,:,zcut,2]
        elif type=='frx':
            c   = 2.997925e+10 #* centimetre/seconde 
            slice=self.fr[:,:,zcut,0]/(c*self.er[:,:,zcut])
        elif type=='fry':
            c   = 2.997925e+10 #* centimetre/seconde 
            slice=self.fr[:,:,zcut,1]/(c*self.er[:,:,zcut])
        elif type=='frz':
            c   = 2.997925e+10 #* centimetre/seconde 
            slice=self.fr[:,:,zcut,2]/(c*self.er[:,:,zcut])
            
        return slice

    def get_slice_yz(self,type='rho',xcut=0,ifx=0):
        if type=='rho':
            slice=self.rho[xcut,:,:]
        elif type=='vx':
            slice=self.u[xcut,:,:,0]
        elif type=='vy':
            slice=self.u[xcut,:,:,1]
        elif type=='vz':
            slice=self.u[xcut,:,:,2]
        elif type=='bx':
            slice=self.B[xcut,:,:,0]
        elif type=='by':
            slice=self.B[xcut,:,:,1]
        elif type=='bz':
            slice=self.B[xcut,:,:,2]
        elif type=='E':
            slice=self.E[xcut,:,:]
        elif type=='P':
            P=self.get_P()
            slice=P[xcut,:,:]
        elif type=='T':
            T=self.get_T()
            slice=T[xcut,:,:]
        elif type=='fx':
            slice=self.fx[xcut,:,:,ifx]
        elif type=='Ed':
            slice=self.ed[xcut,:,:]
        elif type=='Td':
            a_r = 7.56464e-15 #* erg/centimetre**3/degre**4 
            slice=(self.ed[xcut,:,:]/a_r)**0.25
        elif type=='Er':
            slice=self.er[xcut,:,:]
        elif type=='Tr':
            a_r = 7.56464e-15 #* erg/centimetre**3/degre**4 
            slice=(self.er[xcut,:,:]/a_r)**0.25
        elif type=='Frx':
            slice=self.fr[xcut,:,:,0]
        elif type=='Fry':
            slice=self.fr[xcut,:,:,1]
        elif type=='Frz':
            slice=self.fr[xcut,:,:,2]
        elif type=='frx':
            c   = 2.997925e+10 #* centimetre/seconde 
            slice=self.fr[xcut,:,:,0]/(c*self.er[xcut,:,:])
        elif type=='fry':
            c   = 2.997925e+10 #* centimetre/seconde 
            slice=self.fr[xcut,:,:,1]/(c*self.er[xcut,:,:])
        elif type=='frz':
            c   = 2.997925e+10 #* centimetre/seconde 
            slice=self.fr[xcut,:,:,2]/(c*self.er[xcut,:,:])
            
        return slice

    def get_slice_xz(self,type='rho',ycut=0,ifx=0):
        if type=='rho':
            slice=self.rho[:,ycut,:]
        elif type=='vx':
            slice=self.u[:,ycut,:,0]
        elif type=='vy':
            slice=self.u[:,ycut,:,1]
        elif type=='vz':
            slice=self.u[:,ycut,:,2]
        elif type=='bx':
            slice=self.B[:,ycut,:,0]
        elif type=='by':
            slice=self.B[:,ycut,:,1]
        elif type=='bz':
            slice=self.B[:,ycut,:,2]
        elif type=='E':
            slice=self.E[:,ycut,:]
        elif type=='P':
            P=self.get_P()
            slice=P[:,ycut,:]
        elif type=='T':
            T=self.get_T()
            slice=T[:,ycut,:]
        elif type=='fx':
            slice=self.fx[:,ycut,:,ifx]
        elif type=='Ed':
            slice=self.ed[:,ycut,:]
        elif type=='Td':
            a_r = 7.56464e-15 #* erg/centimetre**3/degre**4 
            slice=(self.ed[:,ycut,:]/a_r)**0.25
        elif type=='Er':
            slice=self.er[:,ycut,:]
        elif type=='Tr':
            a_r = 7.56464e-15 #* erg/centimetre**3/degre**4 
            slice=(self.er[:,ycut,:]/a_r)**0.25
        elif type=='Frx':
            slice=self.fr[:,ycut,:,0]
        elif type=='Fry':
            slice=self.fr[:,ycut,:,1]
        elif type=='Frz':
            slice=self.fr[:,ycut,:,2]
        elif type=='frx':
            c   = 2.997925e+10 #* centimetre/seconde 
            slice=self.fr[:,ycut,:,0]/(c*self.er[:,ycut,:])
        elif type=='fry':
            c   = 2.997925e+10 #* centimetre/seconde 
            slice=self.fr[:,ycut,:,1]/(c*self.er[:,ycut,:])
        elif type=='frz':
            c   = 2.997925e+10 #* centimetre/seconde 
            slice=self.fr[:,ycut,:,2]/(c*self.er[:,ycut,:])
            
        return slice

## Read a hdf output on one cpu
#
# self.x,y,z,geometry: local mesh
# self.nx,ny,nz: local size
# self.istart ... kend: position in the global mesh
# self.rho ... B: local fields
# can be optimize with opt, see HeraclesData
# ! it is possible to read only one field without reading the others (hdf output)
class HdfFile:

    def __init__(self,filename="myjob-o-001_005.h5",opt=[1,1,1,1,1,1,1]):
        """ Read HDF5 data output from Heracles."""
        f=tables.openFile(filename)
        
        self.x=f.root.x[:]
        self.y=f.root.y[:]
        self.z=f.root.z[:]

        if opt[0]==1 or opt[1]==1:
            self.rho  = np.transpose(f.root.rho [:,:,:  ],(  2,1,0))
        if opt[1]==1:
            self.rhou = np.transpose(f.root.rhou[:,:,:,:],(3,2,1,0))
        if opt[2]==1:
            self.E    = np.transpose(f.root.E   [:,:,:  ],(  2,1,0))
        
        self.geometry = f.root.geometry[0]

        ray   = f.root.para_int[15]
        diffu = f.root.para_int[16]
        conduc= f.root.para_int[17]
        mhd   = f.root.para_int[18]
        grav  = f.root.para_int[19]

        nbuf = f.root.para_int[20]

        ngrp_ray = f.root.para_int[21]
        ngrp_dif = f.root.para_int[22]

        nfx = f.root.para_int[7]
           
        if opt[3]==1 and nfx>0:
            self.fx=np.transpose(f.root.fx[:,:,:,:],(3,2,1,0))
        if opt[4]==1 and ray:
            self.er=np.transpose(f.root.er[:,:,:  ],(  2,1,0))
            self.fr=np.transpose(f.root.fr[:,:,:,:],(3,2,1,0))
        if opt[5]==1 and diffu:
            self.ed=np.transpose(f.root.erd[:,:,: ],(  2,1,0))
        if opt[6]==1 and mhd:
            self.B =np.transpose(f.root.B   [:,:,:,:],(3,2,1,0))

        self.istart=f.root.para_int[8]  ; self.iend=f.root.para_int[11]
        self.jstart=f.root.para_int[9]  ; self.jend=f.root.para_int[12]
        self.kstart=f.root.para_int[10] ; self.kend=f.root.para_int[13]

        self.nx     = f.root.para_int[4]
        self.ny     = f.root.para_int[5]
        self.nz     = f.root.para_int[6]
        
        f.close()

            
class JobParams:

    def __init__(self,dir='.'):
        """ Read HDF5 file myjob.h5 from Heracles."""
      
        f=tables.openFile(dir+"/myjob.h5")
      
        self.ncpu   = f.root.job_params[0]
        
        self.nxglob = f.root.para_int[0]
        self.nyglob = f.root.para_int[1]
        self.nzglob = f.root.para_int[2]
        self.ndim   = f.root.para_int[3]
        self.nfx    = f.root.para_int[7]

        self.nite   = f.root.para_int[14]

        self.xg=f.root.x_glob[:]
        self.yg=f.root.y_glob[:]
        self.zg=f.root.z_glob[:]
        
        self.time   = f.root.para_real[0]
        self.dt     = f.root.para_real[1]
        self.dtold  = f.root.para_real[2]
        self.gamma  = f.root.para_real[3]
        self.mu     = f.root.para_real[4]
	self.u_M    = f.root.para_real[5]
	self.u_L    = f.root.para_real[6]
	self.u_T    = f.root.para_real[7]

        self.geometry = f.root.geometry[0]
       
        self.ray   = f.root.para_int[15]
        self.diffu = f.root.para_int[16]
        self.conduc= f.root.para_int[17]
        self.mhd   = f.root.para_int[18]
        self.grav  = f.root.para_int[19]

        self.nbuf = f.root.para_int[20]

        self.ngrp_ray = f.root.para_int[21]
        self.ngrp_dif = f.root.para_int[22]
        
        f.close()

## Read a binary output on one cpu
#
# self.x,y,z,geometry: local mesh
# self.nx,ny,nz: local size
# self.istart ... kend: position in the global mesh
# self.rho ... B: local fields
# can be optimize with opt, see HeraclesData
# can change endian and header_prec see HeraclesData
# ! it is not possible to read only one field if all the fields before it are not read (binary output)
class BinFile:

    def __init__(self,filename='liste-00001_000.Bin',endian='@', header_prec='i',opt=[1,1,1,1,1,1,1]):
        #Open file
        f=ff.FortranFile(filename,endian=endian,header_prec=header_prec)
        
        #Get data dimensions from file
        [nxglob,nyglob,nzglob,ncpu,geometry] = f.readType('4i5s')

        self.geometry=geometry

        mesh = f.readReals('d')
                      
        [ndim,nx,ny,nz,nfx,nbuf,ngrp_ray,ngrp_dif] = f.readInts('i')
        N_vit = ndim ; self.nx=nx ; self.ny=ny ; self.nz=nz ; self.ngrp_ray=ngrp_ray ; self.ngrp_dif=ngrp_dif
        
            
        N_vit = ndim
        
        [mhd,ray,diffu,conduc,grav] = f.readInts('i')
        
        self.mhd=mhd ; self.ray=ray ; self.diffu=diffu ; self.conduc=conduc ; self.grav=grav
            
        if mhd:
            N_vit = 3
            
        [self.istart,self.jstart,self.kstart,self.iend,self.jend,self.kend] = f.readInts('i')
        
        local_mesh = f.readReals('d')
        
        self.x  = local_mesh[0              :nxglob+1              ]
        self.y  = local_mesh[nxglob+1       :nxglob+nyglob+2       ]
        self.z  = local_mesh[nxglob+nyglob+2:nxglob+nyglob+nzglob+3]
        
        shift_gr  = f.readReals('d')
        nshift_gr = f.readInts ('i')
        
        nite            = f.readInts ('i')
        [time,dt,dtold] = f.readReals('d')
        [gamma,mu]      = f.readReals('d')
        [u_M,u_L,u_T]   = f.readReals('d')
        
        #Get grid position and state variables

        nbufx = nbuf
        nbufy = 0
        nbufz = 0
        
        if ndim>=2:
            nbufy = nbuf 
        if ndim>=3:
            nbufz = nbuf
        
        size   = (nx+2*nbufx)*(ny+2*nbufy)*(nz+2*nbufz)
        shape  = (nz+2*nbufz,ny+2*nbufy,nx+2*nbufx)
        vshape = (N_vit,nz+2*nbufz,ny+2*nbufy,nx+2*nbufx)
        
        self.rho  = np.transpose((f.readReals('d')).reshape( shape),(  2,1,0))
        
        if opt[1]==1 or opt[2]==1 or opt[3]==1 or opt[4]==1 or opt[5]==1 or opt[6]==1:
            self.rhou = np.transpose((f.readReals('d')).reshape(vshape),(3,2,1,0))
        if opt[2]==1 or opt[3]==1 or opt[4]==1 or opt[5]==1 or opt[6]==1:
            self.E    = np.transpose((f.readReals('d')).reshape( shape),(  2,1,0))
        
        if (opt[3]==1 or opt[4]==1 or opt[5]==1 or opt[6]==1) and nfx>0:
            fxshape = (nfx,nz+2*nbufz,ny+2*nbufy,nx+2*nbufx)
            self.fx = np.transpose((f.readReals('d')).reshape(fxshape),(3,2,1,0))

        if (opt[4]==1 or opt[5]==1 or opt[6]==1) and ray:
            frshape = (ndim,nz+2*nbufz,ny+2*nbufy,nx+2*nbufx)
            self.er = np.transpose((f.readReals('d')).reshape(  shape),(  2,1,0))
            self.fr = np.transpose((f.readReals('d')).reshape(frshape),(3,2,1,0))

        if (opt[5]==1 or opt[6]==1) and diffu:
            self.ed = np.transpose((f.readReals('d')).reshape(  shape),(  2,1,0))

        if opt[6]==1 and mhd:
            self.B  = np.transpose((f.readReals('d')).reshape(vshape),(3,2,1,0)) 
       
        #Close file
        f.close()
